#include "sqlnode.h"
#include "sqlfn.h"
#include "sqltype.h"

/* sparql.sql */

static const char *tbl0 = 
"create table DB.DBA.RDF_QUAD (\n"
"G IRI_ID,\n"
"S IRI_ID,\n"
"P IRI_ID,\n"
"O any,\n"
"primary key (G,S,P,O)\n"
")\n"
"create bitmap index RDF_QUAD_PGOS on DB.DBA.RDF_QUAD (P, G, O, S)\n";

static const char *tbl1 = 
"create table DB.DBA.RDF_URL (\n"
"RU_IID IRI_ID not null primary key,\n"
"RU_QNAME varchar )\n"
"create unique index RU_QNAME on DB.DBA.RDF_URL (RU_QNAME)\n";

static const char *tbl2 = 
"create table DB.DBA.RDF_OBJ (\n"
"RO_ID integeR primary key,\n"
"RO_VAL varchar,\n"
"RO_LONG long varchar\n"
")\n"
"create index RO_VAL on DB.DBA.RDF_OBJ (RO_VAL)\n";

static const char *tbl3 = 
"create table DB.DBA.RDF_DATATYPE (\n"
"RDT_IID IRI_ID not null primary key,\n"
"RDT_TWOBYTE integer not null unique,\n"
"RDT_QNAME varchar )\n";

static const char *tbl4 = 
"create table DB.DBA.RDF_LANGUAGE (\n"
"RL_ID varchar not null primary key,\n"
"RL_TWOBYTE integer not null unique )\n";

static const char *tbl5 = 
"create table DB.DBA.SYS_SPARQL_HOST (\n"
"SH_HOST	varchar not null primary key,\n"
"SH_GRAPH_URI	varchar,\n"
"SH_USER_URI	varchar )\n";

static const char *other0 = 
"sequence_set (\'RDF_URL_IID_NAMED\', 1000000, 1)\n";
static const char *other1 = 
"sequence_set (\'RDF_PREF_SEQ\', 1, 1)\n";
static const char *other2 = 
"sequence_set (\'RDF_URL_IID_BLANK\', 1000000000, 1)\n";
static const char *other3 = 
"sequence_set (\'RDF_RO_ID\', 1, 1)\n";
static const char *other4 = 
"sequence_set (\'RDF_DATATYPE_TWOBYTE\', 258, 1)\n";
static const char *other5 = 
"sequence_set (\'RDF_LANGUAGE_TWOBYTE\', 258, 1)\n";
static const char *proc0 = 
"create procedure DB.DBA.RDF_GLOBAL_RESET ()\n"
"{\n"
"\n"
"__atomic (1);\n"
"iri_id_cache_flush ();\n"
"delete from DB.DBA.RDF_QUAD;\n"
"delete from DB.DBA.RDF_URL;\n"
"delete from DB.DBA.RDF_IRI;\n"
"delete from DB.DBA.RDF_PREFIX;\n"
"delete from DB.DBA.RDF_OBJ;\n"
"delete from DB.DBA.RDF_DATATYPE;\n"
"delete from DB.DBA.RDF_LANGUAGE;\n"
"sequence_set (\'RDF_URL_IID_NAMED\', 1000000, 0);\n"
"sequence_set (\'RDF_URL_IID_BLANK\', 1000000000, 0);\n"
"sequence_set (\'RDF_PREF_SEQ\', 1, 0);\n"
"sequence_set (\'RDF_RO_ID\', 1, 0);\n"
"sequence_set (\'RDF_DATATYPE_TWOBYTE\', 258, 0);\n"
"sequence_set (\'RDF_LANGUAGE_TWOBYTE\', 258, 0);\n"
"__atomic (0);\n"
"\n"
"TTLP (\n"
"cast ( XML_URI_GET (\n"
"\'http://www.openlinksw.com/sparql/virtrdf-data-formats.ttl\', \'\' ) as varchar ),\n"
"\'\', \'http://www.openlinksw.com/schemas/virtrdf#\' );\n"
"TTLP (\'\n"
"@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .\n"
"@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .\n"
"@prefix owl: <http://www.w3.org/2002/07/owl#> .\n"
"@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .\n"
"@prefix virtrdf: <http://www.openlinksw.com/schemas/virtrdf#> .\n"
"@prefix rdfdf: <http://www.openlinksw.com/virtrdf-data-formats#> .\n"
"@prefix atom: <http://atomowl.org/ontologies/atomrdf#> .\n"
"\n"
"virtrdf:DefaultQuadStorage\n"
"rdf:type virtrdf:QuadStorage ;\n"
"virtrdf:qsUserMaps virtrdf:DefaultQuadStorage-UserMaps ;\n"
"virtrdf:qsDefaultMap virtrdf:DefaultQuadMap .\n"
"virtrdf:DefaultQuadStorage-UserMaps\n"
"rdf:type virtrdf:array-of-QuadMap .\n"
"\', \'\', \'http://www.openlinksw.com/schemas/virtrdf#\' );\n"
"delete from SYS_HTTP_SPONGE where HS_PARSER = \'DB.DBA.RDF_LOAD_HTTP_RESPONSE\';\n"
"\n"
"}\n"
"--src sparql.sql:101\n";

static const char *proc1 = 
"create function DB.DBA.RDF_MAKE_IID_OF_QNAME (in qname varchar) returns IRI_ID\n"
"{\n"
"return iri_to_id (qname, 1);\n"
"}\n"
"--src sparql.sql:149\n";

static const char *proc2 = 
"create procedure DB.DBA.RDF_MIGRATE_URL_TO_IRI ()\n"
"{\n"
"declare ctr integer;\n"
"declare iid IRI_ID;\n"
"declare qname, tail varchar;\n"
"__atomic (1);\n"
"declare exit handler for sqlstate \'*\'\n"
"{\n"
"__atomic (0);\n"
"resignal;\n"
"};\n"
"if (not exists (select top 1 1 from DB.DBA.RDF_URL))\n"
"goto rdf_url_is_empty;\n"
"\n"
"again:\n"
"ctr := 0;\n"
"declare cr cursor for select RU_IID, RU_QNAME from DB.DBA.RDF_URL for update;\n"
"open cr (exclusive);\n"
"whenever not found goto rdf_url_is_empty;\n"
"\n"
"next_fetch:\n"
"fetch cr into iid, qname;\n"
"declare parts any;\n"
"declare prefix_id integer;\n"
"parts := iri_to_rdf_prefix_and_local (qname);\n"
"if (parts is null)\n"
"{\n"
"log_message (\'The function DB.DBA.RDF_MIGRATE_URL_TO_IRI() can not migrate an abnormally long URL.\');\n"
"log_message (\'This means that this version of Virtuoso server can not process RDF data stored in the database.\');\n"
"log_message (\'To fix the problem, remove the transaction log and start previous version of Virtuoso.\');\n"
"log_message (\'The example of unsupported RDF node URL is\');\n"
"log_message (qname);\n"
"log_message (\'This error is critical. The server will now exit. Sorry.\');\n"
"raw_exit();\n"
"}\n"
"prefix_id := (select RP_ID from DB.DBA.RDF_PREFIX where RP_NAME = parts[0]);\n"
"if (prefix_id is null)\n"
"{\n"
"prefix_id := sequence_next (\'RDF_PREF_SEQ\');\n"
"insert into DB.DBA.RDF_PREFIX (RP_NAME, RP_ID) values (parts[0], prefix_id);\n"
"}\n"
"tail := parts[1];\n"
"tail[0] := bit_and (255, bit_shift (prefix_id, -24));\n"
"tail[1] := bit_and (255, bit_shift (prefix_id, -16));\n"
"tail[2] := bit_and (255, bit_shift (prefix_id, -8));\n"
"tail[3] := bit_and (255, prefix_id);\n"
"insert into DB.DBA.RDF_IRI (RI_NAME, RI_ID) values (tail, iid);\n"
"delete from DB.DBA.RDF_URL where current of cr;\n"
"ctr := ctr + 1;\n"
"if (ctr > 1000)\n"
"{\n"
"commit work;\n"
"close cr;\n"
"goto again;\n"
"}\n"
"goto next_fetch;\n"
"\n"
"rdf_url_is_empty:\n"
"\n"
"if (exists (select top 1 1 from DB.DBA.RDF_DATATYPE where RDT_TWOBYTE < 512 and RDT_QNAME[0] < 32))\n"
"update DB.DBA.RDF_DATATYPE set RDT_QNAME = id_to_iri (RDT_IID) where RDT_QNAME[0] < 32;\n"
"commit work;\n"
"__atomic (0);\n"
"}\n"
"--src sparql.sql:155\n";

static const char *other6 = 
"DB.DBA.RDF_MIGRATE_URL_TO_IRI ()\n";
static const char *proc3 = 
"create function DB.DBA.RDF_MAKE_IID_OF_QNAME_SAFE (in qname any) returns IRI_ID\n"
"{\n"
"whenever sqlstate \'*\' goto retnull;\n"
"return iri_to_id (qname, 1);\n"
"retnull:\n"
"return null;\n"
"}\n"
"--src sparql.sql:224\n";

static const char *proc4 = 
"create function DB.DBA.RDF_MAKE_IID_OF_LONG (in qname any) returns IRI_ID\n"
"{\n"
"if (193 = __tag (qname))\n"
"qname := DB.DBA.RDF_STRSQLVAL_OF_LONG (qname);\n"
"whenever sqlstate \'*\' goto retnull;\n"
"return iri_to_id (qname, 1);\n"
"retnull:\n"
"return null;\n"
"}\n"
"--src sparql.sql:233\n";

static const char *proc5 = 
"create function DB.DBA.RDF_QNAME_OF_IID (in iid IRI_ID) returns varchar\n"
"{\n"
"if (iid >= #i1000000000)\n"
"return sprintf (\'nodeID://%d\', iri_id_num (iid));\n"
"return id_to_iri (iid);\n"
"}\n"
"--src sparql.sql:244\n";

static const char *proc6 = 
"create function DB.DBA.RDF_IID_OF_QNAME_SAFE (in qname varchar) returns IRI_ID\n"
"{\n"
"whenever sqlstate \'*\' goto retnull;\n"
"return iri_to_id (qname, 0);\n"
"retnull:\n"
"return null;\n"
"}\n"
"--src sparql.sql:252\n";

static const char *proc7 = 
"create function DB.DBA.RDF_IID_OF_QNAME (in qname varchar) returns IRI_ID\n"
"{\n"
"return iri_to_id (qname, 0);\n"
"}\n"
"--src sparql.sql:261\n";

static const char *proc8 = 
"create function DB.DBA.RDF_MAKE_GRAPH_IIDS_OF_QNAMES (in qnames any) returns any\n"
"{\n"
"if (193 <> __tag (qnames))\n"
"return vector ();\n"
"declare res_acc any;\n"
"vectorbld_init (res_acc);\n"
"foreach (any qname in qnames) do\n"
"{\n"
"declare iid IRI_ID;\n"
"whenever sqlstate \'*\' goto skip_acc;\n"
"iid := iri_to_id (qname, 0);\n"
"if (iid is not null)\n"
"vectorbld_acc (res_acc, iid);\n"
"skip_acc: ;\n"
"}\n"
"vectorbld_final (res_acc);\n"
"return res_acc;\n"
"}\n"
"--src sparql.sql:267\n";

static const char *proc9 = 
"create function DB.DBA.RDF_TWOBYTE_OF_DATATYPE (in iid IRI_ID) returns integer\n"
"{\n"
"declare res integer;\n"
"if (iid is null)\n"
"return 257;\n"
"if (not isiri_id (iid))\n"
"{\n"
"declare new_iid IRI_ID;\n"
"new_iid := iri_to_id (iid, 1);\n"
"if (new_iid is NULL or new_iid >= #i1000000000)\n"
"signal (\'RDFXX\', \'Invalid datatype IRI_ID passes as an argument to DB.DBA.RDF_TWOBYTE_OF_DATATYPE()\');\n"
"iid := new_iid;\n"
"}\n"
"whenever not found goto mknew;\n"
"set isolation=\'committed\';\n"
"select RDT_TWOBYTE into res from DB.DBA.RDF_DATATYPE where RDT_IID = iid;\n"
"return res;\n"
"\n"
"mknew:\n"
"set isolation=\'serializable\';\n"
"declare tb_cr cursor for select RDT_TWOBYTE from DB.DBA.RDF_DATATYPE where RDT_IID = iid;\n"
"open tb_cr (exclusive);\n"
"whenever not found goto mknew_ser;\n"
"fetch tb_cr into res;\n"
"return res;\n"
"\n"
"mknew_ser:\n"
"res := sequence_next (\'RDF_DATATYPE_TWOBYTE\');\n"
"if (0 = bit_and (res, 255))\n"
"res := sequence_next (\'RDF_DATATYPE_TWOBYTE\');\n"
"insert into DB.DBA.RDF_DATATYPE\n"
"(RDT_IID, RDT_TWOBYTE, RDT_QNAME)\n"
"values (iid, res, DB.DBA.RDF_QNAME_OF_IID (iid));\n"
"return res;\n"
"}\n"
"--src sparql.sql:290\n";

static const char *proc10 = 
"create function DB.DBA.RDF_TWOBYTE_OF_LANGUAGE (in id varchar) returns integer\n"
"{\n"
"declare res integer;\n"
"if (id is null)\n"
"return 257;\n"
"whenever not found goto mknew;\n"
"set isolation=\'committed\';\n"
"select RL_TWOBYTE into res from DB.DBA.RDF_LANGUAGE where RL_ID = id;\n"
"return res;\n"
"\n"
"mknew:\n"
"set isolation=\'serializable\';\n"
"declare tb_cr cursor for select RL_TWOBYTE from DB.DBA.RDF_LANGUAGE where RL_ID = id;\n"
"open tb_cr (exclusive);\n"
"whenever not found goto mknew_ser;\n"
"fetch tb_cr into res;\n"
"return res;\n"
"\n"
"mknew_ser:\n"
"res := sequence_next (\'RDF_LANGUAGE_TWOBYTE\');\n"
"if (0 = bit_and (res, 255))\n"
"res := sequence_next (\'RDF_LANGUAGE_TWOBYTE\');\n"
"insert into DB.DBA.RDF_LANGUAGE (RL_ID, RL_TWOBYTE) values (id, res);\n"
"return res;\n"
"}\n"
"--src sparql.sql:327\n";

static const char *proc11 = 
"create function DB.DBA.RDF_DATATYPE_OF_TAG (in t integer)\n"
"{\n"
"if (t = 182)\n"
"return UNAME\'http://www.w3.org/2001/XMLSchema#string\';\n"
"if (t = 189)\n"
"return UNAME\'http://www.w3.org/2001/XMLSchema#integer\';\n"
"if (t = 211)\n"
"return UNAME\'http://www.w3.org/2001/XMLSchema#dateTime\';\n"
"if (t = 191)\n"
"return UNAME\'http://www.w3.org/2001/XMLSchema#double\';\n"
"if (t = 204) /* DB NULL */\n"
"return NULL;\n"
"signal (\'RDFXX\', sprintf (\'Unsupported tag in DB.DBA.RDF_DATATYPE_OF_TAG(): %d\', t));\n"
"}\n"
"--src sparql.sql:354\n";

static const char *proc12 = 
"create function DB.DBA.RQ_LONG_OF_O (in o_col any) returns any\n"
"{\n"
"declare t, l, len integer;\n"
"if (not isstring (o_col))\n"
"return o_col;\n"
"l := 20;\n"
"len := length (o_col);\n"
"if (len = (l + 9))\n"
"{\n"
"declare v2 varchar;\n"
"declare id int;\n"
"id := o_col[l+3] + 256 * (o_col [l+4] + 256 * (o_col [l+5] + 256 * o_col [l+6]));\n"
"v2 := (select case (isnull (RO_LONG)) when 0 then blob_to_string (RO_LONG) else RO_VAL end from DB.DBA.RDF_OBJ where RO_ID = id);\n"
"if (v2 is null)\n"
"signal (\'RDFXX\', sprintf (\'Integrity violation in DB.DBA.RQ_LONG_OF_O, bad id %d\', id));\n"
"return vector (\n"
"o_col[0] + 256 * (o_col[1]),\n"
"v2,\n"
"o_col[len-2] + 256 * (o_col[len-1]),\n"
"id,\n"
"o_col );\n"
"}\n"
"if ((len > (l + 5)) or (len < 5))\n"
"signal (\'RDFXX\', sprintf (\'Integrity violation in DB.DBA.RQ_LONG_OF_O, bad string \"%s\"\', o_col));\n"
"return vector (\n"
"o_col[0] + 256 * (o_col[1]),\n"
"subseq (o_col, 2, len-3),\n"
"o_col[len-2] + 256 * (o_col[len-1]),\n"
"0,\n"
"o_col );\n"
"}\n"
"--src sparql.sql:373\n";

static const char *proc13 = 
"create function DB.DBA.RQ_SQLVAL_OF_O (in o_col any) returns any\n"
"{\n"
"declare t, l, len integer;\n"
"if (isiri_id (o_col))\n"
"{\n"
"declare res varchar;\n"
"if (o_col >= #i1000000000)\n"
"return sprintf (\'nodeID://%d\', iri_id_num (o_col));\n"
"res := id_to_iri (o_col);\n"
"\n"
"if (res is null)\n"
"signal (\'RDFXX\', \'Wrong iid in DB.DBA.RQ_SQLVAL_OF_O()\');\n"
"return res;\n"
"}\n"
"if (not isstring (o_col))\n"
"return o_col;\n"
"l := 20;\n"
"len := length (o_col);\n"
"if (len = (l + 9))\n"
"{\n"
"declare v2 varchar;\n"
"declare id int;\n"
"id := o_col[l+3] + 256 * (o_col [l+4] + 256 * (o_col [l+5] + 256 * o_col [l+6]));\n"
"v2 := (select case (isnull (RO_LONG)) when 0 then blob_to_string (RO_LONG) else RO_VAL end from DB.DBA.RDF_OBJ where RO_ID = id);\n"
"if (v2 is null)\n"
"signal (\'RDFXX\', sprintf (\'Integrity violation in DB.DBA.RQ_SQLVAL_OF_O, bad id %d\', id));\n"
"return v2;\n"
"}\n"
"if ((len > (l + 5)) or (len < 5))\n"
"signal (\'RDFXX\', sprintf (\'Integrity violation in DB.DBA.RQ_LONG_OF_O, bad string \"%s\"\', o_col));\n"
"return subseq (o_col, 2, len-3);\n"
"}\n"
"--src sparql.sql:406\n";

static const char *proc14 = 
"create function DB.DBA.RQ_BOOL_OF_O (in o_col any) returns any\n"
"{\n"
"declare t, l, len integer;\n"
"if (isiri_id (o_col))\n"
"return NULL;\n"
"if (isinteger (o_col))\n"
"{\n"
"if (o_col)\n"
"return 1;\n"
"return 0;\n"
"}\n"
"if (not isstring (o_col))\n"
"{\n"
"if (o_col is null)\n"
"return null;\n"
"return neq (o_col, 0.0);\n"
"}\n"
"l := 20;\n"
"len := length (o_col);\n"
"if (((len > (l + 5)) and (len <> (l + 9))) or (len < 5))\n"
"signal (\'RDFXX\', sprintf (\'Integrity violation in DB.DBA.RQ_BOOL_OF_O, bad string \"%s\"\', o_col));\n"
"\n"
"declare twobyte integer;\n"
"declare dtqname any;\n"
"twobyte := o_col[0] + 256 * (o_col[1]);\n"
"if (257 = twobyte)\n"
"goto type_ok;\n"
"whenever not found goto badtype;\n"
"select RDT_QNAME into dtqname from DB.DBA.RDF_DATATYPE where RDT_TWOBYTE = twobyte;\n"
"if (dtqname <> UNAME\'http://www.w3.org/2001/XMLSchema#string\')\n"
"return null;\n"
"\n"
"type_ok:\n"
"return case (len) when 5 then 0 else 1 end;\n"
"\n"
"badtype:\n"
"signal (\'RDFXX\', sprintf (\'Unknown datatype in DB.DBA.RQ_BOOL_OF_O, bad string \"%s\"\', o_col));\n"
"}\n"
"--src sparql.sql:440\n";

static const char *proc15 = 
"create function DB.DBA.RQ_IID_OF_O (in shortobj any) returns IRI_ID\n"
"{\n"
"if (not isiri_id (shortobj))\n"
"return NULL;\n"
"return shortobj;\n"
"}\n"
"--src sparql.sql:480\n";

static const char *proc16 = 
"create function DB.DBA.RQ_O_IS_LIT (in shortobj any) returns integer\n"
"{\n"
"if (isiri_id (shortobj))\n"
"return 0;\n"
"return 1;\n"
"}\n"
"--src sparql.sql:488\n";

static const char *proc17 = 
"create function DB.DBA.RDF_MAKE_RO_ID_OF_STRING (in v varchar) returns integer\n"
"{\n"
"declare llong, id int;\n"
"declare tridgell varchar;\n"
"llong := 1010;\n"
"if (length (v) > llong)\n"
"{\n"
"tridgell := tridgell32 (v, 1);\n"
"set isolation=\'committed\';\n"
"id := (select RO_ID from DB.DBA.RDF_OBJ where RO_VAL = tridgell and blob_to_string (RO_LONG) = v);\n"
"if (id is null)\n"
"{\n"
"set isolation=\'serializable\';\n"
"declare id_cr cursor for select RO_ID from DB.DBA.RDF_OBJ where RO_VAL = tridgell and blob_to_string (RO_LONG) = v;\n"
"open id_cr (exclusive);\n"
"whenever not found goto new_id_for_long;\n"
"fetch id_cr into id;\n"
"return id;\n"
"new_id_for_long:\n"
"id := sequence_next (\'RDF_RO_ID\');\n"
"insert into DB.DBA.RDF_OBJ (RO_ID, RO_VAL, RO_LONG) values (id, tridgell, v);\n"
"}\n"
"}\n"
"else\n"
"{\n"
"set isolation=\'committed\';\n"
"id := (select RO_ID from DB.DBA.RDF_OBJ where RO_VAL = v);\n"
"if (id is null)\n"
"{\n"
"set isolation=\'serializable\';\n"
"declare id_cr cursor for select RO_ID from DB.DBA.RDF_OBJ where RO_VAL = v;\n"
"open id_cr (exclusive);\n"
"whenever not found goto new_id_for_short;\n"
"fetch id_cr into id;\n"
"return id;\n"
"new_id_for_short:\n"
"id := sequence_next (\'RDF_RO_ID\');\n"
"insert into DB.DBA.RDF_OBJ (RO_ID, RO_VAL) values (id, v);\n"
"}\n"
"}\n"
"return id;\n"
"}\n"
"--src sparql.sql:499\n";

static const char *proc18 = 
"create function DB.DBA.RDF_FIND_RO_ID_OF_STRING (in v varchar) returns integer\n"
"{\n"
"declare llong int;\n"
"declare tridgell varchar;\n"
"llong := 1010;\n"
"if (length (v) > llong)\n"
"{\n"
"tridgell := tridgell32 (v, 1);\n"
"return (select RO_ID from DB.DBA.RDF_OBJ where RO_VAL = tridgell and blob_to_string (RO_LONG) = v);\n"
"}\n"
"else\n"
"{\n"
"return (select RO_ID from DB.DBA.RDF_OBJ where RO_VAL = v);\n"
"}\n"
"}\n"
"--src sparql.sql:543\n";

static const char *proc19 = 
"create function DB.DBA.RDF_MAKE_OBJ_OF_SQLVAL (in v any) returns any\n"
"{\n"
"declare l, t int;\n"
"t := __tag (v);\n"
"if (not t in (126, 182, 217, 225, 230))\n"
"return v;\n"
"if (225 = t)\n"
"v := charset_recode (v, \'_WIDE_\', \'UTF-8\');\n"
"else if (217 = t or 126 = t)\n"
"v := cast (v as varchar);\n"
"else if (230 = t)\n"
"v := serialize_to_UTF8_xml (v);\n"
"l := 20;\n"
"if (length (v) > l)\n"
"{\n"
"declare v2 varchar;\n"
"declare id int;\n"
"id := DB.DBA.RDF_MAKE_RO_ID_OF_STRING (v);\n"
"v2 := concat (\'\\001\\001\', subseq (v, 0, l), \'\\0ABCD\\001\\001\');\n"
"v2 [l+3] := bit_and (id, 255);\n"
"v2 [l+4] := bit_and (bit_shift (id, -8), 255);\n"
"v2 [l+5] := bit_and (bit_shift (id, -16), 255);\n"
"v2 [l+6] := bit_and (bit_shift (id, -24), 255);\n"
"return v2;\n"
"}\n"
"return concat (\'\\001\\001\', v, \'\\0\\001\\001\');\n"
"}\n"
"--src sparql.sql:560\n";

static const char *proc20 = 
"create function DB.DBA.RDF_MAKE_OBJ_OF_TYPEDSQLVAL (in v any, in dt_iid IRI_ID, in lang varchar) returns any\n"
"{\n"
"declare l, t, dt_twobyte, lang_twobyte int;\n"
"declare dt_s, lang_s varchar;\n"
"t := __tag (v);\n"
"if (not t in (126, 182, 217, 225, 230))\n"
"signal (\'RDFXX\', \'DB.DBA.RDF_MAKE_OBJ_OF_TYPEDSQLVAL() accepts only string representations of typed values\');\n"
"if (225 = t)\n"
"v := charset_recode (v, \'_WIDE_\', \'UTF-8\');\n"
"else if (217 = t or 126 = t)\n"
"v := cast (v as varchar);\n"
"else if (230 = t)\n"
"v := serialize_to_UTF8_xml (v);\n"
"dt_s := \'\\001\\001\';\n"
"if (dt_iid is not null)\n"
"{\n"
"dt_twobyte := DB.DBA.RDF_TWOBYTE_OF_DATATYPE (dt_iid);\n"
"dt_s[0] := bit_and (dt_twobyte, 255);\n"
"dt_s[1] := bit_and (bit_shift (dt_twobyte, -8), 255);\n"
"}\n"
"lang_s := \'\\001\\001\';\n"
"if (lang is not null)\n"
"{\n"
"lang_twobyte := DB.DBA.RDF_TWOBYTE_OF_LANGUAGE (lang);\n"
"lang_s[0] := bit_and (lang_twobyte, 255);\n"
"lang_s[1] := bit_and (bit_shift (lang_twobyte, -8), 255);\n"
"}\n"
"l := 20;\n"
"if (length (v) > l)\n"
"{\n"
"declare v2 varchar;\n"
"declare id int;\n"
"id := DB.DBA.RDF_MAKE_RO_ID_OF_STRING (v);\n"
"v2 := concat (dt_s, subseq (v, 0, l), \'\\0ABCD\', lang_s);\n"
"v2 [l+3] := bit_and (id, 255);\n"
"v2 [l+4] := bit_and (bit_shift (id, -8), 255);\n"
"v2 [l+5] := bit_and (bit_shift (id, -16), 255);\n"
"v2 [l+6] := bit_and (bit_shift (id, -24), 255);\n"
"return v2;\n"
"}\n"
"return concat (dt_s, v, \'\\0\', lang_s);\n"
"}\n"
"--src sparql.sql:589\n";

static const char *proc21 = 
"create function DB.DBA.RDF_MAKE_OBJ_OF_TYPEDSQLVAL_STRINGS (\n"
"in o_val any, in o_type varchar, in o_lang varchar ) returns any\n"
"{\n"
"if (\'http://www.w3.org/2001/XMLSchema#boolean\' = o_type)\n"
"{\n"
"if ((\'true\' = o_val) or (\'1\' = o_val))\n"
"return 1;\n"
"else if ((\'false\' = o_val) or (\'0\' = o_val))\n"
"return 0;\n"
"else signal (\'RDFXX\', \'Invalid notation of boolean literal\');\n"
"}\n"
"if (\'http://www.w3.org/2001/XMLSchema#dateTime\' = o_type)\n"
"return __xqf_str_parse (\'dateTime\', o_val);\n"
"if (\'http://www.w3.org/2001/XMLSchema#double\' = o_type)\n"
"return cast (o_val as double precision);\n"
"if (\'http://www.w3.org/2001/XMLSchema#float\' = o_type)\n"
"return cast (o_val as float);\n"
"if (\'http://www.w3.org/2001/XMLSchema#integer\' = o_type)\n"
"return cast (o_val as int);\n"
"if (isstring(o_type) or isstring (o_lang))\n"
"{\n"
"return DB.DBA.RDF_MAKE_OBJ_OF_TYPEDSQLVAL (\n"
"o_val,\n"
"DB.DBA.RDF_MAKE_IID_OF_QNAME (o_type),\n"
"o_lang );\n"
"}\n"
"return DB.DBA.RDF_MAKE_OBJ_OF_SQLVAL (o_val);\n"
"}\n"
"--src sparql.sql:633\n";

static const char *proc22 = 
"create function DB.DBA.RDF_LONG_OF_OBJ (in shortobj any) returns any\n"
"{\n"
"declare t, l, len integer;\n"
"if (not isstring (shortobj))\n"
"return shortobj;\n"
"l := 20;\n"
"len := length (shortobj);\n"
"if (len = (l + 9))\n"
"{\n"
"declare v2 varchar;\n"
"declare id int;\n"
"id := shortobj[l+3] + 256 * (shortobj [l+4] + 256 * (shortobj [l+5] + 256 * shortobj [l+6]));\n"
"v2 := (select case (isnull (RO_LONG)) when 0 then blob_to_string (RO_LONG) else RO_VAL end from DB.DBA.RDF_OBJ where RO_ID = id);\n"
"if (v2 is null)\n"
"signal (\'RDFXX\', sprintf (\'Integrity violation in DB.DBA.RQ_LONG_OF_OBJ, bad id %d\', id));\n"
"return vector (\n"
"shortobj[0] + 256 * (shortobj[1]),\n"
"v2,\n"
"shortobj[len-2] + 256 * (shortobj[len-1]),\n"
"id,\n"
"shortobj );\n"
"}\n"
"if ((len > (l + 5)) or (len < 5))\n"
"signal (\'RDFXX\', sprintf (\'Integrity violation in DB.DBA.RQ_LONG_OF_OBJ, bad string \"%s\"\', shortobj));\n"
"return vector (\n"
"shortobj[0] + 256 * (shortobj[1]),\n"
"subseq (shortobj, 2, len-3),\n"
"shortobj[len-2] + 256 * (shortobj[len-1]),\n"
"0,\n"
"shortobj );\n"
"}\n"
"--src sparql.sql:663\n";

static const char *proc23 = 
"create function DB.DBA.RDF_DATATYPE_OF_OBJ (in shortobj any) returns any\n"
"{\n"
"declare l, len, twobyte integer;\n"
"declare res any;\n"
"if (not isstring (shortobj))\n"
"{\n"
"if (isiri_id (shortobj))\n"
"return NULL;\n"
"return DB.DBA.RDF_DATATYPE_OF_TAG (__tag (shortobj));\n"
"}\n"
"l := 20;\n"
"len := length (shortobj);\n"
"if (((len <> (l + 9)) and (len > (l + 5))) or (len < 5))\n"
"signal (\'RDFXX\', sprintf (\'Integrity violation in DB.DBA.RQ_DATATYPE_OF_OBJ, bad string \"%s\"\', shortobj));\n"
"twobyte := shortobj[0] + 256 * (shortobj[1]);\n"
"if (257 = twobyte)\n"
"return null;\n"
"whenever not found goto badtype;\n"
"select RDT_QNAME into res from DB.DBA.RDF_DATATYPE where RDT_TWOBYTE = twobyte;\n"
"return res;\n"
"\n"
"badtype:\n"
"signal (\'RDFXX\', sprintf (\'Unknown datatype in DB.DBA.RQ_DATATYPE_OF_OBJ, bad string \"%s\"\', shortobj));\n"
"}\n"
"--src sparql.sql:696\n";

static const char *proc24 = 
"create function DB.DBA.RDF_LANGUAGE_OF_OBJ (in shortobj any) returns any\n"
"{\n"
"declare l, len, twobyte integer;\n"
"declare res varchar;\n"
"if (not isstring (shortobj))\n"
"return null;\n"
"l := 20;\n"
"len := length (shortobj);\n"
"if (((len <> (l + 9)) and (len > (l + 5))) or (len < 5))\n"
"signal (\'RDFXX\', sprintf (\'Integrity violation in DB.DBA.RDF_LANGUAGE_OF_OBJ, bad string \"%s\"\', shortobj));\n"
"twobyte := shortobj[len-2] + 256 * (shortobj[len-1]);\n"
"if (257 = twobyte)\n"
"return null;\n"
"whenever not found goto badtype;\n"
"select RL_ID into res from DB.DBA.RDF_LANGUAGE where RL_TWOBYTE = twobyte;\n"
"return res;\n"
"\n"
"badtype:\n"
"signal (\'RDFXX\', sprintf (\'Unknown language in DB.DBA.RDF_LANGUAGE_OF_OBJ, bad string \"%s\"\', shortobj));\n"
"}\n"
"--src sparql.sql:722\n";

static const char *proc25 = 
"create function DB.DBA.RDF_SQLVAL_OF_OBJ (in shortobj any) returns any\n"
"{\n"
"declare t, l, len integer;\n"
"if (isiri_id (shortobj))\n"
"{\n"
"if (shortobj >= #i1000000000)\n"
"return sprintf (\'nodeID://%d\', iri_id_num (shortobj));\n"
"else\n"
"{\n"
"declare res varchar;\n"
"res := id_to_iri (shortobj);\n"
"\n"
"if (res is null)\n"
"signal (\'RDFXX\', \'Wrong iid in DB.DBA.RDF_SQLVAL_OF_OBJ()\');\n"
"return res;\n"
"}\n"
"}\n"
"if (not isstring (shortobj))\n"
"return shortobj;\n"
"l := 20;\n"
"len := length (shortobj);\n"
"if (len = (l + 9))\n"
"{\n"
"declare v2 varchar;\n"
"declare id int;\n"
"id := shortobj[l+3] + 256 * (shortobj [l+4] + 256 * (shortobj [l+5] + 256 * shortobj [l+6]));\n"
"v2 := (select case (isnull (RO_LONG)) when 0 then blob_to_string (RO_LONG) else RO_VAL end from DB.DBA.RDF_OBJ where RO_ID = id);\n"
"if (v2 is null)\n"
"signal (\'RDFXX\', sprintf (\'Integrity violation in DB.DBA.RQ_SQLVAL_OF_O, bad id %d\', id));\n"
"return v2;\n"
"}\n"
"if ((len > (l + 5)) or (len < 5))\n"
"signal (\'RDFXX\', sprintf (\'Integrity violation in DB.DBA.RQ_LONG_OF_O, bad string \"%s\"\', shortobj));\n"
"return subseq (shortobj, 2, len-3);\n"
"}\n"
"--src sparql.sql:744\n";

static const char *proc26 = 
"create function DB.DBA.RDF_BOOL_OF_OBJ (in shortobj any) returns any\n"
"{\n"
"declare t, l, len integer;\n"
"if (shortobj is null)\n"
"return null;\n"
"if (isiri_id (shortobj))\n"
"return null;\n"
"if (isinteger (shortobj))\n"
"{\n"
"if (shortobj)\n"
"return 1;\n"
"return 0;\n"
"}\n"
"if (not isstring (shortobj))\n"
"{\n"
"if (shortobj is null)\n"
"return null;\n"
"return neq (shortobj, 0.0);\n"
"}\n"
"l := 20;\n"
"len := length (shortobj);\n"
"if (((len > (l + 5)) and (len <> (l + 9))) or (len < 5))\n"
"signal (\'RDFXX\', sprintf (\'Integrity violation in DB.DBA.RDF_BOOL_OF_OBJ, bad string \"%s\"\', shortobj));\n"
"\n"
"declare twobyte integer;\n"
"declare dtqname any;\n"
"twobyte := shortobj[0] + 256 * (shortobj[1]);\n"
"if (257 = twobyte)\n"
"goto type_ok;\n"
"whenever not found goto badtype;\n"
"select RDT_QNAME into dtqname from DB.DBA.RDF_DATATYPE where RDT_TWOBYTE = twobyte;\n"
"if (dtqname <> UNAME\'http://www.w3.org/2001/XMLSchema#string\')\n"
"return null;\n"
"\n"
"type_ok:\n"
"return case (len) when 5 then 0 else 1 end;\n"
"\n"
"badtype:\n"
"signal (\'RDFXX\', sprintf (\'Unknown datatype in DB.DBA.RDF_BOOL_OF_OBJ, bad string \"%s\"\', shortobj));\n"
"}\n"
"--src sparql.sql:781\n";

static const char *proc27 = 
"create function DB.DBA.RDF_QNAME_OF_OBJ (in shortobj any) returns varchar\n"
"{\n"
"if (isiri_id (shortobj))\n"
"{\n"
"if (shortobj >= #i1000000000)\n"
"return sprintf (\'nodeID://%d\', iri_id_num (shortobj));\n"
"else\n"
"{\n"
"declare res varchar;\n"
"res := id_to_iri (shortobj);\n"
"\n"
"if (res is null)\n"
"signal (\'RDFXX\', \'Wrong iid in DB.DBA.RDF_QNAME_OF_OBJ()\');\n"
"return res;\n"
"}\n"
"}\n"
"return NULL;\n"
"}\n"
"--src sparql.sql:823\n";

static const char *proc28 = 
"create function DB.DBA.RDF_STRSQLVAL_OF_OBJ (in shortobj any)\n"
"{\n"
"declare t, l, len integer;\n"
"if (isiri_id (shortobj))\n"
"{\n"
"declare res varchar;\n"
"if (shortobj >= #i1000000000)\n"
"return sprintf (\'nodeID://%d\', iri_id_num (shortobj));\n"
"res := id_to_iri (shortobj);\n"
"\n"
"if (res is null)\n"
"signal (\'RDFXX\', \'Wrong iid in DB.DBA.RDF_STRSQLVAL_OF_OBJ()\');\n"
"return res;\n"
"}\n"
"if (not isstring (shortobj))\n"
"{\n"
"if (211 = __tag (shortobj))\n"
"{\n"
"declare vc varchar;\n"
"vc := cast (shortobj as varchar);\n"
"return replace (vc, \' \', \'T\');\n"
"}\n"
"if (225 =  __tag (shortobj))\n"
"return charset_recode (shortobj, \'_WIDE_\', \'UTF-8\');\n"
"return cast (shortobj as varchar);\n"
"}\n"
"l := 20;\n"
"len := length (shortobj);\n"
"if (len = (l + 9))\n"
"{\n"
"declare v2 varchar;\n"
"declare id int;\n"
"id := shortobj[l+3] + 256 * (shortobj [l+4] + 256 * (shortobj [l+5] + 256 * shortobj [l+6]));\n"
"v2 := (select case (isnull (RO_LONG)) when 0 then blob_to_string (RO_LONG) else RO_VAL end from DB.DBA.RDF_OBJ where RO_ID = id);\n"
"if (v2 is null)\n"
"signal (\'RDFXX\', sprintf (\'Integrity violation in DB.DBA.RDF_STRSQLVAL_OF_OBJ, bad id %d\', id));\n"
"return v2;\n"
"}\n"
"if ((len > (l + 5)) or (len < 5))\n"
"signal (\'RDFXX\', sprintf (\'Integrity violation in DB.DBA.RDF_STRSQLVAL_OF_OBJ, bad string \"%s\"\', shortobj));\n"
"return subseq (shortobj, 2, len-3);\n"
"}\n"
"--src sparql.sql:843\n";

static const char *proc29 = 
"create function DB.DBA.RDF_OBJ_OF_LONG (in longobj any) returns any\n"
"{\n"
"if (193 <> __tag(longobj))\n"
"return longobj;\n"
"if (length (longobj) > 5)\n"
"return call (longobj[5] || \'_OBJ_OF_LONG\')(longobj);\n"
"else if ((length (longobj) = 5) and isstring (longobj[4]))\n"
"return longobj[4];\n"
"else\n"
"{\n"
"declare l, dt_twobyte, lang_twobyte int;\n"
"declare v, dt_s, lang_s varchar;\n"
"v := longobj[1];\n"
"dt_s := \'\\001\\001\';\n"
"if (257 <> longobj[0])\n"
"{\n"
"dt_twobyte := longobj[0];\n"
"dt_s[0] := bit_and (dt_twobyte, 255);\n"
"dt_s[1] := bit_and (bit_shift (dt_twobyte, -8), 255);\n"
"}\n"
"lang_s := \'\\001\\001\';\n"
"if (257 <> longobj[2])\n"
"{\n"
"lang_twobyte := longobj[2];\n"
"lang_s[0] := bit_and (lang_twobyte, 255);\n"
"lang_s[1] := bit_and (bit_shift (lang_twobyte, -8), 255);\n"
"}\n"
"l := 20;\n"
"if (length (v) > l)\n"
"{\n"
"declare v2 varchar;\n"
"declare id int;\n"
"id := DB.DBA.RDF_MAKE_RO_ID_OF_STRING (v);\n"
"v2 := concat (dt_s, subseq (v, 0, l), \'\\0ABCD\', lang_s);\n"
"v2 [l+3] := bit_and (id, 255);\n"
"v2 [l+4] := bit_and (bit_shift (id, -8), 255);\n"
"v2 [l+5] := bit_and (bit_shift (id, -16), 255);\n"
"v2 [l+6] := bit_and (bit_shift (id, -24), 255);\n"
"return v2;\n"
"}\n"
"return concat (dt_s, v, \'\\0\', lang_s);\n"
"}\n"
"}\n"
"--src sparql.sql:888\n";

static const char *proc30 = 
"create function DB.DBA.RDF_OBJ_OF_SQLVAL (in v any) returns any\n"
"{\n"
"declare l, t int;\n"
"t := __tag (v);\n"
"if (not t in (182, 217, 225))\n"
"return v;\n"
"if (225 = t)\n"
"v := charset_recode (v, \'_WIDE_\', \'UTF-8\');\n"
"else if (217 = t)\n"
"v := cast (v as varchar);\n"
"l := 20;\n"
"if (length (v) > l)\n"
"{\n"
"declare v2 varchar;\n"
"declare id int;\n"
"id := DB.DBA.RDF_FIND_RO_ID_OF_STRING (v);\n"
"if (id is null)\n"
"return null;\n"
"v2 := concat (\'\\001\\001\', subseq (v, 0, l), \'\\0ABCD\\001\\001\');\n"
"v2 [l+3] := bit_and (id, 255);\n"
"v2 [l+4] := bit_and (bit_shift (id, -8), 255);\n"
"v2 [l+5] := bit_and (bit_shift (id, -16), 255);\n"
"v2 [l+6] := bit_and (bit_shift (id, -24), 255);\n"
"return v2;\n"
"}\n"
"return concat (\'\\001\\001\', v, \'\\0\\001\\001\');\n"
"}\n"
"--src sparql.sql:933\n";

static const char *proc31 = 
"create function DB.DBA.RDF_MAKE_LONG_OF_SQLVAL (in v any) returns any\n"
"{\n"
"declare l, t int;\n"
"t := __tag (v);\n"
"if (not t in (126, 182, 217, 225, 230))\n"
"return v;\n"
"if (225 = t)\n"
"v := charset_recode (v, \'_WIDE_\', \'UTF-8\');\n"
"else if (217 = t or 126 = t)\n"
"v := cast (v as varchar);\n"
"else if (230 = t)\n"
"v := serialize_to_UTF8_xml (v);\n"
"l := 20;\n"
"if (length (v) > l)\n"
"{\n"
"declare v2 varchar;\n"
"declare id int;\n"
"id := DB.DBA.RDF_MAKE_RO_ID_OF_STRING (v);\n"
"v2 := concat (\'\\001\\001\', subseq (v, 0, l), \'\\0ABCD\\001\\001\');\n"
"v2 [l+3] := bit_and (id, 255);\n"
"v2 [l+4] := bit_and (bit_shift (id, -8), 255);\n"
"v2 [l+5] := bit_and (bit_shift (id, -16), 255);\n"
"v2 [l+6] := bit_and (bit_shift (id, -24), 255);\n"
"return vector (257, v, 257, id, v2);\n"
"}\n"
"return vector (257, v, 257, 0, concat (\'\\001\\001\', v, \'\\0\\001\\001\'));\n"
"}\n"
"--src sparql.sql:965\n";

static const char *proc32 = 
"create function DB.DBA.RDF_MAKE_LONG_OF_TYPEDSQLVAL (in v any, in dt_iid IRI_ID, in lang varchar) returns any\n"
"{\n"
"declare l, t, dt_twobyte, lang_twobyte int;\n"
"declare dt_s, lang_s varchar;\n"
"t := __tag (v);\n"
"if (not t in (182, 217, 225, 230))\n"
"signal (\'RDFXX\', \'DB.DBA.RDF_MAKE_LONG_OF_TYPEDSQLVAL() accepts only string representations of typed values\');\n"
"if (225 = t)\n"
"v := charset_recode (v, \'_WIDE_\', \'UTF-8\');\n"
"else if (217 = t)\n"
"v := cast (v as varchar);\n"
"else if (230 = t)\n"
"v := serialize_to_UTF8_xml (v);\n"
"dt_s := \'\\001\\001\';\n"
"if (dt_iid is not null)\n"
"{\n"
"dt_twobyte := DB.DBA.RDF_TWOBYTE_OF_DATATYPE (dt_iid);\n"
"dt_s[0] := bit_and (dt_twobyte, 255);\n"
"dt_s[1] := bit_and (bit_shift (dt_twobyte, -8), 255);\n"
"}\n"
"else\n"
"dt_twobyte := 257;\n"
"lang_s := \'\\001\\001\';\n"
"if (lang is not null)\n"
"{\n"
"lang_twobyte := DB.DBA.RDF_TWOBYTE_OF_LANGUAGE (lang);\n"
"lang_s[0] := bit_and (lang_twobyte, 255);\n"
"lang_s[1] := bit_and (bit_shift (lang_twobyte, -8), 255);\n"
"}\n"
"else\n"
"lang_twobyte := 257;\n"
"l := 20;\n"
"if (length (v) > l)\n"
"{\n"
"declare v2 varchar;\n"
"declare id int;\n"
"id := DB.DBA.RDF_MAKE_RO_ID_OF_STRING (v);\n"
"v2 := concat (dt_s, subseq (v, 0, l), \'\\0ABCD\', lang_s);\n"
"v2 [l+3] := bit_and (id, 255);\n"
"v2 [l+4] := bit_and (bit_shift (id, -8), 255);\n"
"v2 [l+5] := bit_and (bit_shift (id, -16), 255);\n"
"v2 [l+6] := bit_and (bit_shift (id, -24), 255);\n"
"return vector (dt_twobyte, v, lang_twobyte, id, v2);\n"
"}\n"
"return vector (dt_twobyte, v, lang_twobyte, 0, concat (dt_s, v, \'\\0\', lang_s));\n"
"}\n"
"--src sparql.sql:995\n";

static const char *proc33 = 
"create function DB.DBA.RDF_MAKE_LONG_OF_TYPEDSQLVAL_STRINGS (\n"
"in o_val any, in o_type varchar, in o_lang varchar ) returns any\n"
"{\n"
"if (\'http://www.w3.org/2001/XMLSchema#boolean\' = o_type)\n"
"{\n"
"if ((\'true\' = o_val) or (\'1\' = o_val))\n"
"return 1;\n"
"else if ((\'false\' = o_val) or (\'0\' = o_val))\n"
"return 0;\n"
"else signal (\'RDFXX\', \'Invalid notation of boolean literal\');\n"
"}\n"
"if (\'http://www.w3.org/2001/XMLSchema#dateTime\' = o_type)\n"
"return __xqf_str_parse (\'dateTime\', o_val);\n"
"if (\'http://www.w3.org/2001/XMLSchema#double\' = o_type)\n"
"return cast (o_val as double precision);\n"
"if (\'http://www.w3.org/2001/XMLSchema#float\' = o_type)\n"
"return cast (o_val as float);\n"
"if (\'http://www.w3.org/2001/XMLSchema#integer\' = o_type)\n"
"return cast (o_val as int);\n"
"if (isstring(o_type) or isstring (o_lang))\n"
"{\n"
"return DB.DBA.RDF_MAKE_LONG_OF_TYPEDSQLVAL (\n"
"o_val,\n"
"DB.DBA.RDF_MAKE_IID_OF_QNAME (o_type),\n"
"o_lang );\n"
"}\n"
"return DB.DBA.RDF_MAKE_LONG_OF_SQLVAL (o_val);\n"
"}\n"
"--src sparql.sql:1043\n";

static const char *proc34 = 
"create function DB.DBA.RDF_SQLVAL_OF_LONG (in longobj any) returns any\n"
"{\n"
"\n"
"if (isiri_id (longobj))\n"
"{\n"
"if (longobj >= #i1000000000)\n"
"return sprintf (\'nodeID://%d\', iri_id_num (longobj));\n"
"else\n"
"{\n"
"declare res varchar;\n"
"res := id_to_iri (longobj);\n"
"\n"
"if (res is null)\n"
"signal (\'RDFXX\', \'Wrong iid in DB.DBA.RDF_SQLVAL_OF_LONG()\');\n"
"return res;\n"
"}\n"
"}\n"
"if (193 = __tag (longobj))\n"
"{\n"
"declare v2 varchar;\n"
"if (isstring (longobj[1]))\n"
"return longobj[1];\n"
"if (length (longobj) > 5)\n"
"return call (longobj[5] || \'_SQLVAL_OF_LONG\')(longobj);\n"
"v2 := (select case (isnull (RO_LONG)) when 0 then blob_to_string (RO_LONG) else RO_VAL end from DB.DBA.RDF_OBJ where RO_ID = longobj[3]);\n"
"if (v2 is null)\n"
"signal (\'RDFXX\', sprintf (\'Integrity violation in DB.DBA.RDF_SQLVAL_OF_LONG, bad id %d\', longobj[3]));\n"
"return v2;\n"
"}\n"
"return longobj;\n"
"}\n"
"--src sparql.sql:1073\n";

static const char *proc35 = 
"create function DB.DBA.RDF_BOOL_OF_LONG (in longobj any) returns any\n"
"{\n"
"if (longobj is null)\n"
"return null;\n"
"if (isiri_id (longobj))\n"
"return NULL;\n"
"if (isinteger (longobj))\n"
"{\n"
"if (longobj)\n"
"return 1;\n"
"return 0;\n"
"}\n"
"if (193 <> __tag (longobj))\n"
"return neq (longobj, 0.0);\n"
"declare dtqname any;\n"
"if (257 = longobj[0])\n"
"goto type_ok;\n"
"whenever not found goto badtype;\n"
"select RDT_QNAME into dtqname from DB.DBA.RDF_DATATYPE where RDT_TWOBYTE = longobj[0];\n"
"if (dtqname <> UNAME\'http://www.w3.org/2001/XMLSchema#string\')\n"
"return null;\n"
"\n"
"type_ok:\n"
"return case (length (longobj[1])) when 0 then 0 else 1 end;\n"
"\n"
"badtype:\n"
"signal (\'RDFXX\', sprintf (\'Unknown datatype in DB.DBA.RDF_BOOL_OF_LONG (code %d)\', longobj[0]));\n"
"}\n"
"--src sparql.sql:1106\n";

static const char *proc36 = 
"create function DB.DBA.RDF_DATATYPE_OF_LONG (in longobj any) returns any\n"
"{\n"
"if (193 = __tag (longobj))\n"
"{\n"
"declare twobyte integer;\n"
"declare res varchar;\n"
"twobyte := longobj[0];\n"
"if (257 = twobyte)\n"
"return null;\n"
"whenever not found goto badtype;\n"
"select RDT_QNAME into res from DB.DBA.RDF_DATATYPE where RDT_TWOBYTE = twobyte;\n"
"return res;\n"
"\n"
"badtype:\n"
"signal (\'RDFXX\', sprintf (\'Unknown datatype in DB.DBA.RDF_DATATYPE_OF_LONG, bad id %d\', twobyte));\n"
"}\n"
"if (isiri_id (longobj))\n"
"return NULL;\n"
"return DB.DBA.RDF_DATATYPE_OF_TAG (__tag (longobj));\n"
"}\n"
"--src sparql.sql:1136\n";

static const char *proc37 = 
"create function DB.DBA.RDF_LANGUAGE_OF_LONG (in longobj any) returns any\n"
"{\n"
"if (193 = __tag (longobj))\n"
"{\n"
"declare twobyte integer;\n"
"declare res varchar;\n"
"twobyte := longobj[2];\n"
"if (257 = twobyte)\n"
"return null;\n"
"whenever not found goto badlang;\n"
"select RL_ID into res from DB.DBA.RDF_LANGUAGE where RL_TWOBYTE = twobyte;\n"
"return res;\n"
"\n"
"badlang:\n"
"signal (\'RDFXX\', sprintf (\'Unknown language in DB.DBA.RDF_LANGUAGE_OF_LONG, bad id %d\', twobyte));\n"
"}\n"
"return NULL;\n"
"}\n"
"--src sparql.sql:1158\n";

static const char *proc38 = 
"create function DB.DBA.RDF_STRSQLVAL_OF_LONG (in longobj any)\n"
"{\n"
"declare t, l, len integer;\n"
"if (193 = __tag (longobj))\n"
"{\n"
"declare v2 varchar;\n"
"if (isstring (longobj[1]))\n"
"return longobj[1];\n"
"if (length (longobj) > 5)\n"
"return call (longobj[5] || \'_STRSQLVAL_OF_LONG\')(longobj);\n"
"v2 := (select case (isnull (RO_LONG)) when 0 then blob_to_string (RO_LONG) else RO_VAL end from DB.DBA.RDF_OBJ where RO_ID = longobj[3]);\n"
"if (v2 is null)\n"
"signal (\'RDFXX\', sprintf (\'Integrity violation in DB.DBA.RDF_STRSQLVAL_OF_LONG, bad id %d\', longobj[3]));\n"
"return v2;\n"
"}\n"
"if (isiri_id (longobj))\n"
"{\n"
"declare res varchar;\n"
"if (longobj >= #i1000000000)\n"
"return sprintf (\'nodeID://%d\', iri_id_num (longobj));\n"
"res := id_to_iri (longobj);\n"
"\n"
"if (res is null)\n"
"signal (\'RDFXX\', \'Wrong iid in DB.DBA.RDF_STRSQLVAL_OF_LONG()\');\n"
"return res;\n"
"}\n"
"if (211 = __tag (longobj))\n"
"{\n"
"declare vc varchar;\n"
"vc := cast (longobj as varchar);\n"
"return replace (vc, \' \', \'T\');\n"
"}\n"
"if (225 = __tag (longobj))\n"
"return charset_recode (longobj, \'_WIDE_\', \'UTF-8\');\n"
"return cast (longobj as varchar);\n"
"}\n"
"--src sparql.sql:1178\n";

static const char *proc39 = 
"create function DB.DBA.RDF_LONG_OF_SQLVAL (in v any) returns any\n"
"{\n"
"declare l, t int;\n"
"t := __tag (v);\n"
"if (not t in (182, 217, 225))\n"
"return v;\n"
"if (225 = t)\n"
"v := charset_recode (v, \'_WIDE_\', \'UTF-8\');\n"
"else if (217 = t)\n"
"v := cast (v as varchar);\n"
"l := 20;\n"
"if (length (v) > l)\n"
"{\n"
"declare v2 varchar;\n"
"declare id int;\n"
"id := DB.DBA.RDF_FIND_RO_ID_OF_STRING (v);\n"
"if (id is null)\n"
"return vector (257, v, 257, null, null);\n"
"v2 := concat (\'\\001\\001\', subseq (v, 0, l), \'\\0ABCD\\001\\001\');\n"
"v2 [l+3] := bit_and (id, 255);\n"
"v2 [l+4] := bit_and (bit_shift (id, -8), 255);\n"
"v2 [l+5] := bit_and (bit_shift (id, -16), 255);\n"
"v2 [l+6] := bit_and (bit_shift (id, -24), 255);\n"
"return vector (257, v, 257, id, v2);\n"
"}\n"
"return vector (257, v, 257, 0, concat (\'\\001\\001\', v, \'\\0\\001\\001\'));\n"
"}\n"
"--src sparql.sql:1216\n";

static const char *proc40 = 
" create function DB.DBA.RDF_STRSQLVAL_OF_SQLVAL (in sqlval any)\n"
"{\n"
"declare t, l, len integer;\n"
"if (193 = __tag (sqlval))\n"
"{\n"
"signal (\'RDFXX\', \'Long object in DB.DBA.RDF_STRSQLVAL_OF_SQLVAL()\');\n"
"}\n"
"if (isiri_id (sqlval))\n"
"{\n"
"declare res varchar;\n"
"if (sqlval >= #i1000000000)\n"
"return sprintf (\'nodeID://%d\', iri_id_num (sqlval));\n"
"res := id_to_iri (sqlval);\n"
"\n"
"if (res is null)\n"
"signal (\'RDFXX\', \'Wrong iid in DB.DBA.RDF_STRSQLVAL_OF_SQLVAL()\');\n"
"return res;\n"
"}\n"
"if (211 = __tag (sqlval))\n"
"{\n"
"declare vc varchar;\n"
"vc := cast (sqlval as varchar);\n"
"return replace (vc, \' \', \'T\');\n"
"}\n"
"if (225 = __tag (sqlval))\n"
"return charset_recode (sqlval, \'_WIDE_\', \'UTF-8\');\n"
"return cast (sqlval as varchar);\n"
"}\n"
"--src sparql.sql:1249\n";

static const char *proc41 = 
" create function DB.DBA.RDF_DATATYPE_OF_SQLVAL (in v any) returns any\n"
"{\n"
"declare t int;\n"
"t := __tag (v);\n"
"if (not t in (182, 217, 225))\n"
"return DB.DBA.RDF_DATATYPE_OF_TAG (t);\n"
"return UNAME\'http://www.w3.org/2001/XMLSchema#string\';\n"
"}\n"
"--src sparql.sql:1280\n";

static const char *proc42 = 
" create function DB.DBA.RDF_LANGUAGE_OF_SQLVAL (in v any) returns any\n"
"{\n"
"declare t int;\n"
"return NULL;\n"
"\n"
"\n"
"\n"
"\n"
"}\n"
"--src sparql.sql:1291\n";

static const char *proc43 = 
" create function DB.DBA.RDF_IS_BLANK_REF (in v any) returns any\n"
"{\n"
"if (not isstring (v))\n"
"return 0;\n"
"if (\"LEFT\" (v, 9) <> \'nodeID://\')\n"
"return 0;\n"
"return 1;\n"
"}\n"
"--src sparql.sql:1303\n";

static const char *proc44 = 
" create function DB.DBA.RDF_IS_URI_REF (in v any) returns any\n"
"{\n"
"if (not isstring (v))\n"
"return 0;\n"
"if (\"LEFT\" (v, 9) = \'nodeID://\')\n"
"return 0;\n"
"return 1;\n"
"}\n"
"--src sparql.sql:1314\n";

static const char *proc45 = 
" create function DB.DBA.\"http://www.w3.org/2001/XMLSchema#boolean\" (in strg any) returns integer\n"
"{\n"
"if (isstring (strg))\n"
"{\n"
"if ((\'true\' = strg) or (\'1\' = strg))\n"
"return 1;\n"
"if ((\'false\' = strg) or (\'0\' = strg))\n"
"return 0;\n"
"}\n"
"if (isinteger (strg))\n"
"return case (strg) when 0 then 0 else 1 end;\n"
"return NULL;\n"
"}\n"
"--src sparql.sql:1328\n";

static const char *proc46 = 
" create function DB.DBA.\"http://www.w3.org/2001/XMLSchema#dateTime\" (in strg any) returns datetime\n"
"{\n"
"if (211 = __tag (strg))\n"
"return strg;\n"
"if (isstring (strg))\n"
"return __xqf_str_parse (\'dateTime\', strg);\n"
"}\n"
"--src sparql.sql:1344\n";

static const char *proc47 = 
" create function DB.DBA.\"http://www.w3.org/2001/XMLSchema#double\" (in strg varchar) returns double precision\n"
"{\n"
"return cast (strg as double precision);\n"
"}\n"
"--src sparql.sql:1354\n";

static const char *proc48 = 
" create function DB.DBA.\"http://www.w3.org/2001/XMLSchema#float\" (in strg varchar) returns float\n"
"{\n"
"return cast (strg as float);\n"
"}\n"
"--src sparql.sql:1361\n";

static const char *proc49 = 
" create function DB.DBA.\"http://www.w3.org/2001/XMLSchema#integer\" (in strg varchar) returns integer\n"
"{\n"
"return cast (strg as integer);\n"
"}\n"
"--src sparql.sql:1368\n";

static const char *proc50 = 
" create function DB.DBA.__and (in e1 any, in e2 any) returns integer\n"
"{\n"
"if (e1 and e2)\n"
"return 1;\n"
"return 0;\n"
"}\n"
"--src sparql.sql:1379\n";

static const char *proc51 = 
" create function DB.DBA.__or (in e1 any, in e2 any) returns integer\n"
"{\n"
"if (e1 or e2)\n"
"return 1;\n"
"return 0;\n"
"}\n"
"--src sparql.sql:1388\n";

static const char *proc52 = 
" create function DB.DBA.__not (in e1 any) returns integer\n"
"{\n"
"if (e1)\n"
"return 0;\n"
"return 1;\n"
"}\n"
"--src sparql.sql:1397\n";

static const char *proc53 = 
"create procedure DB.DBA.RDF_QUAD_URI (in g_uri varchar, in s_uri varchar, in p_uri varchar, in o_uri varchar)\n"
"{\n"
"insert replacing DB.DBA.RDF_QUAD (G,S,P,O)\n"
"values (\n"
"DB.DBA.RDF_MAKE_IID_OF_QNAME (g_uri),\n"
"DB.DBA.RDF_MAKE_IID_OF_QNAME (s_uri),\n"
"DB.DBA.RDF_MAKE_IID_OF_QNAME (p_uri),\n"
"DB.DBA.RDF_MAKE_IID_OF_QNAME (o_uri) );\n"
"}\n"
"--src sparql.sql:1409\n";

static const char *proc54 = 
"create procedure DB.DBA.RDF_QUAD_URI_L (in g_uri varchar, in s_uri varchar, in p_uri varchar, in o_lit any)\n"
"{\n"
"insert replacing DB.DBA.RDF_QUAD (G,S,P,O)\n"
"values (\n"
"DB.DBA.RDF_MAKE_IID_OF_QNAME (g_uri),\n"
"DB.DBA.RDF_MAKE_IID_OF_QNAME (s_uri),\n"
"DB.DBA.RDF_MAKE_IID_OF_QNAME (p_uri),\n"
"DB.DBA.RDF_MAKE_OBJ_OF_SQLVAL (o_lit) );\n"
"}\n"
"--src sparql.sql:1420\n";

static const char *proc55 = 
"create procedure DB.DBA.RDF_QUAD_URI_L_TYPED (in g_uri varchar, in s_uri varchar, in p_uri varchar, in o_lit any, in dt any, in lang varchar)\n"
"{\n"
"insert replacing DB.DBA.RDF_QUAD (G,S,P,O)\n"
"values (\n"
"DB.DBA.RDF_MAKE_IID_OF_QNAME (g_uri),\n"
"DB.DBA.RDF_MAKE_IID_OF_QNAME (s_uri),\n"
"DB.DBA.RDF_MAKE_IID_OF_QNAME (p_uri),\n"
"DB.DBA.RDF_MAKE_OBJ_OF_TYPEDSQLVAL (\n"
"o_lit, DB.DBA.RDF_MAKE_IID_OF_QNAME (dt), lang ) );\n"
"}\n"
"--src sparql.sql:1431\n";

static const char *proc56 = 
"create procedure DB.DBA.TTLP_EXEC_NEW_GRAPH (in g varchar, inout app_env any) {\n"
"\n"
" ;\n"
"}\n"
"--src sparql.sql:1443\n";

static const char *proc57 = 
"create function DB.DBA.TTLP_EXEC_NEW_BLANK (in g varchar, inout app_env any) returns IRI_ID {\n"
"declare res IRI_ID;\n"
"res := iri_id_from_num (sequence_next (\'RDF_URL_IID_BLANK\'));\n"
"\n"
"return res;\n"
"}\n"
"--src sparql.sql:1449\n";

static const char *proc58 = 
"create function DB.DBA.TTLP_EXEC_GET_IID (in uri varchar, in g varchar, inout app_env any) returns IRI_ID {\n"
"declare res IRI_ID;\n"
"\n"
"res := DB.DBA.RDF_MAKE_IID_OF_QNAME (uri);\n"
"\n"
"return res;\n"
"}\n"
"--src sparql.sql:1457\n";

static const char *proc59 = 
"create procedure DB.DBA.TTLP_EXEC_TRIPLE (\n"
"in g_uri varchar, in g_iid IRI_ID,\n"
"in s_uri varchar, in s_iid IRI_ID,\n"
"in p_uri varchar, in p_iid IRI_ID,\n"
"in o_uri varchar, in o_iid IRI_ID,\n"
"inout app_env any )\n"
"{\n"
"insert replacing DB.DBA.RDF_QUAD (G,S,P,O)\n"
"values (g_iid, s_iid, p_iid, o_iid);\n"
"}\n"
"--src sparql.sql:1466\n";

static const char *proc60 = 
"create procedure DB.DBA.TTLP_EXEC_TRIPLE_L (\n"
"in g_uri varchar, in g_iid IRI_ID,\n"
"in s_uri varchar, in s_iid IRI_ID,\n"
"in p_uri varchar, in p_iid IRI_ID,\n"
"in o_val any, in o_type varchar, in o_lang varchar,\n"
"inout app_env any )\n"
"{\n"
"if (\'http://www.w3.org/2001/XMLSchema#boolean\' = o_type)\n"
"{\n"
"if ((\'true\' = o_val) or (\'1\' = o_val))\n"
"o_val := 1;\n"
"else if ((\'false\' = o_val) or (\'0\' = o_val))\n"
"o_val := 0;\n"
"else signal (\'RDFXX\', \'Invalid notation of boolean literal\');\n"
"}\n"
"else if (\'http://www.w3.org/2001/XMLSchema#dateTime\' = o_type)\n"
"{\n"
"o_val := __xqf_str_parse (\'dateTime\', o_val);\n"
"}\n"
"else if (\'http://www.w3.org/2001/XMLSchema#double\' = o_type)\n"
"{\n"
"o_val := cast (o_val as double precision);\n"
"}\n"
"else if (\'http://www.w3.org/2001/XMLSchema#float\' = o_type)\n"
"{\n"
"o_val := cast (o_val as float);\n"
"}\n"
"else if (\'http://www.w3.org/2001/XMLSchema#integer\' = o_type)\n"
"{\n"
"o_val := cast (o_val as int);\n"
"}\n"
"else if (isstring (o_type) or isstring (o_lang))\n"
"{\n"
"if (not isstring (o_type))\n"
"o_type := null;\n"
"if (not isstring (o_lang))\n"
"o_lang := null;\n"
"insert replacing DB.DBA.RDF_QUAD (G,S,P,O)\n"
"values (g_iid, s_iid, p_iid,\n"
"DB.DBA.RDF_MAKE_OBJ_OF_TYPEDSQLVAL (\n"
"o_val,\n"
"DB.DBA.RDF_MAKE_IID_OF_QNAME (o_type),\n"
"o_lang ) );\n"
"return;\n"
"}\n"
"insert replacing DB.DBA.RDF_QUAD (G,S,P,O)\n"
"values (g_iid, s_iid, p_iid, DB.DBA.RDF_MAKE_OBJ_OF_SQLVAL (o_val));\n"
"}\n"
"--src sparql.sql:1478\n";

static const char *proc61 = 
"create procedure DB.DBA.TTLP (in strg varchar, in base varchar, in graph varchar := null, in flags integer := 0)\n"
"{\n"
"return rdf_load_turtle (blob_to_string (strg), base, graph, flags,\n"
"vector (\n"
"\'DB.DBA.TTLP_EXEC_NEW_GRAPH(?,?)\',\n"
"\'select DB.DBA.TTLP_EXEC_NEW_BLANK(?,?)\',\n"
"\'select DB.DBA.TTLP_EXEC_GET_IID(?,?,?)\',\n"
"\'DB.DBA.TTLP_EXEC_TRIPLE(?,?, ?,?, ?,?, ?,?, ?)\',\n"
"\'DB.DBA.TTLP_EXEC_TRIPLE_L(?,?, ?,?, ?,?, ?,?,?, ?)\',\n"
"\'commit work\' ),\n"
"\'app-env\');\n"
"}\n"
"--src sparql.sql:1528\n";

static const char *proc62 = 
"create procedure DB.DBA.RDF_TTL2HASH_EXEC_NEW_GRAPH (in g varchar, inout app_env any) {\n"
"\n"
" ;\n"
"}\n"
"--src sparql.sql:1542\n";

static const char *proc63 = 
"create function DB.DBA.RDF_TTL2HASH_EXEC_NEW_BLANK (in g varchar, inout app_env any) returns IRI_ID {\n"
"declare res IRI_ID;\n"
"res := iri_id_from_num (sequence_next (\'RDF_URL_IID_BLANK\'));\n"
"\n"
"return res;\n"
"}\n"
"--src sparql.sql:1548\n";

static const char *proc64 = 
"create function DB.DBA.RDF_TTL2HASH_EXEC_GET_IID (in uri varchar, in g varchar, inout app_env any) returns IRI_ID {\n"
"declare res IRI_ID;\n"
"\n"
"res := DB.DBA.RDF_MAKE_IID_OF_QNAME (uri);\n"
"\n"
"return res;\n"
"}\n"
"--src sparql.sql:1556\n";

static const char *proc65 = 
"create procedure DB.DBA.RDF_TTL2HASH_EXEC_TRIPLE (\n"
"in g_uri varchar, in g_iid IRI_ID,\n"
"in s_uri varchar, in s_iid IRI_ID,\n"
"in p_uri varchar, in p_iid IRI_ID,\n"
"in o_uri varchar, in o_iid IRI_ID,\n"
"inout app_env any )\n"
"{\n"
"dict_put (app_env, vector (s_iid, p_iid, o_iid), 0);\n"
"}\n"
"--src sparql.sql:1565\n";

static const char *proc66 = 
"create procedure DB.DBA.RDF_TTL2HASH_EXEC_TRIPLE_L (\n"
"in g_uri varchar, in g_iid IRI_ID,\n"
"in s_uri varchar, in s_iid IRI_ID,\n"
"in p_uri varchar, in p_iid IRI_ID,\n"
"in o_val any, in o_type varchar, in o_lang varchar,\n"
"inout app_env any )\n"
"{\n"
"if (not isstring (o_type))\n"
"o_type := null;\n"
"if (not isstring (o_lang))\n"
"o_lang := null;\n"
"dict_put (app_env,\n"
"vector (s_iid, p_iid, DB.DBA.RDF_MAKE_LONG_OF_TYPEDSQLVAL_STRINGS (o_val, o_type, o_lang)),\n"
"0 );\n"
"}\n"
"--src sparql.sql:1576\n";

static const char *proc67 = 
"create function DB.DBA.RDF_TTL2HASH (in str varchar, in base varchar, in graph varchar := null, in flags integer := 0) returns any\n"
"{\n"
"declare res any;\n"
"res := dict_new ();\n"
"rdf_load_turtle (str, base, graph, flags,\n"
"vector (\n"
"\'DB.DBA.RDF_TTL2HASH_EXEC_NEW_GRAPH(?,?)\',\n"
"\'select DB.DBA.RDF_TTL2HASH_EXEC_NEW_BLANK(?,?)\',\n"
"\'select DB.DBA.RDF_TTL2HASH_EXEC_GET_IID(?,?,?)\',\n"
"\'DB.DBA.RDF_TTL2HASH_EXEC_TRIPLE(?,?, ?,?, ?,?, ?,?, ?)\',\n"
"\'DB.DBA.RDF_TTL2HASH_EXEC_TRIPLE_L(?,?, ?,?, ?,?, ?,?,?, ?)\',\n"
"\'isinteger(0)\' ),\n"
"res);\n"
"return res;\n"
"}\n"
"--src sparql.sql:1593\n";

static const char *proc68 = 
"create procedure DB.DBA.RDF_LOAD_RDFXML (in strg varchar, in base varchar, in graph varchar)\n"
"{\n"
"rdf_load_rdfxml (strg, 0,\n"
"graph,\n"
"vector (\n"
"\'DB.DBA.TTLP_EXEC_NEW_GRAPH(?,?)\',\n"
"\'select DB.DBA.TTLP_EXEC_NEW_BLANK(?,?)\',\n"
"\'select DB.DBA.TTLP_EXEC_GET_IID(?,?,?)\',\n"
"\'DB.DBA.TTLP_EXEC_TRIPLE(?,?, ?,?, ?,?, ?,?, ?)\',\n"
"\'DB.DBA.TTLP_EXEC_TRIPLE_L(?,?, ?,?, ?,?, ?,?,?, ?)\',\n"
"\'commit work\' ),\n"
"\'app-env\',\n"
"base );\n"
"return graph;\n"
"}\n"
"--src sparql.sql:1610\n";

static const char *proc69 = 
"create procedure DB.DBA.RDF_RDFXML_TO_DICT (in strg varchar, in base varchar, in graph varchar)\n"
"{\n"
"declare res any;\n"
"res := dict_new ();\n"
"rdf_load_rdfxml (strg, 0,\n"
"graph,\n"
"vector (\n"
"\'DB.DBA.TTL2HASH_EXEC_NEW_GRAPH(?,?)\',\n"
"\'select DB.DBA.RDF_TTL2HASH_EXEC_NEW_BLANK(?,?)\',\n"
"\'select DB.DBA.RDF_TTL2HASH_EXEC_GET_IID(?,?,?)\',\n"
"\'DB.DBA.RDF_TTL2HASH_EXEC_TRIPLE(?,?, ?,?, ?,?, ?,?, ?)\',\n"
"\'DB.DBA.RDF_TTL2HASH_EXEC_TRIPLE_L(?,?, ?,?, ?,?, ?,?,?, ?)\',\n"
"\'isinteger(0)\' ),\n"
"res,\n"
"base );\n"
"return res;\n"
"}\n"
"--src sparql.sql:1627\n";

static const char *proc70 = 
"create procedure DB.DBA.RDF_LONG_TO_TTL (inout obj any, inout ses any)\n"
"{\n"
"declare res varchar;\n"
"if (obj is null)\n"
"signal (\'RDFXX\', \'DB.DBA.TRIPLES_TO_TTL(): object is NULL\');\n"
"if (isiri_id (obj))\n"
"{\n"
"if (obj >= #i1000000000)\n"
"http (sprintf (\'_:b%d \', iri_id_num (obj)), ses);\n"
"else\n"
"{\n"
"res := coalesce (id_to_iri (obj), sprintf (\'_:bad_iid_%d\', iri_id_num (obj)));\n"
"\n"
"http (\'<\', ses);\n"
"http_escape (res, 12, ses, 1, 1);\n"
"http (\'> \', ses);\n"
"}\n"
"}\n"
"else if (193 = __tag (obj))\n"
"{\n"
"http (\'\"\', ses);\n"
"http_escape (obj[1], 11, ses, 1, 1);\n"
"if (257 <> obj[0])\n"
"{\n"
"res := coalesce ((select RDT_QNAME from DB.DBA.RDF_DATATYPE where RDT_TWOBYTE = obj[0]));\n"
"http (\'\"^^<\', ses);\n"
"http_escape (res, 12, ses, 1, 1);\n"
"http (\'> \', ses);\n"
"}\n"
"else if (257 <> obj[2])\n"
"{\n"
"res := coalesce ((select RL_ID from DB.DBA.RDF_LANGUAGE where RL_TWOBYTE = obj[2]));\n"
"http (\'\"@\', ses); http (res, ses); http (\' \', ses);\n"
"}\n"
"else\n"
"http (\'\"\', ses);\n"
"}\n"
"else if (182 = __tag (obj))\n"
"{\n"
"http (\'\"\', ses);\n"
"http_escape (obj, 11, ses, 1, 1);\n"
"http (\'\" \', ses);\n"
"}\n"
"else\n"
"{\n"
"http (\'\"\', ses);\n"
"http_escape (DB.DBA.RDF_STRSQLVAL_OF_LONG (obj), 11, ses, 1, 1);\n"
"http (\'\"^^<\', ses);\n"
"http_escape (cast (DB.DBA.RDF_DATATYPE_OF_TAG (__tag (obj)) as varchar), 12, ses, 1, 1);\n"
"http (\'> \', ses);\n"
"}\n"
"}\n"
"--src sparql.sql:1650\n";

static const char *proc71 = 
"create procedure DB.DBA.RDF_TRIPLES_TO_TTL (inout triples any, inout ses any)\n"
"{\n"
"declare tcount, tctr integer;\n"
"declare res varchar;\n"
"tcount := length (triples);\n"
"\n"
"if (0 = tcount)\n"
"{\n"
"http (\'# Empty TURTLE\', ses);\n"
"return;\n"
"}\n"
"for (tctr := 0; tctr < tcount; tctr := tctr + 1)\n"
"{\n"
"declare subj,pred,obj any;\n"
"subj := triples[tctr][0];\n"
"pred := triples[tctr][1];\n"
"obj := triples[tctr][2];\n"
"if (not isiri_id (subj))\n"
"{\n"
"if (subj is null)\n"
"signal (\'RDFXX\', \'DB.DBA.TRIPLES_TO_TTL(): subject is NULL\');\n"
"signal (\'RDFXX\', \'DB.DBA.TRIPLES_TO_TTL(): subject is literal\');\n"
"}\n"
"if (subj >= #i1000000000)\n"
"http (sprintf (\'_:b%d \', iri_id_num (subj)), ses);\n"
"else\n"
"{\n"
"res := id_to_iri (subj);\n"
"\n"
"http (\'<\', ses);\n"
"http_escape (res, 12, ses, 1, 1);\n"
"http (\'> \', ses);\n"
"}\n"
"if (not isiri_id (pred))\n"
"{\n"
"if (pred is null)\n"
"signal (\'RDFXX\', \'DB.DBA.TRIPLES_TO_TTL(): predicate is NULL\');\n"
"signal (\'RDFXX\', \'DB.DBA.TRIPLES_TO_TTL(): predicate is literal\');\n"
"}\n"
"if (pred >= #i1000000000)\n"
"signal (\'RDFXX\', \'DB.DBA.TRIPLES_TO_TTL(): blank node as predicate\');\n"
"else\n"
"{\n"
"res := id_to_iri (pred);\n"
"\n"
"http (\'<\', ses);\n"
"http_escape (res, 12, ses, 1, 1);\n"
"http (\'> \', ses);\n"
"}\n"
"DB.DBA.RDF_LONG_TO_TTL (obj, ses);\n"
"http (\'.\\n\', ses);\n"
"}\n"
"}\n"
"--src sparql.sql:1705\n";

static const char *proc72 = 
"create procedure DB.DBA.RDF_TRIPLES_TO_RDF_XML_TEXT (inout triples any, in print_top_level integer, inout ses any)\n"
"{\n"
"declare tcount, tctr integer;\n"
"tcount := length (triples);\n"
"if (print_top_level)\n"
"{\n"
"http (\'<rdf:RDF xmlns:rdf=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\">\', ses);\n"
"}\n"
"for (tctr := 0; tctr < tcount; tctr := tctr + 1)\n"
"{\n"
"declare subj, pred, obj any;\n"
"declare pred_tagname varchar;\n"
"declare res varchar;\n"
"subj := triples[tctr][0];\n"
"pred := triples[tctr][1];\n"
"obj := triples[tctr][2];\n"
"http (\'\\n<rdf:Description\', ses);\n"
"if (not isiri_id (subj))\n"
"{\n"
"if (subj is null)\n"
"signal (\'RDFXX\', \'DB.DBA.TRIPLES_TO_RDF_XML_TEXT(): subject is NULL\');\n"
"signal (\'RDFXX\', \'DB.DBA.TRIPLES_TO_RDF_XML_TEXT(): subject is literal\');\n"
"}\n"
"if (subj >= #i1000000000)\n"
"http (sprintf (\' rdf:nodeID=\"b%d\">\', iri_id_num (subj)), ses);\n"
"else\n"
"{\n"
"res := id_to_iri (subj);\n"
"\n"
"http (\' rdf:about=\"\', ses); http_value (res, 0, ses); http (\'\">\', ses);\n"
"}\n"
"if (not isiri_id (pred))\n"
"{\n"
"if (pred is null)\n"
"signal (\'RDFXX\', \'DB.DBA.TRIPLES_TO_RDF_XML_TEXT(): predicate is NULL\');\n"
"signal (\'RDFXX\', \'DB.DBA.TRIPLES_TO_RDF_XML_TEXT(): predicate is literal\');\n"
"}\n"
"if (pred >= #i1000000000)\n"
"signal (\'RDFXX\', \'DB.DBA.TRIPLES_TO_RDF_XML_TEXT(): blank node as predicate\');\n"
"else\n"
"{\n"
"res := id_to_iri (pred);\n"
"\n"
"declare delim integer;\n"
"delim := __max (strrchr (res, \'/\'), strrchr (res, \'#\'), strrchr (res, \':\'));\n"
"if (delim is null)\n"
"{\n"
"pred_tagname := res;\n"
"http (\'<\', ses); http (pred_tagname, ses);\n"
"}\n"
"else\n"
"{\n"
"pred_tagname := \'ns0pred:\' || subseq (res, delim+1);\n"
"http (\'<\', ses); http (pred_tagname, ses);\n"
"http (\' xmlns:ns0pred=\"\', ses); http_value (subseq (res, 0, delim+1), 0, ses);\n"
"http (\'\"\', ses);\n"
"}\n"
"}\n"
"if (obj is null)\n"
"signal (\'RDFXX\', \'DB.DBA.TRIPLES_TO_RDF_XML_TEXT(): object is NULL\');\n"
"if (isiri_id (obj))\n"
"{\n"
"if (obj >= #i1000000000)\n"
"http (sprintf (\' rdf:nodeID=\"b%d\"/>\', iri_id_num (subj)), ses);\n"
"else\n"
"{\n"
"res := coalesce (id_to_iri(obj), sprintf (\'_:bad_iid_%d\', iri_id_num (obj)));\n"
"\n"
"http (\' rdf:resource=\"\', ses); http_value (res, 0, ses); http (\'\"/>\', ses);\n"
"}\n"
"}\n"
"else if (193 = __tag (obj))\n"
"{\n"
"if (257 <> obj[0])\n"
"{\n"
"res := coalesce ((select RDT_QNAME from DB.DBA.RDF_DATATYPE where RDT_TWOBYTE = obj[0]));\n"
"http (\' rdf:type=\"\', ses); http_value (res, 0, ses); http (\'\"\', ses);\n"
"}\n"
"else if (257 <> obj[2])\n"
"{\n"
"res := coalesce ((select RDT_QNAME from DB.DBA.RDF_DATATYPE where RDT_TWOBYTE = obj[0]));\n"
"http (\' xml:lang=\"\', ses); http_value (res, 0, ses); http (\'\"\', ses);\n"
"}\n"
"if (230 = __tag (obj[1]))\n"
"{\n"
"http (\' rdf:parseType=\"Literal\">\', ses);\n"
"http_value (obj[1], 0, ses);\n"
"http (\'</\', ses); http (pred_tagname, ses); http (\'>\', ses);\n"
"}\n"
"else\n"
"{\n"
"http (\'>\', ses);\n"
"http_value (DB.DBA.RDF_STRSQLVAL_OF_LONG (obj), 0, ses);\n"
"http (\'</\', ses); http (pred_tagname, ses); http (\'>\', ses);\n"
"}\n"
"}\n"
"else if (182 = __tag (obj))\n"
"{\n"
"http (\'>\', ses);\n"
"http_value (obj, 0, ses);\n"
"http (\'</\', ses); http (pred_tagname, ses); http (\'>\', ses);\n"
"}\n"
"else\n"
"{\n"
"http (\' rdf:type=\"\', ses);\n"
"http_value (DB.DBA.RDF_DATATYPE_OF_TAG (__tag (obj)), 0, ses);\n"
"http (\'\">\', ses);\n"
"http_value (DB.DBA.RDF_STRSQLVAL_OF_LONG (obj), 0, ses);\n"
"http (\'</\', ses); http (pred_tagname, ses); http (\'>\', ses);\n"
"}\n"
"http (\'</rdf:Description>\', ses);\n"
"}\n"
"if (print_top_level)\n"
"{\n"
"http (\'\\n</rdf:RDF>\', ses);\n"
"}\n"
"}\n"
"--src sparql.sql:1773\n";

static const char *proc73 = 
"create procedure DB.DBA.RDF_FORMAT_RESULT_SET_AS_TTL_INIT (inout _env any)\n"
"{\n"
"_env := string_output();\n"
"http (\'@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .\n"
"@prefix rs: <http://www.w3.org/2005/sparql-results#> .\n"
"@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .\n"
"[ rdf:type rs:results ;\', _env);\n"
"}\n"
"--src sparql.sql:1908\n";

static const char *proc74 = 
"create procedure DB.DBA.RDF_FORMAT_RESULT_SET_AS_TTL_ACC (inout _env any, inout colvalues any, inout colnames any)\n"
"{\n"
"declare col_ctr, col_count integer;\n"
"declare blank_ids any;\n"
"if (185 <> __tag(_env))\n"
"DB.DBA.RDF_FORMAT_RESULT_SET_AS_TTL_INIT (_env);\n"
"http (\'\\n  rs:result [\', _env);\n"
"col_count := length (colnames);\n"
"for (col_ctr := 0; col_ctr < col_count; col_ctr := col_ctr + 1)\n"
"{\n"
"declare _name varchar;\n"
"declare _val any;\n"
"_name := colnames[col_ctr];\n"
"_val := colvalues[col_ctr];\n"
"if (_val is null)\n"
"goto end_of_binding;\n"
"http (\'\\n      rs:binding [ rs:name \"\', _env);\n"
"http_value (colnames[col_ctr], 0, _env);\n"
"http (\'\" ; rs:value \', _env);\n"
"if (isiri_id (_val))\n"
"{\n"
"if (_val >= #i1000000000)\n"
"{\n"
"http (sprintf (\'_:nodeID%d ] ;\', iri_id_num (_val)), _env);\n"
"}\n"
"else\n"
"{\n"
"declare res varchar;\n"
"res := id_to_iri (_val);\n"
"\n"
"if (res is null)\n"
"res := sprintf (\'<bad://%d>\', iri_id_num (_val));\n"
"http (sprintf (\'<%V> ] ;\', res), _env);\n"
"}\n"
"}\n"
"else\n"
"{\n"
"declare lang, dt varchar;\n"
"lang := DB.DBA.RDF_LANGUAGE_OF_LONG (_val);\n"
"dt := DB.DBA.RDF_DATATYPE_OF_LONG (_val);\n"
"http (\'\"\', _env);\n"
"http_escape (DB.DBA.RDF_SQLVAL_OF_LONG (_val), 11, _env, 1, 1);\n"
"if (lang is not null)\n"
"{\n"
"if (dt is not null)\n"
"http (sprintf (\'\"@\"%V\"^^<%V> ] ;\',\n"
"cast (lang as varchar), cast (dt as varchar)), _env);\n"
"else\n"
"http (sprintf (\'\"@\"%V\" ] ;\',\n"
"cast (lang as varchar)), _env);\n"
"}\n"
"else\n"
"{\n"
"if (dt is not null)\n"
"http (sprintf (\'\"^^<%V> ] ;\',\n"
"cast (dt as varchar)), _env);\n"
"else\n"
"http (sprintf (\'\" ] ;\'), _env);\n"
"}\n"
"}\n"
"end_of_binding: ;\n"
"\n"
"}\n"
"http (\'\\n      ] ;\', _env);\n"
"}\n"
"--src sparql.sql:1918\n";

static const char *proc75 = 
"create function DB.DBA.RDF_FORMAT_RESULT_SET_AS_TTL_FIN (inout _env any) returns long varchar\n"
"{\n"
"if (185 <> __tag(_env))\n"
"DB.DBA.RDF_FORMAT_RESULT_SET_AS_TTL_INIT (_env);\n"
"\n"
"http (\'\\n    ] .\', _env);\n"
"return string_output_string (_env);\n"
"}\n"
"--src sparql.sql:1985\n";

static const char *other7 = 
"create aggregate DB.DBA.RDF_FORMAT_RESULT_SET_AS_TTL (in colvalues any, in colnames any) returns long varchar\n"
"from DB.DBA.RDF_FORMAT_RESULT_SET_AS_TTL_INIT, DB.DBA.RDF_FORMAT_RESULT_SET_AS_TTL_ACC, DB.DBA.RDF_FORMAT_RESULT_SET_AS_TTL_FIN\n";

static const char *proc76 = 
"create procedure DB.DBA.RDF_FORMAT_RESULT_SET_AS_RDF_XML_INIT (inout _env any)\n"
"{\n"
"_env := string_output();\n"
"http (\'<rdf:RDF\n"
"xmlns:rdf=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\"\n"
"xmlns:rs=\"http://www.w3.org/2005/sparql-results#\"\n"
"xmlns:xsd=\"http://www.w3.org/2001/XMLSchema#\" >\n"
"<rs:results rdf:nodeID=\"rset\">\', _env);\n"
"}\n"
"--src sparql.sql:1999\n";

static const char *proc77 = 
"create procedure DB.DBA.RDF_FORMAT_RESULT_SET_AS_RDF_XML_ACC (inout _env any, inout colvalues any, inout colnames any)\n"
"{\n"
"declare sol_id varchar;\n"
"declare col_ctr, col_count integer;\n"
"declare blank_ids any;\n"
"if (185 <> __tag(_env))\n"
"DB.DBA.RDF_FORMAT_RESULT_SET_AS_RDF_XML_INIT (_env);\n"
"sol_id := cast (length (_env) as varchar);\n"
"http (\'\\n  <rs:result rdf:nodeID=\"sol\' || sol_id || \'\">\', _env);\n"
"col_count := length (colnames);\n"
"for (col_ctr := 0; col_ctr < col_count; col_ctr := col_ctr + 1)\n"
"{\n"
"declare _name varchar;\n"
"declare _val any;\n"
"_name := colnames[col_ctr];\n"
"_val := colvalues[col_ctr];\n"
"if (_val is null)\n"
"goto end_of_binding;\n"
"http (\'\\n   <rs:binding rdf:nodeID=\"sol\' || sol_id || \'-\' || cast (col_ctr as varchar) || \'\" rs:name=\"\', _env);\n"
"http_value (colnames[col_ctr], 0, _env);\n"
"http (\'\"><rs:value\', _env);\n"
"if (isiri_id (_val))\n"
"{\n"
"if (_val >= #i1000000000)\n"
"{\n"
"http (sprintf (\' rdf:nodeID=\"%d\"/></rs:binding>\', iri_id_num (_val)), _env);\n"
"}\n"
"else\n"
"{\n"
"declare res varchar;\n"
"res := id_to_iri (_val);\n"
"\n"
"if (res is null)\n"
"res := sprintf (\'bad://%d\', iri_id_num (_val));\n"
"http (sprintf (\' rdf:resource=\"%V\"/></rs:binding>\', res), _env);\n"
"}\n"
"}\n"
"else\n"
"{\n"
"declare lang, dt varchar;\n"
"lang := DB.DBA.RDF_LANGUAGE_OF_LONG (_val);\n"
"dt := DB.DBA.RDF_DATATYPE_OF_LONG (_val);\n"
"if (lang is not null)\n"
"{\n"
"if (dt is not null)\n"
"http (sprintf (\' xml:lang=\"%V\" rdf:datatype=\"%V\">\',\n"
"cast (lang as varchar), cast (dt as varchar)), _env);\n"
"else\n"
"http (sprintf (\' xml:lang=\"%V\">\',\n"
"cast (lang as varchar)), _env);\n"
"}\n"
"else\n"
"{\n"
"if (dt is not null)\n"
"http (sprintf (\' rdf:datatype=\"%V\">\',\n"
"cast (dt as varchar)), _env);\n"
"else\n"
"http (sprintf (\'>\'), _env);\n"
"}\n"
"http_value (DB.DBA.RDF_SQLVAL_OF_LONG (_val), 0, _env);\n"
"http (\'</rs:value></rs:binding>\', _env);\n"
"}\n"
"end_of_binding: ;\n"
"\n"
"}\n"
"http (\'\\n  </rs:result>\', _env);\n"
"}\n"
"--src sparql.sql:2010\n";

static const char *proc78 = 
"create function DB.DBA.RDF_FORMAT_RESULT_SET_AS_RDF_XML_FIN (inout _env any) returns long varchar\n"
"{\n"
"if (185 <> __tag(_env))\n"
"DB.DBA.RDF_FORMAT_RESULT_SET_AS_RDF_XML_INIT (_env);\n"
"\n"
"http (\'\\n </rs:results>\\n</rdf:RDF>\', _env);\n"
"return string_output_string (_env);\n"
"}\n"
"--src sparql.sql:2079\n";

static const char *other8 = 
"create aggregate DB.DBA.RDF_FORMAT_RESULT_SET_AS_RDF_XML (in colvalues any, in colnames any) returns long varchar\n"
"from DB.DBA.RDF_FORMAT_RESULT_SET_AS_RDF_XML_INIT, DB.DBA.RDF_FORMAT_RESULT_SET_AS_RDF_XML_ACC, DB.DBA.RDF_FORMAT_RESULT_SET_AS_RDF_XML_FIN\n";

static const char *proc79 = 
"create function DB.DBA.RDF_FORMAT_TRIPLE_DICT_AS_TTL (inout triples_dict any) returns long varchar\n"
"{\n"
"declare triples, ses any;\n"
"ses := string_output ();\n"
"if (214 <> __tag (triples_dict))\n"
"{\n"
"triples := vector ();\n"
"}\n"
"else\n"
"triples := dict_list_keys (triples_dict, 1);\n"
"DB.DBA.RDF_TRIPLES_TO_TTL (triples, ses);\n"
"return ses;\n"
"}\n"
"--src sparql.sql:2093\n";

static const char *proc80 = 
"create function DB.DBA.RDF_FORMAT_TRIPLE_DICT_AS_RDF_XML (inout triples_dict any) returns long varchar\n"
"{\n"
"declare triples, ses any;\n"
"ses := string_output ();\n"
"if (214 <> __tag (triples_dict))\n"
"{\n"
"triples := vector ();\n"
"}\n"
"else\n"
"triples := dict_list_keys (triples_dict, 1);\n"
"DB.DBA.RDF_TRIPLES_TO_RDF_XML_TEXT (triples, 1, ses);\n"
"return ses;\n"
"}\n"
"--src sparql.sql:2108\n";

static const char *proc81 = 
" create procedure DB.DBA.RDF_FORMAT_BOOL_RESULT_AS_RDF_XML_INIT (inout _env any)\n"
"{\n"
"_env := 0;\n"
"}\n"
"--src sparql.sql:2124\n";

static const char *proc82 = 
" create procedure DB.DBA.RDF_FORMAT_BOOL_RESULT_AS_RDF_XML_ACC (inout _env any, inout one any)\n"
"{\n"
"_env := 1;\n"
"}\n"
"--src sparql.sql:2131\n";

static const char *proc83 = 
" create function DB.DBA.RDF_FORMAT_BOOL_RESULT_AS_RDF_XML_FIN (inout _env any) returns long varchar\n"
"{\n"
"declare ses any;\n"
"declare ans varchar;\n"
"ses := string_output ();\n"
"if (isinteger (_env) and _env)\n"
"ans := \'1\';\n"
"else\n"
"ans := \'0\';\n"
"http (\'<rdf:RDF\n"
"xmlns:rdf=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\"\n"
"xmlns:rs=\"http://www.w3.org/2005/sparql-results#\"\n"
"xmlns:xsd=\"http://www.w3.org/2001/XMLSchema#\" >\n"
"<rs:results rdf:nodeID=\"rset\">\n"
"<rs:boolean rdf:datatype=\"http://www.w3.org/2001/XMLSchema#boolean\">\' || ans || \'</rs:boolean></results></rdf:RDF>\', ses);\n"
"return ses;\n"
"}\n"
"--src sparql.sql:2138\n";

static const char *other9 = 
"create aggregate DB.DBA.RDF_FORMAT_BOOL_RESULT_AS_RDF_XML (inout one any) returns long varchar\n"
"from DB.DBA.RDF_FORMAT_BOOL_RESULT_AS_RDF_XML_INIT, DB.DBA.RDF_FORMAT_BOOL_RESULT_AS_RDF_XML_ACC, DB.DBA.RDF_FORMAT_BOOL_RESULT_AS_RDF_XML_FIN\n";

static const char *proc84 = 
" create procedure DB.DBA.RDF_FORMAT_BOOL_RESULT_AS_TTL_INIT (inout _env any)\n"
"{\n"
"_env := 0;\n"
"}\n"
"--src sparql.sql:2162\n";

static const char *proc85 = 
" create procedure DB.DBA.RDF_FORMAT_BOOL_RESULT_AS_TTL_ACC (inout _env any, inout one any)\n"
"{\n"
"_env := 1;\n"
"}\n"
"--src sparql.sql:2169\n";

static const char *proc86 = 
" create function DB.DBA.RDF_FORMAT_BOOL_RESULT_AS_TTL_FIN (inout _env any) returns long varchar\n"
"{\n"
"declare ses any;\n"
"declare ans varchar;\n"
"ses := string_output ();\n"
"if (isinteger (_env) and _env)\n"
"ans := \'TRUE\';\n"
"else\n"
"ans := \'FALSE\';\n"
"http (\'@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .\n"
"@prefix rs: <http://www.w3.org/2005/sparql-results#> .\\n\', ses);\n"
"http (sprintf (\'[ rdf:type rs:results ; rs:boolean %s ]\', ans), ses);\n"
"return ses;\n"
"}\n"
"--src sparql.sql:2176\n";

static const char *other10 = 
"create aggregate DB.DBA.RDF_FORMAT_BOOL_RESULT_AS_TTL (inout one any) returns long varchar\n"
"from DB.DBA.RDF_FORMAT_BOOL_RESULT_AS_TTL_INIT, DB.DBA.RDF_FORMAT_BOOL_RESULT_AS_TTL_ACC, DB.DBA.RDF_FORMAT_BOOL_RESULT_AS_TTL_FIN\n";

static const char *proc87 = 
"create function DB.DBA.RDF_INSERT_TRIPLES (in graph_iri any, in triples any)\n"
"{\n"
"declare ctr integer;\n"
"if (not isiri_id (graph_iri))\n"
"graph_iri := DB.DBA.RDF_MAKE_IID_OF_QNAME (graph_iri);\n"
"for (ctr := length (triples) - 1; ctr >= 0; ctr := ctr - 1)\n"
"{\n"
"\n"
"insert replacing DB.DBA.RDF_QUAD (G,S,P,O)\n"
"values (graph_iri, triples[ctr][0], triples[ctr][1], DB.DBA.RDF_OBJ_OF_LONG (triples[ctr][2]));\n"
"}\n"
"}\n"
"--src sparql.sql:2199\n";

static const char *proc88 = 
"create function DB.DBA.RDF_DELETE_TRIPLES (in graph_iri any, in triples any)\n"
"{\n"
"declare ctr integer;\n"
"if (not isiri_id (graph_iri))\n"
"graph_iri := DB.DBA.RDF_MAKE_IID_OF_QNAME (graph_iri);\n"
"for (ctr := length (triples) - 1; ctr >= 0; ctr := ctr - 1)\n"
"{\n"
"delete from DB.DBA.RDF_QUAD\n"
"where G = graph_iri and\n"
"S = triples[ctr][0] and\n"
"P = triples[ctr][1] and\n"
"O = DB.DBA.RDF_OBJ_OF_LONG (triples[ctr][2]);\n"
"}\n"
"}\n"
"--src sparql.sql:2213\n";

static const char *proc89 = 
"create function DB.DBA.SPARQL_INSERT_DICT_CONTENT (in graph_iri any, in triples_dict any)\n"
"{\n"
"declare triples any;\n"
"declare len integer;\n"
"triples := dict_list_keys (triples_dict, 1);\n"
"len := length (triples);\n"
"DB.DBA.RDF_INSERT_TRIPLES (graph_iri, triples);\n"
"return len;\n"
"}\n"
"--src sparql.sql:2229\n";

static const char *proc90 = 
"create function DB.DBA.SPARQL_DELETE_DICT_CONTENT (in graph_iri any, in triples_dict any)\n"
"{\n"
"declare triples any;\n"
"declare len integer;\n"
"triples := dict_list_keys (triples_dict, 1);\n"
"len := length (triples);\n"
"DB.DBA.RDF_DELETE_TRIPLES (graph_iri, triples);\n"
"return len;\n"
"}\n"
"--src sparql.sql:2240\n";

static const char *proc91 = 
" create function DB.DBA.RDF_REGEX (in s varchar, in p varchar, in coll varchar := null)\n"
"{\n"
"\n"
"if (regexp_match (p, s, 0) is not null)\n"
"return 1;\n"
"return 0;\n"
"}\n"
"--src sparql.sql:2255\n";

static const char *proc92 = 
" create function DB.DBA.RDF_LANGMATCHES (in r varchar, in t varchar)\n"
"{\n"
"if (\'*\' = t)\n"
"{\n"
"if (r <> \'\')\n"
"return 1;\n"
"return 0;\n"
"}\n"
"if ((t is null) or (r is null))\n"
"return 0;\n"
"t := toupper (t);\n"
"r := toupper (r);\n"
"if (r = t)\n"
"return 1;\n"
"if (r like t || \'-%\')\n"
"return 1;\n"
"return 0;\n"
"}\n"
"--src sparql.sql:2265\n";

static const char *proc93 = 
" create procedure DB.DBA.SPARQL_CONSTRUCT_INIT (inout _env any)\n"
"{\n"
"_env := 0;\n"
"}\n"
"--src sparql.sql:2286\n";

static const char *proc94 = 
" create procedure DB.DBA.SPARQL_CONSTRUCT_ACC (inout _env any, in opcodes any, in vars any, in stats any)\n"
"{\n"
"declare triple_ctr integer;\n"
"declare blank_ids any;\n"
"if (214 <> __tag(_env))\n"
"{\n"
"_env := dict_new ();\n"
"if (0 < length (stats))\n"
"DB.DBA.SPARQL_CONSTRUCT_ACC (_env, stats, vector(), vector());\n"
"}\n"
"blank_ids := 0;\n"
"for (triple_ctr := length (opcodes) - 1; triple_ctr >= 0; triple_ctr := triple_ctr-1)\n"
"{\n"
"declare fld_ctr integer;\n"
"declare triple_vec any;\n"
"triple_vec := vector (0,0,0);\n"
"for (fld_ctr := 2; fld_ctr >= 0; fld_ctr := fld_ctr - 1)\n"
"{\n"
"declare op integer;\n"
"declare arg any;\n"
"op := opcodes[triple_ctr][fld_ctr * 2];\n"
"arg := opcodes[triple_ctr][fld_ctr * 2 + 1];\n"
"if (1 = op)\n"
"{\n"
"declare i IRI_ID;\n"
"i := vars[arg];\n"
"if (i is null)\n"
"goto end_of_adding_triple;\n"
"if ((2 > fld_ctr) and not isiri_id (i))\n"
"signal (\'RDF01\',\n"
"sprintf (\'Bad variable value in CONSTRUCT: \"%.100s\" is not a valid %s, only object of a triple can be a literal\',\n"
"DB.DBA.RDF_STRSQLVAL_OF_LONG (i),\n"
"case (fld_ctr) when 1 then \'predicate\' else \'subject\' end ) );\n"
"if ((1 = fld_ctr) and isiri_id (i) and (i >= #i1000000000))\n"
"signal (\'RDF01\', \'Bad variable value in CONSTRUCT: blank node can not be used as predicate\');\n"
"triple_vec[fld_ctr] := i;\n"
"}\n"
"else if (2 = op)\n"
"{\n"
"if (isinteger (blank_ids))\n"
"blank_ids := vector (iri_id_from_num (sequence_next (\'RDF_URL_IID_BLANK\')));\n"
"while (arg >= length (blank_ids))\n"
"blank_ids := vector_concat (blank_ids, vector (iri_id_from_num (sequence_next (\'RDF_URL_IID_BLANK\'))));\n"
"if (1 = fld_ctr)\n"
"signal (\'RDF01\', \'Bad triple for CONSTRUCT: blank node can not be used as predicate\');\n"
"triple_vec[fld_ctr] := blank_ids[arg];\n"
"}\n"
"else if (3 = op)\n"
"{\n"
"if (arg is null)\n"
"goto end_of_adding_triple;\n"
"if ((2 > fld_ctr) and not isiri_id (arg))\n"
"signal (\'RDF01\', sprintf (\'Bad const value in CONSTRUCT: \"%.100s\" is not a valid %s, only object of a triple can be a literal\',\n"
"DB.DBA.RDF_STRSQLVAL_OF_LONG (arg),\n"
"case (fld_ctr) when 1 then \'predicate\' else \'subject\' end ) );\n"
"if ((1 = fld_ctr) and isiri_id (arg) and (arg >= #i1000000000))\n"
"signal (\'RDF01\', \'Bad const value in CONSTRUCT: blank node can not be used as predicate\');\n"
"triple_vec[fld_ctr] := arg;\n"
"}\n"
"else signal (\'RDFXX\', \'Bad opcode in DB.DBA.SPARQL_CONSTRUCT()\');\n"
"}\n"
"\n"
"dict_put (_env, triple_vec, 0);\n"
"end_of_adding_triple: ;\n"
"}\n"
"}\n"
"--src sparql.sql:2293\n";

static const char *proc95 = 
" create procedure DB.DBA.SPARQL_CONSTRUCT_FIN (inout _env any)\n"
"{\n"
"if (214 <> __tag(_env))\n"
"_env := dict_new ();\n"
"return _env;\n"
"}\n"
"--src sparql.sql:2362\n";

static const char *other11 = 
" create aggregate DB.DBA.SPARQL_CONSTRUCT (in opcodes any, in vars any, in stats any) returns any\n"
"from DB.DBA.SPARQL_CONSTRUCT_INIT, DB.DBA.SPARQL_CONSTRUCT_ACC, DB.DBA.SPARQL_CONSTRUCT_FIN\n";

static const char *proc96 = 
"create procedure DB.DBA.SPARQL_DESCRIBE_INIT (inout _env any)\n"
"{\n"
"_env := 0;\n"
"}\n"
"--src sparql.sql:2375\n";

static const char *proc97 = 
"create procedure DB.DBA.SPARQL_DESCRIBE_ACC (inout _env any, in vars any, in stats any, in options any)\n"
"{\n"
"declare var_ctr integer;\n"
"declare blank_ids any;\n"
"if (193 <> __tag(_env))\n"
"{\n"
"_env := vector (dict_new (), options);\n"
"if (0 < length (stats))\n"
"DB.DBA.SPARQL_DESCRIBE_ACC (_env, stats, vector(), vector());\n"
"}\n"
"for (var_ctr := length (vars) - 1; var_ctr >= 0; var_ctr := var_ctr - 1)\n"
"{\n"
"declare i any;\n"
"i := vars[var_ctr];\n"
"if (isiri_id (i))\n"
"dict_put (_env[0], i, 0);\n"
"}\n"
"}\n"
"--src sparql.sql:2381\n";

static const char *proc98 = 
"create procedure DB.DBA.SPARQL_DESCRIBE_FIN (inout _env any)\n"
"{\n"
"declare subjects, options, res any;\n"
"declare subj_ctr integer;\n"
"if (193 <> __tag(_env))\n"
"return dict_new ();\n"
"subjects := dict_list_keys (_env[0], 1);\n"
"options := _env[1];\n"
"res := dict_new ();\n"
"for (subj_ctr := length (subjects) - 1; subj_ctr >= 0; subj_ctr := subj_ctr - 1)\n"
"{\n"
"declare subj any;\n"
"subj := subjects[subj_ctr];\n"
"DB.DBA.RDF_DESCRIBE_PUT (res, subj, options);\n"
"}\n"
"\n"
"\n"
"\n"
"\n"
"\n"
"\n"
"\n"
"\n"
"\n"
"\n"
"\n"
"return res;\n"
"}\n"
"--src sparql.sql:2401\n";

static const char *other12 = 
"create aggregate DB.DBA.SPARQL_DESCRIBE (in opcodes any, in vars any, in stats any) returns any\n"
"from DB.DBA.SPARQL_DESCRIBE_INIT, DB.DBA.SPARQL_DESCRIBE_ACC, DB.DBA.SPARQL_DESCRIBE_FIN\n";

static const char *proc99 = 
"create procedure DB.DBA.RDF_DESCRIBE_PUT (in dict any, in subj IRI_ID, inout options any)\n"
"{\n"
"\n"
"for (select G as g1, P as p1, O as obj1 from DB.DBA.RDF_QUAD where S = subj) do\n"
"{\n"
"dict_put (dict, vector (subj, p1, DB.DBA.RDF_LONG_OF_OBJ (obj1)), 0);\n"
"if (isiri_id (obj1))\n"
"{\n"
"for (select P as p2, O as obj2\n"
"from DB.DBA.RDF_QUAD\n"
"where G = g1 and S = obj1 and not (isiri_id (O)) ) do\n"
"{\n"
"dict_put (dict, vector (obj1, p2, DB.DBA.RDF_LONG_OF_OBJ (obj2)), 0);\n"
"}\n"
"}\n"
"}\n"
"}\n"
"--src sparql.sql:2435\n";

static const char *proc100 = 
"create function DB.DBA.RDF_TYPEMIN_OF_OBJ (in obj any) returns any\n"
"{\n"
"declare tag integer;\n"
"if (obj is null)\n"
"return NULL;\n"
"tag := __tag (obj);\n"
"if (tag = 182)\n"
"\n"
"return concat (subseq (obj, 0, 1), \'\\0\\0\\0\');\n"
"if (tag in (189, 191, 219))\n"
"return -3.40282347e+38;\n"
"if (tag = 211)\n"
"return cast (\'0101-01-01\' as datetime);\n"
"return NULL;\n"
"}\n"
"--src sparql.sql:2458\n";

static const char *proc101 = 
"create function DB.DBA.RDF_TYPEMAX_OF_OBJ (in obj any) returns any\n"
"{\n"
"declare tag integer;\n"
"if (obj is null)\n"
"return NULL;\n"
"tag := __tag (obj);\n"
"if (tag = 182)\n"
"\n"
"return concat (subseq (obj, 0, 1), \'\\377\\377\\377\\377\\377\\377\\377\\377\\377\\377\\377\\377\\0\\377\\377\');\n"
"if (tag in (189, 191, 219))\n"
"return 3.40282347e+38;\n"
"if (tag = 211)\n"
"return cast (\'9999-12-30\' as datetime);\n"
"return NULL;\n"
"}\n"
"--src sparql.sql:2475\n";

static const char *proc102 = 
"create function DB.DBA.RDF_IID_CMP (in obj1 any, in obj2 any) returns integer\n"
"{\n"
"return NULL;\n"
"}\n"
"--src sparql.sql:2492\n";

static const char *proc103 = 
"create function DB.DBA.RDF_OBJ_CMP (in obj1 any, in obj2 any) returns integer\n"
"{\n"
"declare tag1, tag2 integer;\n"
"if (obj1 is null or obj2 is null)\n"
"return NULL;\n"
"tag1 := __tag (obj1);\n"
"tag2 := __tag (obj2);\n"
"if (tag1 = 182)\n"
"{\n"
"if (tag2 = 182)\n"
"{\n"
"declare l, len1, len2, id1, id2 integer;\n"
"declare begin1, begin2, full1, full2 varchar;\n"
"if (obj1 = obj2)\n"
"return 0;\n"
"if (subseq (obj1, 0, 1) <> subseq (obj1, 0, 1))\n"
"return null;\n"
"len1 := length (obj1);\n"
"len2 := length (obj2);\n"
"\n"
"\n"
"l := 20;\n"
"if (len1 = (l + 9))\n"
"{\n"
"begin1 := subseq (obj1, 2, len1-7);\n"
"id1 := obj1 [l+3] + 256 * (obj1 [l+4] + 256 * (obj1 [l+5] + 256 * obj1 [l+6]));\n"
"}\n"
"else\n"
"{\n"
"begin1 := subseq (obj1, 2, len1-3);\n"
"id1 := 0;\n"
"}\n"
"if (len2 = (l + 9))\n"
"{\n"
"begin2 := subseq (obj2, 2, len2-7);\n"
"id2 := obj2 [l+3] + 256 * (obj2 [l+4] + 256 * (obj2 [l+5] + 256 * obj2 [l+6]));\n"
"}\n"
"else\n"
"{\n"
"begin2 := subseq (obj2, 2, len2-3);\n"
"id2 := 0;\n"
"}\n"
"if (begin1 <> begin2)\n"
"{\n"
"if (begin1 < begin2)\n"
"return -1;\n"
"return 1;\n"
"}\n"
"if (id1 = id2)\n"
"return 0;\n"
"full1 := (select case (isnull (RO_LONG)) when 0 then blob_to_string (RO_LONG) else RO_VAL end from DB.DBA.RDF_OBJ where RO_ID = id1);\n"
"if (full1 is null)\n"
"signal (\'RDFXX\', sprintf (\'Integrity violation in DB.DBA.RDF_OBJ_CMP, bad id %d\', id1));\n"
"full2 := (select case (isnull (RO_LONG)) when 0 then blob_to_string (RO_LONG) else RO_VAL end from DB.DBA.RDF_OBJ where RO_ID = id2);\n"
"if (full2 is null)\n"
"signal (\'RDFXX\', sprintf (\'Integrity violation in DB.DBA.RDF_OBJ_CMP, bad id %d\', id2));\n"
"if (full1 <> full2)\n"
"{\n"
"if (full1 < full2)\n"
"return -1;\n"
"return 1;\n"
"}\n"
"return 0;\n"
"}\n"
"return null;\n"
"}\n"
"if (tag1 in (189, 191, 219))\n"
"{\n"
"if (tag2 in (189, 191, 219))\n"
"{\n"
"if (obj1 <> obj2)\n"
"{\n"
"if (obj1 < obj2)\n"
"return -1;\n"
"return 1;\n"
"}\n"
"return 0;\n"
"}\n"
"return null;\n"
"}\n"
"if (tag1 = 211)\n"
"{\n"
"if (tag2 = 211)\n"
"{\n"
"if (obj1 <> obj2)\n"
"{\n"
"if (obj1 < obj2)\n"
"return -1;\n"
"return 1;\n"
"}\n"
"return 0;\n"
"}\n"
"return null;\n"
"}\n"
"return NULL;\n"
"}\n"
"--src sparql.sql:2498\n";

static const char *proc104 = 
"create function DB.DBA.RDF_LONG_CMP (in long1 any, in long2 any) returns integer\n"
"{\n"
"declare tag1, tag2 integer;\n"
"if (long1 is null or long2 is null)\n"
"return NULL;\n"
"tag1 := __tag (long1);\n"
"tag2 := __tag (long2);\n"
"if (tag1 = 193)\n"
"{\n"
"if (tag2 = 193)\n"
"{\n"
"declare full1, full2 varchar;\n"
"if (long1[0] <> long2[0])\n"
"return null;\n"
"full1 := long1 [1];\n"
"full2 := long2 [1];\n"
"if (full1 <> full2)\n"
"{\n"
"if (full1 < full2)\n"
"return -1;\n"
"return 1;\n"
"}\n"
"return 0;\n"
"}\n"
"return null;\n"
"}\n"
"if (tag1 in (189, 191, 219))\n"
"{\n"
"if (tag2 in (189, 191, 219))\n"
"{\n"
"if (long1 <> long2)\n"
"{\n"
"if (long1 < long2)\n"
"return -1;\n"
"return 1;\n"
"}\n"
"return 0;\n"
"}\n"
"return null;\n"
"}\n"
"if (tag1 = 211)\n"
"{\n"
"if (tag2 = 211)\n"
"{\n"
"if (long1 <> long2)\n"
"{\n"
"if (long1 < long2)\n"
"return -1;\n"
"return 1;\n"
"}\n"
"return 0;\n"
"}\n"
"return null;\n"
"}\n"
"return NULL;\n"
"}\n"
"--src sparql.sql:2596\n";

static const char *proc105 = 
"create function JSO_MAKE_INHERITANCE (in jgraph varchar, in class varchar, in rootinst varchar, in destinst varchar, in dest_iid iri_id, inout noinherits any, inout inh_stack any)\n"
"{\n"
"declare base_iid iri_id;\n"
"declare baseinst varchar;\n"
"\n"
"inh_stack := vector_concat (inh_stack, vector (destinst));\n"
"baseinst := null;\n"
"if (not exists (sparql\n"
"define input:storage \"\"\n"
"prefix virtrdf: <http://www.openlinksw.com/schemas/virtrdf#>\n"
"ask where {\n"
"graph ?:jgraph { ?:dest_iid rdf:type `iri(?:class)`\n"
"} } ) )\n"
"signal (\'22023\', \'JSO_MAKE_INHERITANCE has not found object <\' || destinst || \'> of type <\' || class || \'>\');\n"
"/* This fails. !!!TBD: fix sparql2sql.c to preserve data about equalities, fixed values and globals when triples are moved from gp to gp\n"
"for (sparql\n"
"define input:storage \"\"\n"
"prefix virtrdf: <http://www.openlinksw.com/schemas/virtrdf#>\n"
"select ?srcinst\n"
"where {\n"
"graph ?:jgraph {\n"
"{ {\n"
"?destnode rdf:type `iri(?:class)` .\n"
"filter (?destnode = iri(?:destinst)) }\n"
"union\n"
"{\n"
"?destnode rdf:type `iri(?:class)` .\n"
"?destnode rdf:name `iri(?:destinst)` } } .\n"
"?destnode virtrdf:inheritFrom ?srcinst .\n"
"?srcinst rdf:type `iri(?:class)` .\n"
"} } ) do\n"
"*/\n"
"for (sparql\n"
"define input:storage \"\"\n"
"define output:valmode \"LONG\"\n"
"prefix virtrdf: <http://www.openlinksw.com/schemas/virtrdf#>\n"
"select ?src_iid\n"
"where {\n"
"graph ?:jgraph { ?:dest_iid virtrdf:inheritFrom ?src_iid } } ) do\n"
"{\n"
"declare srcinst varchar;\n"
"srcinst := DB.DBA.RDF_QNAME_OF_IID (\"src_iid\");\n"
"if (baseinst is null)\n"
"{\n"
"if (not exists (sparql\n"
"define input:storage \"\"\n"
"prefix virtrdf: <http://www.openlinksw.com/schemas/virtrdf#>\n"
"ask where { graph ?:jgraph { ?:\"src_iid\" rdf:type `iri(?:class)` } } ) )\n"
"signal (\'22023\', \'JSO_MAKE_INHERITANCE has found that the object <\' || destinst || \'> has wrong virtrdf:inheritFrom <\' || srcinst || \'> that is not an instance of type <\' || class || \'>\');\n"
"base_iid := \"src_iid\";\n"
"baseinst := srcinst;\n"
"}\n"
"else if (baseinst <> srcinst)\n"
"signal (\'22023\', \'JSO_MAKE_INHERITANCE has found that the object <\' || destinst || \'> has multiple virtrdf:inheritFrom declarations: <\' || baseinst || \'> and <\' || srcinst || \'>\');\n"
"}\n"
"if (position (baseinst, inh_stack))\n"
"signal (\'22023\', \'JSO_MAKE_INHERITANCE has found that the object <\' || baseinst || \'> is recursively inherited from itself\');\n"
"/* This fails. !!!TBD: fix sparql2sql.c to preserve data about equalities, fixed values and globals when triples are moved from gp to gp\n"
"for (sparql\n"
"define input:storage \"\"\n"
"prefix virtrdf: <http://www.openlinksw.com/schemas/virtrdf#>\n"
"select ?pred\n"
"where {\n"
"graph ?:jgraph {\n"
"{ {\n"
"?destnode rdf:type `iri(?:class)` .\n"
"filter (?destnode = iri(?:destinst)) }\n"
"union\n"
"{\n"
"?destnode rdf:type `iri(?:class)` .\n"
"?destnode rdf:name `iri(?:destinst)` } } .\n"
"?destnode virtrdf:noInherit ?pred .\n"
"} } ) do\n"
"*/\n"
"for (sparql\n"
"define input:storage \"\"\n"
"prefix virtrdf: <http://www.openlinksw.com/schemas/virtrdf#>\n"
"select ?pred\n"
"where {\n"
"graph ?:jgraph {\n"
"?:dest_iid virtrdf:noInherit ?pred\n"
"} } ) do\n"
"{\n"
"if (baseinst is null)\n"
"signal (\'22023\', \'JSO_MAKE_INHERITANCE has found that the object <\' || destinst || \'> has set virtrdf:noInherit but has no virtrdf:inheritFrom\');\n"
"dict_put (noinherits, \"pred\", destinst);\n"
"}\n"
"if (baseinst is null)\n"
"return;\n"
"for (select \"pred\", \"predval\"\n"
"from (sparql\n"
"define input:storage \"\"\n"
"define output:valmode \"LONG\"\n"
"prefix virtrdf: <http://www.openlinksw.com/schemas/virtrdf#>\n"
"select ?pred, ?predval\n"
"where {\n"
"graph ?:jgraph {\n"
"?:base_iid ?pred ?predval\n"
"} } ) as \"t00\"\n"
"where not exists (sparql\n"
"define input:storage \"\"\n"
"prefix virtrdf: <http://www.openlinksw.com/schemas/virtrdf#>\n"
"ask where { graph ?:jgraph { ?:\"t00\".\"pred\" virtrdf:loadAs virtrdf:jsoTriple } } )\n"
") do\n"
"{\n"
"\"pred\" := DB.DBA.RDF_QNAME_OF_IID (\"pred\");\n"
"if (DB.DBA.RDF_LANGUAGE_OF_LONG (\"predval\") is not null)\n"
"signal (\'22023\', \'JSO_MAKE_INHERITANCE does not support language marks on objects\');\n"
"if (\'http://www.w3.org/1999/02/22-rdf-syntax-ns#type\' = \"pred\")\n"
" ;\n"
"else if (\'http://www.w3.org/1999/02/22-rdf-syntax-ns#name\' = \"pred\")\n"
" ;\n"
"else if (\'http://www.openlinksw.com/schemas/virtrdf#inheritFrom\' = \"pred\")\n"
" ;\n"
"else if (\'http://www.openlinksw.com/schemas/virtrdf#noInherit\' = \"pred\")\n"
" ;\n"
"else if (dict_get (noinherits, \"pred\", baseinst) = baseinst)\n"
"{\n"
"jso_set (class, rootinst, \"pred\", DB.DBA.RDF_SQLVAL_OF_LONG (\"predval\"), isiri_id (\"predval\"));\n"
"dict_put (noinherits, \"pred\", baseinst);\n"
"}\n"
"}\n"
"JSO_MAKE_INHERITANCE (jgraph, class, rootinst, baseinst, base_iid, noinherits, inh_stack);\n"
"}\n"
"--src sparql.sql:2658\n";

static const char *proc106 = 
"create function JSO_LOAD_INSTANCE (in jgraph varchar, in jinst varchar, in delete_first integer, in make_new integer, in jsubj_iid iri_id := 0)\n"
"{\n"
"declare jinst_iid, jgraph_iid IRI_ID;\n"
"declare jclass varchar;\n"
"declare noinherits, inh_stack any;\n"
"\n"
"noinherits := dict_new ();\n"
"jinst_iid := DB.DBA.RDF_MAKE_IID_OF_QNAME (jinst);\n"
"jgraph_iid := DB.DBA.RDF_MAKE_IID_OF_QNAME (jgraph);\n"
"if (jsubj_iid is null)\n"
"{\n"
"jsubj_iid := (sparql\n"
"define input:storage \"\"\n"
"define output:valmode \"LONG\"\n"
"select ?s\n"
"where { graph ?:jgraph { ?s rdf:name ?:jinst } } );\n"
"if (jsubj_iid is null)\n"
"jsubj_iid := jinst_iid;\n"
"}\n"
"jclass := (sparql\n"
"define input:storage \"\"\n"
"select ?t\n"
"where {\n"
"graph ?:jgraph { ?:jsubj_iid rdf:type ?t } } );\n"
"if (jclass is null)\n"
"{\n"
"if (exists (sparql\n"
"define input:storage \"\"\n"
"select ?x\n"
"where { graph ?:jgraph {\n"
"{ ?:jinst ?x ?o }\n"
"union\n"
"{ ?x rdf:name ?ji .\n"
"filter (str (?ji) = ?:jinst)\n"
"} } } ) )\n"
"signal (\'22023\', \'JSO_LOAD_INSTANCE can not detect the type of <\' || jinst || \'>\');\n"
"else\n"
"signal (\'22023\', \'JSO_LOAD_INSTANCE can not find an object <\' || jinst || \'>\');\n"
"}\n"
"if (delete_first)\n"
"jso_delete (jclass, jinst, 1);\n"
"if (make_new)\n"
"jso_new (jclass, jinst);\n"
"for (select \"p\", coalesce (\"o2\", \"o1\") as \"o\"\n"
"from (sparql\n"
"define input:storage \"\"\n"
"define output:valmode \"LONG\"\n"
"select ?p ?o1 ?o2\n"
"where {\n"
"graph ?:jgraph {\n"
"{ ?:jsubj_iid ?p ?o1 }  optional { ?o1 rdf:name ?o2 }\n"
"} }\n"
") as \"t00\"\n"
"where not exists (sparql\n"
"define input:storage \"\"\n"
"prefix virtrdf: <http://www.openlinksw.com/schemas/virtrdf#>\n"
"ask where { graph ?:jgraph_iid { ?:\"t00\".\"p\" virtrdf:loadAs virtrdf:jsoTriple } } )\n"
") do\n"
"{\n"
"\"p\" := DB.DBA.RDF_QNAME_OF_IID (\"p\");\n"
"if (DB.DBA.RDF_LANGUAGE_OF_LONG (\"o\") is not null)\n"
"signal (\'22023\', \'JSO_LOAD_INSTANCE does not support language marks on objects\');\n"
"if (\'http://www.w3.org/1999/02/22-rdf-syntax-ns#type\' = \"p\")\n"
"{\n"
"if (DB.DBA.RDF_SQLVAL_OF_LONG (\"o\") <> jclass)\n"
"signal (\'22023\', \'JSO_LOAD_INSTANCE has found that the object <\' || jinst || \'> has multiple type declarations\');\n"
"}\n"
"else if (\'http://www.w3.org/1999/02/22-rdf-syntax-ns#name\' = \"p\")\n"
" ;\n"
"else if (\'http://www.openlinksw.com/schemas/virtrdf#inheritFrom\' = \"p\")\n"
" ;\n"
"else if (\'http://www.openlinksw.com/schemas/virtrdf#noInherit\' = \"p\")\n"
" ;\n"
"else\n"
"{\n"
"jso_set (jclass, jinst, \"p\", DB.DBA.RDF_SQLVAL_OF_LONG (\"o\"), isiri_id (\"o\"));\n"
"dict_put (noinherits, \"p\", jinst);\n"
"}\n"
"}\n"
"inh_stack := vector ();\n"
"JSO_MAKE_INHERITANCE (jgraph, jclass, jinst, jinst, jsubj_iid, noinherits, inh_stack);\n"
"}\n"
"--src sparql.sql:2784\n";

static const char *proc107 = 
"create procedure JSO_LIST_INSTANCES_OF_GRAPH (in jgraph varchar, out instances any)\n"
"{\n"
"\n"
"instances := (\n"
"select vector_agg (\n"
"vector (\n"
"DB.DBA.RDF_QNAME_OF_IID (\"jclass\"),\n"
"DB.DBA.RDF_QNAME_OF_IID (\"jinst\"),\n"
"coalesce (\"s\", \"jinst\") ) )\n"
"from ( sparql\n"
"define output:valmode \"LONG\"\n"
"define input:storage \"\"\n"
"select ?jclass ?jinst ?s\n"
"where {\n"
"graph ?:jgraph {\n"
"{ ?jinst rdf:type ?jclass .\n"
"filter (!isBLANK (?jinst)) }\n"
"union\n"
"{ ?s rdf:type ?jclass .\n"
"?s rdf:name ?jinst .\n"
"filter (isBLANK (?s))\n"
"} } }\n"
") as inst );\n"
"\n"
"}\n"
"--src sparql.sql:2868\n";

static const char *proc108 = 
"create function JSO_LOAD_GRAPH (in jgraph varchar, in pin_now integer := 1)\n"
"{\n"
"declare jgraph_iid IRI_ID;\n"
"declare instances, chk any;\n"
"\n"
"jgraph_iid := DB.DBA.RDF_MAKE_IID_OF_QNAME (jgraph);\n"
"JSO_LIST_INSTANCES_OF_GRAPH (jgraph, instances);\n"
"/* Pass 1. Deleting all obsolete instances. */\n"
"foreach (any j in instances) do\n"
"jso_delete (j[0], j[1], 1);\n"
"/* Pass 2. Creating all instances. */\n"
"foreach (any j in instances) do\n"
"jso_new (j[0], j[1]);\n"
"/* Pass 3. Loading all instances, including loading inherited values. */\n"
"foreach (any j in instances) do\n"
"JSO_LOAD_INSTANCE (jgraph, j[1], 0, 0, j[2]);\n"
"/* Pass 4. Validation all instances. */\n"
"foreach (any j in instances) do\n"
"jso_validate (j[0], j[1], 1);\n"
"/* Pass 5. Pin all instances. */\n"
"if (pin_now)\n"
"{\n"
"foreach (any j in instances) do\n"
"jso_pin (j[0], j[1]);\n"
"}\n"
"/* Pass 6. Load all separate triples */\n"
"for (sparql\n"
"define input:storage \"\"\n"
"prefix virtrdf: <http://www.openlinksw.com/schemas/virtrdf#>\n"
"select ?s ?p ?o\n"
"where { graph ?:jgraph_iid { ?p virtrdf:loadAs virtrdf:jsoTriple . ?s ?p ?o } } ) do\n"
"jso_triple_add (\"s\", \"p\", \"o\");\n"
"chk := jso_triple_get_objs (\n"
"UNAME\'http://www.openlinksw.com/schemas/virtrdf#loadAs\',\n"
"UNAME\'http://www.openlinksw.com/schemas/virtrdf#loadAs\' );\n"
"if ((1 <> length (chk)) or (cast (chk[0] as varchar) <> \'http://www.openlinksw.com/schemas/virtrdf#jsoTriple\'))\n"
"signal (\'22023\', \'JSO_LOAD_GRAPH has not found expected metadata in the graph\');\n"
"}\n"
"--src sparql.sql:2895\n";

static const char *proc109 = 
"create function JSO_PIN_GRAPH (in jgraph varchar)\n"
"{\n"
"declare instances any;\n"
"JSO_LIST_INSTANCES_OF_GRAPH (jgraph, instances);\n"
"foreach (any j in instances) do\n"
"jso_pin (j[0], j[1]);\n"
"}\n"
"--src sparql.sql:2935\n";

static const char *proc110 = 
"create function JSO_SYS_GRAPH () returns varchar\n"
"{\n"
"return \'http://www.openlinksw.com/schemas/virtrdf#\';\n"
"}\n"
"--src sparql.sql:2944\n";

static const char *proc111 = 
"create function JSO_DUMP_IRI (in v varchar, inout ses any)\n"
"{\n"
"\n"
"\n"
"if (v like \'http://www.w3.org/2000/01/rdf-schema#%\')\n"
"{ http (\'rdfs:\' || subseq (v, 37), ses); return; }\n"
"\n"
"\n"
"if (v like \'http://www.w3.org/1999/02/22-rdf-syntax-ns#%\')\n"
"{ http (\'rdf:\' || subseq (v, 43), ses); return; }\n"
"\n"
"\n"
"if (v like \'http://www.w3.org/2002/07/owl#%\')\n"
"{ http (\'owl:\' || subseq (v, 30), ses); return; }\n"
"\n"
"\n"
"if (v like \'http://www.w3.org/2001/XMLSchema#%\')\n"
"{ http (\'xsd:\' || subseq (v, 33), ses); return; }\n"
"\n"
"\n"
"if (v like \'http://www.openlinksw.com/schemas/virtrdf#%\')\n"
"{ http (\'virtrdf:\' || subseq (v, 42), ses); return; }\n"
"\n"
"\n"
"if (v like \'http://www.openlinksw.com/virtrdf-data-formats#%\')\n"
"{ http (\'rdfdf:\' || subseq (v, 47), ses); return; }\n"
"http (\'<\', ses);\n"
"http_escape (v, 12, ses, 1, 1);\n"
"http (\'>\', ses);\n"
"}\n"
"--src sparql.sql:2950\n";

static const char *proc112 = 
"create function JSO_DUMP_FLD (in v any, inout ses any)\n"
"{\n"
"declare v_tag integer;\n"
"v_tag := __tag(v);\n"
"if (v_tag = 217)\n"
"JSO_DUMP_IRI (cast (v as varchar), ses);\n"
"else if (v_tag = 203)\n"
"http (jso_dbg_dump_rtti (v), ses);\n"
"else if (v_tag = 182)\n"
"{\n"
"http (\'\"\', ses);\n"
"http_escape (v, 11, ses, 1, 1);\n"
"http (\'\"\', ses);\n"
"}\n"
"else if (isinteger (v))\n"
"http_value (v, 0, ses);\n"
"else\n"
"{\n"
"http (\'\"\', ses);\n"
"http_escape (DB.DBA.RDF_STRSQLVAL_OF_LONG (v), 11, ses, 1, 1);\n"
"http (\'\"^^<\', ses);\n"
"http_escape (cast (DB.DBA.RDF_DATATYPE_OF_TAG (__tag (v)) as varchar), 12, ses, 1, 1);\n"
"http (\'>\', ses);\n"
"}\n"
"}\n"
"--src sparql.sql:2982\n";

static const char *proc113 = 
"create function JSO_DUMP_ALL () returns any\n"
"{\n"
"declare proplist, ses any;\n"
"declare prev_obj any;\n"
"declare ctr, len integer;\n"
"ses := string_output ();\n"
"proplist := jso_proplist ();\n"
"gvector_sort (proplist, 1, 0, 1);\n"
"prev_obj := null;\n"
"len := length (proplist);\n"
"for (ctr := 0; ctr < len; ctr := ctr+1)\n"
"{\n"
"declare obj any;\n"
"obj := proplist[ctr][0];\n"
"if (obj = prev_obj)\n"
"http (\';\\n  \', ses);\n"
"else\n"
"{\n"
"if (prev_obj is null)\n"
"http (\n"
"\'@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .\n"
"@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .\n"
"@prefix owl: <http://www.w3.org/2002/07/owl#> .\n"
"@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .\n"
"@prefix virtrdf: <http://www.openlinksw.com/schemas/virtrdf#> .\n"
"@prefix rdfdf: <http://www.openlinksw.com/virtrdf-data-formats#> .\n"
"\', ses );\n"
"else\n"
"http (\'.\\n\', ses);\n"
"prev_obj := obj;\n"
"JSO_DUMP_FLD (obj, ses);\n"
"http (\'\\n  \', ses);\n"
"}\n"
"JSO_DUMP_FLD (proplist[ctr][1], ses);\n"
"http (\'\\t\', ses);\n"
"JSO_DUMP_FLD (proplist[ctr][2], ses);\n"
"}\n"
"if (prev_obj is not null)\n"
"http (\'.\\n\', ses);\n"
"return ses;\n"
"}\n"
"--src sparql.sql:3009\n";

static const char *proc114 = 
"create procedure DB.DBA.RDF_QM_CHANGE (in warninglist any)\n"
"{\n"
"declare STATE, MESSAGE varchar;\n"
"result_names (STATE, MESSAGE);\n"
"foreach (any warnings in warninglist) do\n"
"{\n"
"foreach (any warning in warnings) do\n"
"result (warning[0], warning[1]);\n"
"}\n"
"commit work;\n"
"}\n"
"--src sparql.sql:3055\n";

static const char *proc115 = 
"create function DB.DBA.RDF_QM_APPLY_CHANGES (in deleted any, in affected any) returns any\n"
"{\n"
"declare ctr, len integer;\n"
"declare graphiri varchar;\n"
"graphiri := JSO_SYS_GRAPH ();\n"
"commit work;\n"
"DB.DBA.JSO_LOAD_GRAPH (graphiri, 0);\n"
"DB.DBA.JSO_PIN_GRAPH (graphiri);\n"
"len := length (deleted);\n"
"for (ctr := 0; ctr < len; ctr := ctr + 2)\n"
"jso_delete (deleted [ctr], deleted [ctr+1], 1);\n"
"len := length (affected);\n"
"for (ctr := 0; ctr < len; ctr := ctr + 1)\n"
"jso_mark_affected (affected [ctr]);\n"
"return vector (vector (\'00000\', \'Transaction committed, SPARQL compiler re-configured\'));\n"
"}\n"
"--src sparql.sql:3068\n";

static const char *proc116 = 
"create function DB.DBA.RDF_QM_ASSERT_JSO_TYPE (in inst varchar, in expected varchar, in allow_missing integer := 0) returns integer\n"
"{\n"
"declare actual varchar;\n"
"if (expected is null)\n"
"{\n"
"actual := coalesce ((sparql\n"
"define input:storage \"\"\n"
"prefix virtrdf: <http://www.openlinksw.com/schemas/virtrdf#>\n"
"select ?t where {\n"
"graph `sql:JSO_SYS_GRAPH NIL` { `iri(?:inst)` rdf:type ?t } } ));\n"
"if (actual is not null)\n"
"signal (\'22023\', \'The RDF QM schema object <\' || inst || \'> already exists, type <\' || cast (actual as varchar) || \'>\');\n"
"}\n"
"else\n"
"{\n"
"declare hit integer;\n"
"hit := 0;\n"
"for (sparql\n"
"define input:storage \"\"\n"
"prefix virtrdf: <http://www.openlinksw.com/schemas/virtrdf#>\n"
"select ?t where {\n"
"graph `sql:JSO_SYS_GRAPH NIL` { `iri(?:inst)` rdf:type ?t } } ) do\n"
"{\n"
"if (\"t\" <> expected)\n"
"signal (\'22023\', \'The RDF QM schema object <\' || inst || \'> has type <\' || cast (actual as varchar) || \'>, cannot use same identifier for <\' || expected || \'>\');\n"
"hit := 1;\n"
"}\n"
"if (not hit)\n"
"{\n"
"if (allow_missing)\n"
"return 0;\n"
"signal (\'22023\', \'The RDF QM schema object <\' || inst || \'> does not exists, should be of type <\' || expected || \'>\');\n"
"}\n"
"}\n"
"return 1;\n"
"}\n"
"--src sparql.sql:3086\n";

static const char *proc117 = 
"create procedure DB.DBA.RDF_QM_ASSERT_STORAGE_FLAG (in storage varchar, in req_flag integer)\n"
"{\n"
"declare graphiri varchar;\n"
"graphiri := DB.DBA.JSO_SYS_GRAPH ();\n"
"DB.DBA.RDF_QM_ASSERT_JSO_TYPE (storage, \'http://www.openlinksw.com/schemas/virtrdf#QuadStorage\');\n"
"for (sparql define input:storage \"\"\n"
"prefix rdfdf: <http://www.openlinksw.com/virtrdf-data-formats#>\n"
"select ?startdt where {\n"
"graph ?:graphiri {\n"
"`iri(?:storage)` virtrdf:qsAlterInProgress ?startdt .\n"
"} } ) do\n"
"{\n"
"if (req_flag)\n"
"return;\n"
"signal (\'22023\', \'The quad storage \"\' || storage || \'\" is edited by other client, started \' || cast (\"startdt\" as varchar));\n"
"}\n"
"if (not req_flag)\n"
"return;\n"
"signal (\'22023\', \'The quad storage \"\' || storage || \'\" is not flagged as being edited, cannot change it\' );\n"
"}\n"
"--src sparql.sql:3124\n";

static const char *proc118 = 
"create procedure DB.DBA.RDF_QM_ASSERT_STORAGE_CONTAINS_MAPPING (in storage varchar, in qmid varchar, in must_contain integer)\n"
"{\n"
"declare graphiri varchar;\n"
"graphiri := DB.DBA.JSO_SYS_GRAPH ();\n"
"if (exists (sparql define input:storage \"\"\n"
"ask where {\n"
"graph ?:graphiri {\n"
"{ `iri(?:storage)` virtrdf:qsDefaultMap `iri(?:qmid)` }\n"
"union\n"
"{ `iri(?:storage)` virtrdf:qsUserMaps ?qmlist .\n"
"?qmlist ?p `iri(?:qmid)` .\n"
"} } } ) )\n"
"{\n"
"if (must_contain)\n"
"return;\n"
"signal (\'22023\', \'The quad storage \"\' || storage || \'\" contains quad map \' || qmid );\n"
"}\n"
"if (not must_contain)\n"
"return;\n"
"signal (\'22023\', \'The quad storage \"\' || storage || \'\" does not contains quad map \' || qmid );\n"
"}\n"
"--src sparql.sql:3146\n";

static const char *proc119 = 
"create procedure DB.DBA.RDF_QM_ASSERT_STORAGE_IS_FLAGGED (in storage varchar)\n"
"{\n"
"if (not DB.DBA.RDF_QM_GET_STORAGE_FLAG (storage))\n"
"signal (\'22023\', \'The quad storage \"\' || storage || \'\" is not flagged as being edited\' );\n"
"}\n"
"--src sparql.sql:3169\n";

static const char *proc120 = 
"create function DB.DBA.RDF_QM_GC_SUBTREE (in id varchar) returns integer\n"
"{\n"
"declare graphiri varchar;\n"
"declare objs any;\n"
"\n"
"graphiri := DB.DBA.JSO_SYS_GRAPH ();\n"
"for (sparql define input:storage \"\" select ?s where {\n"
"graph ?g { ?s ?p `iri(?:id)` } } ) do\n"
"{\n"
"\n"
"return \"s\";\n"
"}\n"
"objs := (select VECTOR_AGG (sub.\"o\") from\n"
"(sparql define input:storage \"\"\n"
"select ?o where {\n"
"graph ?:graphiri { `iri(?:id)` ?p ?o . filter (!isliteral (?o)) } } ) as sub );\n"
"for (sparql define input:storage \"\"\n"
"delete from graph ?:graphiri { `iri(?:id)` ?p ?o }\n"
"where { graph ?:graphiri { `iri(?:id)` ?p ?o } } ) do {;}\n"
"foreach (varchar obj in objs) do\n"
"DB.DBA.RDF_QM_GC_SUBTREE (obj);\n"
"return NULL;\n"
"}\n"
"--src sparql.sql:3176\n";

static const char *proc121 = 
"create function DB.DBA.RDF_QM_DROP_MAPPING (in storage varchar, in mapname any) returns any\n"
"{\n"
"declare qmid, qmgraph varchar;\n"
"qmid := get_keyword_ucase (\'ID\', mapname, NULL);\n"
"qmgraph := get_keyword_ucase (\'GRAPH\', mapname, NULL);\n"
"if (qmid is null)\n"
"{\n"
"qmid := coalesce ((sparql\n"
"define input:storage \"\"\n"
"prefix virtrdf: <http://www.openlinksw.com/schemas/virtrdf#>\n"
"select ?s where {\n"
"graph `sql:JSO_SYS_GRAPH NIL` {\n"
"?s rdf:type virtrdf:QuadMap .\n"
"?s virtrdf:qmGraphRange-rvrFixedValue `iri(?:qmgraph)` .\n"
"?s virtrdf:qmTableName \"\" .\n"
"} } ));\n"
"if (qmid is null)\n"
"return vector (vector (\'00100\', \'Quad map for graph <\' || qmgraph || \'> is not found\'));\n"
"}\n"
"DB.DBA.RDF_QM_ASSERT_JSO_TYPE (qmid, \'http://www.openlinksw.com/schemas/virtrdf#QuadMap\');\n"
"if (storage is null)\n"
"{\n"
"delete from RDF_QUAD\n"
"where G = DB.DBA.RDF_MAKE_IID_OF_QNAME (DB.DBA.JSO_SYS_GRAPH()) and\n"
"O = DB.DBA.RDF_MAKE_IID_OF_QNAME (qmid);\n"
"DB.DBA.RDF_QM_GC_SUBTREE (qmid);\n"
"delete from RDF_QUAD\n"
"where G = DB.DBA.RDF_MAKE_IID_OF_QNAME (DB.DBA.JSO_SYS_GRAPH()) and\n"
"S = DB.DBA.RDF_MAKE_IID_OF_QNAME (qmid);\n"
"return vector (vector (\'00000\', \'Quad map <\' || qmid || \'> is deleted\'));\n"
"}\n"
"else\n"
"{\n"
"\n"
"\n"
"\n"
"\n"
"\n"
"\n"
"\n"
"\n"
"\n"
"\n"
"\n"
"\n"
"\n"
"DB.DBA.RDF_QM_DELETE_MAPPING_FROM_STORAGE (storage, NULL, qmid);\n"
"return vector (vector (\'00000\', \'Quad map <\' || qmid || \'> is no longer used in storage <\' || storage || \'>\'));\n"
"}\n"
"}\n"
"--src sparql.sql:3201\n";

static const char *proc122 = 
"create function DB.DBA.RDF_QM_DEFINE_IRI_CLASS_FORMAT (in classiri varchar, in iritmpl varchar, in arglist any) returns any\n"
"{\n"
"declare graphiri varchar;\n"
"declare superformatsid varchar;\n"
"declare res any;\n"
"declare arglist_len integer;\n"
"graphiri := DB.DBA.JSO_SYS_GRAPH ();\n"
"if (strstr (iritmpl, \'^{URIQADefaultHost}^\') is not null)\n"
"{\n"
"declare host varchar;\n"
"host := registry_get (\'URIQADefaultHost\');\n"
"if (not isstring (host))\n"
"signal (\'22023\', \'Can not use ^{URIQADefaultHost}^ in IRI template if there is no DefaultHost parameter in [URIQA] section of Virtuoso configuration file\');\n"
"iritmpl := replace (iritmpl, \'^{URIQADefaultHost}^\', host);\n"
"}\n"
"superformatsid := classiri || \'--SuperFormats\';\n"
"foreach (any arg in arglist) do\n"
"if (UNAME\'in\' <> arg[0])\n"
"signal (\'22023\', \'Only \"in\" parameters are now supported in argument lists of class formats, \"\' || arg[0] || \'\" is not supported in CREATE IRI CLASS <\' || classiri || \'>\' );\n"
"if (length (arglist) = 0)\n"
"signal (\'22023\', \'Empty argument list in CREATE IRI CLASS <\' || classiri || \'>\');\n"
"if (DB.DBA.RDF_QM_ASSERT_JSO_TYPE (classiri, \'http://www.openlinksw.com/schemas/virtrdf#QuadMapFormat\', 1))\n"
"{\n"
"declare side_s varchar;\n"
"side_s := DB.DBA.RDF_QM_GC_SUBTREE (classiri);\n"
"if (side_s is not null)\n"
"signal (\'22023\', \'Can not change iri class <\' || classiri || \'> because it is used by other quad map objects, e.g., <\' || side_s || \'>\');\n"
"res := vector (vector (\'00000\', \'Previous definition of IRI class <\' || classiri || \'> has been dropped\'));\n"
"}\n"
"else\n"
"res := vector ();\n"
"arglist_len := length (arglist);\n"
"if (arglist_len > 1)\n"
"{\n"
"for (sparql define input:storage \"\"\n"
"prefix rdfdf: <http://www.openlinksw.com/virtrdf-data-formats#>\n"
"insert in graph ?:graphiri\n"
"{\n"
"`iri(?:classiri)`\n"
"rdf:type virtrdf:QuadMapFormat ;\n"
"virtrdf:inheritFrom rdfdf:multipart-uri ;\n"
"virtrdf:noInherit virtrdf:qmfCustomString1 ;\n"
"virtrdf:qmfCustomString1 `?:iritmpl` ;\n"
"virtrdf:qmfColumnCount ?:arglist_len ;\n"
"virtrdf:qmfSuperFormats `iri(?:superformatsid)` .\n"
"`iri(?:superformatsid)`\n"
"rdf:type virtrdf:array-of-QuadMapFormat .\n"
"}\n"
"where {} ) do {;}\n"
"return vector_concat (res, vector (vector (\'00000\', \'IRI class <\' || classiri || \'> has been defined (inherited from rdfdf:multipart-uri)\')));\n"
"}\n"
"else /* arglist is 1 item long */\n"
"{\n"
"declare basetype, basetypeiri varchar;\n"
"basetype := lower (arglist[0][2]);\n"
"if (not (basetype in (\'integer\', \'varchar\', \'date\', \'datetime\', \'doubleprecision\')))\n"
"signal (\'22023\', \'The datatype \"\' || basetype || \'\" is not supported in CREATE IRI CLASS <\' || classiri || \'>\' );\n"
"basetype := \'sql-\' || basetype || \'-uri\';\n"
"if (not (coalesce (arglist[0][3], 0)))\n"
"basetype := basetype || \'-nullable\';\n"
"basetypeiri := \'http://www.openlinksw.com/virtrdf-data-formats#\' || basetype;\n"
"for (sparql define input:storage \"\"\n"
"prefix rdfdf: <http://www.openlinksw.com/virtrdf-data-formats#>\n"
"insert in graph ?:graphiri\n"
"{\n"
"`iri(?:classiri)`\n"
"rdf:type virtrdf:QuadMapFormat;\n"
"virtrdf:inheritFrom `iri(?:basetypeiri)`;\n"
"virtrdf:noInherit virtrdf:qmfCustomString1;\n"
"virtrdf:qmfCustomString1 ?:iritmpl;\n"
"virtrdf:qmfColumnCount 1 ;\n"
"virtrdf:qmfSuperFormats `iri(?:superformatsid)` .\n"
"`iri(?:superformatsid)`\n"
"rdf:type virtrdf:array-of-QuadMapFormat .\n"
"}\n"
"where {} ) do {;}\n"
"return vector_concat (res, vector (vector (\'00000\', \'IRI class <\' || classiri || \'> has been defined (inherited from rdfdf:\' || basetype || \')\')));\n"
"}\n"
"return vector_concat (res, vector (vector (\'22023\', \'DB.DBA.RDF_QM_DEFINE_IRI_CLASS_FORMAT() is not yet implemented\')));\n"
"}\n"
"--src sparql.sql:3253\n";

static const char *proc123 = 
"create function DB.DBA.RDF_QM_DEFINE_IRI_CLASS_FUNCTIONS (in classiri varchar, in uriprint any, in uriparse any) returns any\n"
"{\n"
"/*\n"
"uriprint is, say,\n"
"vector ( \'DB.DBA.RDF_DF_GRANTEE_ID_URI\' ,\n"
"vector (\n"
"vector ( 306,  \'id\' ,  \'integer\' ,  NULL ) ),  \'varchar\' ,  NULL ),\n"
"uriparse is, say,\n"
"vector ( \'DB.DBA.RDF_DF_GRANTEE_ID_URI_INVERSE\' ,\n"
"vector (\n"
"vector ( 306,  \'id_iri\' ,  \'varchar\' ,  NULL ) ),  \'integer\' ,  NULL ) ),\n"
"*/\n"
"declare uriprintname, uriparsename varchar;\n"
"declare numofnotnulls integer;\n"
"declare graphiri varchar;\n"
"declare superformatsid varchar;\n"
"declare res any;\n"
"graphiri := DB.DBA.JSO_SYS_GRAPH ();\n"
"superformatsid := classiri || \'--SuperFormats\';\n"
"DB.DBA.RDF_QM_ASSERT_JSO_TYPE (classiri, NULL);\n"
"uriprintname := uriprint[0];\n"
"uriparsename := uriparse[0];\n"
"if (uriparsename <> uriprintname || \'_INVERSE\')\n"
"signal (\'22023\', \'Name of inverse function should be \"\' || uriprintname || \'_INVERSE\", not \"\' || uriparsename || \'\", other variants are not supported by the current version\' );\n"
"if (uriprint[2] <> \'varchar\')\n"
"signal (\'22023\', \'IRI composing function \"\' || uriprintname || \'\" should return varchar, not \' || uriprint[2]);\n"
"if (1 <> length (uriprint[1]))\n"
"signal (\'22023\', \'IRI composing function \"\' || uriprintname || \'\" should have exactly one argument\');\n"
"foreach (any arg in uriprint[1]) do\n"
"if (UNAME\'in\' <> arg[0])\n"
"signal (\'22023\', \'Only \"in\" parameters are now supported in argument lists of IRI class functions, not \"\' || arg[0] || \'\"\');\n"
"if (1 <> length (uriparse[1]))\n"
"signal (\'22023\', \'IRI parsing function \"\' || uriparsename || \'\" should have exactly one argument\');\n"
"foreach (any arg in uriparse[1]) do\n"
"if (UNAME\'in\' <> arg[0])\n"
"signal (\'22023\', \'Only \"in\" parameters are now supported in argument lists of IRI class functions, not \"\' || arg[0] || \'\"\');\n"
"if (uriparse[1][0][2] <> \'varchar\')\n"
"signal (\'22023\', \'IRI parsing function \"\' || uriparsename || \'\" should have argument of type varchar, not \' || uriparse[1][0][2]);\n"
"if (uriparse[2] <> uriprint[1][0][2])\n"
"signal (\'22023\', \'The return value of \"\' || uriparsename || \'\" and the argument of \"\' || uriprintname || \'\" should be of the same data type\');\n"
"numofnotnulls :=\n"
"coalesce (uriprint[1][0][3], 0) + coalesce (uriprint[3], 0) +\n"
"coalesce (uriparse[1][0][3], 0) + coalesce (uriparse[3], 0);\n"
"if ((0 <> numofnotnulls) and (4 <> numofnotnulls))\n"
"signal (\'22023\', \'Both arguments and return values of both \"\' || uriparsename || \'\" and \"\' || uriprintname || \'\" should be either all four nullable or all four NOT NULL\');\n"
"declare arglist, basetype, basetypeiri varchar;\n"
"arglist := uriprint[1];\n"
"basetype := lower (arglist[0][2]);\n"
"if (not (basetype in (\'integer\', \'varchar\' /*, \'date\', \'doubleprecision\'*/)))\n"
"signal (\'22023\', \'The datatype \"\' || basetype || \'\" is not supported in CREATE IRI CLASS <\' || classiri || \'> USING FUNCTION\' );\n"
"basetype := \'sql-\' || basetype || \'-uri-fn\';\n"
"if (not (coalesce (arglist[0][3], 0)))\n"
"basetype := basetype || \'-nullable\';\n"
"basetypeiri := \'http://www.openlinksw.com/virtrdf-data-formats#\' || basetype;\n"
"if (DB.DBA.RDF_QM_ASSERT_JSO_TYPE (classiri, \'http://www.openlinksw.com/schemas/virtrdf#QuadMapFormat\', 1))\n"
"{\n"
"declare side_s varchar;\n"
"side_s := DB.DBA.RDF_QM_GC_SUBTREE (classiri);\n"
"if (side_s is not null)\n"
"signal (\'22023\', \'Can not change iri class <\' || classiri || \'> because it is used by other quad map objects, e.g., <\' || side_s || \'>\');\n"
"res := vector (vector (\'00000\', \'Previous definition of IRI class <\' || classiri || \'> has been dropped\'));\n"
"}\n"
"else\n"
"res := vector ();\n"
"for (sparql define input:storage \"\"\n"
"prefix rdfdf: <http://www.openlinksw.com/virtrdf-data-formats#>\n"
"insert in graph ?:graphiri\n"
"{\n"
"`iri(?:classiri)`\n"
"rdf:type virtrdf:QuadMapFormat;\n"
"virtrdf:inheritFrom `iri(?:basetypeiri)`;\n"
"virtrdf:noInherit virtrdf:qmfCustomString1;\n"
"virtrdf:qmfCustomString1 ?:uriprintname;\n"
"virtrdf:qmfSuperFormats `iri(?:superformatsid)` .\n"
"`iri(?:superformatsid)`\n"
"rdf:type virtrdf:array-of-QuadMapFormat .\n"
"}\n"
"where {} ) do {;}\n"
"return vector_concat (res, vector (vector (\'00000\', \'IRI class <\' || classiri || \'> has been defined (inherited from rdfdf:\' || basetype || \') using \' || uriprintname)));\n"
"}\n"
"--src sparql.sql:3335\n";

static const char *proc124 = 
"create function DB.DBA.RDF_QM_DEFINE_SUBCLASS (in subclassiri varchar, in superclassiri varchar) returns any\n"
"{\n"
"declare graphiri varchar;\n"
"graphiri := DB.DBA.JSO_SYS_GRAPH ();\n"
"DB.DBA.RDF_QM_ASSERT_JSO_TYPE (subclassiri, \'http://www.openlinksw.com/schemas/virtrdf#QuadMapFormat\');\n"
"DB.DBA.RDF_QM_ASSERT_JSO_TYPE (superclassiri, \'http://www.openlinksw.com/schemas/virtrdf#QuadMapFormat\');\n"
"for (sparql define input:storage \"\"\n"
"prefix rdfdf: <http://www.openlinksw.com/virtrdf-data-formats#>\n"
"insert in graph ?:graphiri\n"
"{\n"
"`iri(?:subclassiri)` virtrdf:isSubclassOf `iri(?:superclassiri)` .\n"
"}\n"
"where {} ) do {;}\n"
"return vector (vector (\'00000\', \'IRI class <\' || subclassiri || \'> is now known as a subclass of <\' || superclassiri || \'>\'));\n"
"}\n"
"--src sparql.sql:3417\n";

static const char *proc125 = 
"create function DB.DBA.RDF_QM_DROP_QUAD_STORAGE (in storage varchar) returns any\n"
"{\n"
"declare graphiri varchar;\n"
"graphiri := DB.DBA.JSO_SYS_GRAPH ();\n"
"DB.DBA.RDF_QM_ASSERT_STORAGE_FLAG (storage, 0);\n"
"DB.DBA.RDF_QM_GC_SUBTREE (storage);\n"
"for (sparql define input:storage \"\"\n"
"prefix rdfdf: <http://www.openlinksw.com/virtrdf-data-formats#>\n"
"delete from graph ?:graphiri {\n"
"`iri(?:storage)` ?p ?o\n"
"}\n"
"where { graph ?:graphiri {\n"
"`iri(?:storage)` ?p ?o .\n"
"} } ) do {;}\n"
"return vector (vector (\'00000\', \'Quad storage <\' || storage || \'> is removed from the quad mapping schema\'));\n"
"}\n"
"--src sparql.sql:3434\n";

static const char *proc126 = 
"create function DB.DBA.RDF_QM_DEFINE_QUAD_STORAGE (in storage varchar) returns any\n"
"{\n"
"declare graphiri, qsusermaps varchar;\n"
"graphiri := DB.DBA.JSO_SYS_GRAPH ();\n"
"DB.DBA.RDF_QM_ASSERT_JSO_TYPE (storage, NULL);\n"
"qsusermaps := storage || \'--UserMaps\';\n"
"for (sparql define input:storage \"\"\n"
"prefix rdfdf: <http://www.openlinksw.com/virtrdf-data-formats#>\n"
"insert in graph ?:graphiri {\n"
"`iri(?:storage)`\n"
"rdf:type virtrdf:QuadStorage ;\n"
"virtrdf:qsUserMaps `iri(?:qsusermaps)` .\n"
"`iri(?:qsusermaps)`\n"
"rdf:type virtrdf:array-of-QuadMap .\n"
"}\n"
"where {} ) do {;}\n"
"return vector (vector (\'00000\', \'A new empty quad storage <\' || storage || \'> is added to the quad mapping schema\'));\n"
"}\n"
"--src sparql.sql:3452\n";

static const char *proc127 = 
"create function DB.DBA.RDF_QM_BEGIN_ALTER_QUAD_STORAGE (in storage varchar) returns any\n"
"{\n"
"declare graphiri varchar;\n"
"graphiri := DB.DBA.JSO_SYS_GRAPH ();\n"
"DB.DBA.RDF_QM_ASSERT_STORAGE_FLAG (storage, 0);\n"
"for (sparql define input:storage \"\"\n"
"prefix rdfdf: <http://www.openlinksw.com/virtrdf-data-formats#>\n"
"insert in graph ?:graphiri {\n"
"`iri(?:storage)` virtrdf:qsAlterInProgress `bif:now NIL` .\n"
"}\n"
"where {} ) do {;}\n"
"return vector (vector (\'00000\', \'Quad storage <\' || storage || \'> is flagged as being edited\'));\n"
"}\n"
"--src sparql.sql:3472\n";

static const char *proc128 = 
"create function DB.DBA.RDF_QM_END_ALTER_QUAD_STORAGE (in storage varchar) returns any\n"
"{\n"
"declare graphiri varchar;\n"
"graphiri := DB.DBA.JSO_SYS_GRAPH ();\n"
"DB.DBA.RDF_QM_ASSERT_JSO_TYPE (storage, \'http://www.openlinksw.com/schemas/virtrdf#QuadStorage\');\n"
"for (sparql define input:storage \"\"\n"
"prefix rdfdf: <http://www.openlinksw.com/virtrdf-data-formats#>\n"
"delete from graph ?:graphiri {\n"
"`iri(?:storage)` virtrdf:qsAlterInProgress ?dtstart\n"
"}\n"
"where { graph ?:graphiri {\n"
"`iri(?:storage)` virtrdf:qsAlterInProgress ?dtstart .\n"
"} } ) do {;}\n"
"return vector (vector (\'00000\', \'Quad storage <\' || storage || \'> is unflagged and can be edited by other transactions\'));\n"
"}\n"
"--src sparql.sql:3487\n";

static const char *proc129 = 
"create function DB.DBA.RDF_QM_DEFINE_MAP_VALUE (in qmv any, in fldname varchar, inout tablename varchar) returns varchar\n"
"{\n"
"/* iqi qmv: vector ( UNAME\'http://www.openlinksw.com/schemas/oplsioc#user_iri\' , vector ( vector (\'DB.DBA.SYS_USERS\', \'alias1\', \'U_ID\') ) ) */\n"
"declare graphiri varchar;\n"
"declare atables, sqlcols, conds any;\n"
"declare atablectr, atablecount integer;\n"
"declare colctr, colcount integer;\n"
"declare condctr, condcount integer;\n"
"declare fmtid, iriclassid, qmvid, qmvatablesid, qmvcolsid, qmvcondsid varchar;\n"
"graphiri := DB.DBA.JSO_SYS_GRAPH ();\n"
"fmtid := qmv[0];\n"
"if (fmtid <> UNAME\'literal\')\n"
"DB.DBA.RDF_QM_ASSERT_JSO_TYPE (fmtid, \'http://www.openlinksw.com/schemas/virtrdf#QuadMapFormat\');\n"
"atables := qmv[1];\n"
"sqlcols := qmv[2];\n"
"conds := qmv[3];\n"
"atablecount := length (atables);\n"
"colcount := length (sqlcols);\n"
"condcount := length (conds);\n"
"for (colctr := 0; colctr < colcount; colctr := colctr + 1)\n"
"{\n"
"declare sqlcol any;\n"
"sqlcol := sqlcols [colctr];\n"
"if (not exists (select top 1 1 from DB.DBA.SYS_COLS where \"TABLE\" = sqlcol[0] and \"COLUMN\" = sqlcol[2]))\n"
"{\n"
"if (sqlcol[1] is not null)\n"
"signal (\'22023\', \'No column \' || sqlcol[2] || \' in table \' || sqlcol[0] || \' (alias \' || sqlcol[1] || \') in database, please check spelling and character case\');\n"
"else\n"
"signal (\'22023\', \'No column \' || sqlcol[2] || \' in table \' || sqlcol[0] || \' in database, please check spelling and character case\');\n"
"}\n"
"if (tablename is null)\n"
"tablename := sqlcol[0];\n"
"else if (tablename <> sqlcol[0])\n"
"tablename := \'\';\n"
"}\n"
"if (fmtid = UNAME\'literal\')\n"
"{\n"
"declare coldtp, colnullable integer;\n"
"declare coltype varchar;\n"
"select COL_DTP, COL_NULLABLE into coldtp, colnullable\n"
"from DB.DBA.SYS_COLS where \"TABLE\" = sqlcols[0][0] and \"COLUMN\" = sqlcols[0][2];\n"
"coltype := case (coldtp)\n"
"when 125 then \'longvarchar\'\n"
"when 128 then \'datetime\'\n"
"when 129 then \'date\'\n"
"when 131 then \'longvarbinary\'\n"
"when 189 then \'integer\'\n"
"when 182 then \'varchar\'\n"
"when 211 then \'datetime\'\n"
"else NULL end;\n"
"if (coltype is null)\n"
"signal (\'22023\', \'The datatype of column \"\' || sqlcols[0][2] ||\n"
"\'\" of table \"\' || sqlcols[0][0] || \'\" (COL_DTP=\' || cast (coldtp as varchar) ||\n"
"\') can not be mapped to an RDF literal in current version of Virtuoso\' );\n"
"fmtid := \'http://www.openlinksw.com/virtrdf-data-formats#sql-\' || coltype;\n"
"if (colnullable)\n"
"fmtid := fmtid || \'-nullable\';\n"
"iriclassid := null;\n"
"}\n"
"else\n"
"iriclassid := fmtid;\n"
"qmvid := \'sys:qmv-\' || md5 (serialize (vector (fmtid, sqlcols)));\n"
"qmvatablesid := qmvid || \'-atables\';\n"
"qmvcolsid := qmvid || \'-cols\';\n"
"qmvcondsid := qmvid || \'-conds\';\n"
"/* Trick to avoid repeating re-declarations */\n"
"if (exists (sparql define input:storage \"\"\n"
"prefix rdfdf: <http://www.openlinksw.com/virtrdf-data-formats#>\n"
"ask where {\n"
"graph ?:graphiri {\n"
"?:qmvid\n"
"rdf:type virtrdf:QuadMapValue ;\n"
"virtrdf:qmvATables `iri(?:qmvatablesid)` ;\n"
"virtrdf:qmvColumns `iri(?:qmvcolsid)` ;\n"
"virtrdf:qmvConds `iri(?:qmvcondsid)` ;\n"
"virtrdf:qmvFormat `iri(?:fmtid)` . } } ) )\n"
"return qmvid;\n"
"/* Create everything if qmv has not been found */\n"
"for (sparql define input:storage \"\"\n"
"prefix rdfdf: <http://www.openlinksw.com/virtrdf-data-formats#>\n"
"delete from graph ?:graphiri {\n"
"`iri(?:qmvid)` ?p ?o . }\n"
"where { graph ?:graphiri {\n"
"`iri(?:qmvid)` ?p ?o .\n"
"} } ) do {;}\n"
"for (sparql define input:storage \"\"\n"
"prefix rdfdf: <http://www.openlinksw.com/virtrdf-data-formats#>\n"
"select ?atable where { graph ?:graphiri {\n"
"`iri(?:qmvatablesid)` ?p ?atable } } ) do {\n"
"DB.DBA.RDF_QM_GC_SUBTREE (\"atable\");\n"
"}\n"
"for (sparql define input:storage \"\"\n"
"prefix rdfdf: <http://www.openlinksw.com/virtrdf-data-formats#>\n"
"delete from graph ?:graphiri {\n"
"`iri(?:qmvatablesid)` ?p ?o }\n"
"where { graph ?:graphiri {\n"
"`iri(?:qmvatablesid)` ?p ?o .\n"
"} } ) do {;}\n"
"for (sparql define input:storage \"\"\n"
"prefix rdfdf: <http://www.openlinksw.com/virtrdf-data-formats#>\n"
"select ?col where { graph ?:graphiri {\n"
"`iri(?:qmvcolsid)` ?p ?col } } ) do {\n"
"DB.DBA.RDF_QM_GC_SUBTREE (\"col\");\n"
"}\n"
"for (sparql define input:storage \"\"\n"
"prefix rdfdf: <http://www.openlinksw.com/virtrdf-data-formats#>\n"
"delete from graph ?:graphiri {\n"
"`iri(?:qmvcolsid)` ?p ?o }\n"
"where { graph ?:graphiri {\n"
"`iri(?:qmvcolsid)` ?p ?o .\n"
"} } ) do {;}\n"
"for (sparql define input:storage \"\"\n"
"prefix rdfdf: <http://www.openlinksw.com/virtrdf-data-formats#>\n"
"delete from graph ?:graphiri {\n"
"`iri(?:qmvcondsid)` ?p ?o }\n"
"where { graph ?:graphiri {\n"
"`iri(?:qmvcondsid)` ?p ?o .\n"
"} } ) do {;}\n"
"if (0 = atablecount)\n"
"qmvatablesid := NULL;\n"
"if (0 = condcount)\n"
"qmvcondsid := NULL;\n"
"for (sparql define input:storage \"\"\n"
"prefix rdfdf: <http://www.openlinksw.com/virtrdf-data-formats#>\n"
"insert in graph ?:graphiri {\n"
"`iri(?:qmvid)`\n"
"rdf:type virtrdf:QuadMapValue;\n"
"virtrdf:qmvTableName ?:tablename;\n"
"virtrdf:qmvATables `iri(?:qmvatablesid)`;\n"
"virtrdf:qmvColumns `iri(?:qmvcolsid)`;\n"
"virtrdf:qmvConds `iri(?:qmvcondsid)`;\n"
"virtrdf:qmvFormat `iri(?:fmtid)`;\n"
"virtrdf:qmvIriClass `iri(?:iriclassid)`;\n"
"virtrdf:qmvColumnsFormKey 0 .  #!!!TBD: detection of unique keys.\n"
"`iri(?:qmvatablesid)`\n"
"rdf:type virtrdf:array-of-QuadMapATable .\n"
"`iri(?:qmvcolsid)`\n"
"rdf:type virtrdf:array-of-QuadMapColumn .\n"
"`iri(?:qmvcondsid)`\n"
"rdf:type virtrdf:array-of-string .\n"
"}\n"
"where {} ) do {;}\n"
"for (atablectr := 0; atablectr < atablecount; atablectr := atablectr + 1)\n"
"{\n"
"declare pair any;\n"
"declare qtable, alias, inner_id varchar;\n"
"pair := atables [atablectr];\n"
"alias := pair[0];\n"
"qtable := pair[1];\n"
"inner_id := qmvid || \'-atable-\' || alias || \'-\' || qtable;\n"
"for (sparql define input:storage \"\"\n"
"prefix rdfdf: <http://www.openlinksw.com/virtrdf-data-formats#>\n"
"insert in graph ?:graphiri {\n"
"`iri(?:qmvatablesid)`\n"
"`iri (bif:sprintf (\"%s%d\", str (rdf:_), ?:atablectr+1))` `iri(?:inner_id)` .\n"
"`iri(?:inner_id)`  \n"
"rdf:type virtrdf:QuadMapATable ;\n"
"virtrdf:qmvaAlias ?:alias ;\n"
"virtrdf:qmvaTableName ?:qtable .\n"
"}\n"
"where {} ) do {;}\n"
"}\n"
"for (colctr := 0; colctr < colcount; colctr := colctr + 1)\n"
"{\n"
"declare sqlcol any;\n"
"declare qtable, alias, colname, inner_id varchar;\n"
"sqlcol := sqlcols [colctr];\n"
"alias := sqlcol[1];\n"
"colname := sqlcol[2];\n"
"inner_id := qmvid || \'-col-\' || alias || \'-\' || colname;\n"
"for (sparql define input:storage \"\"\n"
"prefix rdfdf: <http://www.openlinksw.com/virtrdf-data-formats#>\n"
"insert in graph ?:graphiri {\n"
"`iri(?:qmvcolsid)`\n"
"`iri (bif:sprintf (\"%s%d\", str (rdf:_), ?:colctr+1))` `iri(?:inner_id)` .\n"
"`iri(?:inner_id)`  \n"
"rdf:type virtrdf:QuadMapColumn ;\n"
"virtrdf:qmvcAlias ?:alias ;\n"
"virtrdf:qmvcColumnName ?:colname .\n"
"}\n"
"where {} ) do {;}\n"
"}\n"
"for (condctr := 0; condctr < condcount; condctr := condctr + 1)\n"
"{\n"
"declare sqlcond varchar;\n"
"sqlcond := conds [condctr];\n"
"for (sparql define input:storage \"\"\n"
"prefix rdfdf: <http://www.openlinksw.com/virtrdf-data-formats#>\n"
"insert in graph ?:graphiri {\n"
"`iri(?:qmvcondsid)`\n"
"`iri (bif:sprintf (\"%s%d\", str (rdf:_), ?:condctr+1))` ?:sqlcond .\n"
"}\n"
"where {} ) do {;}\n"
"}\n"
"return qmvid;\n"
"}\n"
"--src sparql.sql:3504\n";

static const char *proc130 = 
"create procedure DB.DBA.RDF_QM_NORMALIZE_QMV (\n"
"inout qmv any, inout qmvfix any, inout qmvid any,\n"
"in can_be_literal integer, in fldname varchar, inout tablename varchar )\n"
"{\n"
"qmvid := qmvfix := NULL;\n"
"if ((193 = __tag (qmv)) and (4 = length (qmv)))\n"
"qmvid := DB.DBA.RDF_QM_DEFINE_MAP_VALUE (qmv, fldname, tablename);\n"
"else if (217 = __tag (qmv))\n"
"qmvfix := DB.DBA.RDF_MAKE_IID_OF_QNAME (qmv);\n"
"else if (qmv is not null and not can_be_literal)\n"
"signal (\'22023\', sprintf (\'Quad map declaration can not specify a literal (non-IRI) constant for its %s (tag %d, length %d)\',\n"
"fldname, __tag (qmv), length (qmv) ) );\n"
"else if (193 = __tag (qmv))\n"
"signal (\'22023\', sprintf (\'Quad map declaration contains constant %s of unsupported type (tag %d, length %d)\',\n"
"fldname, __tag (qmv), length (qmv) ) );\n"
"else\n"
"qmvfix := qmv;\n"
"}\n"
"--src sparql.sql:3702\n";

static const char *proc131 = 
"create function DB.DBA.RDF_QM_DEFINE_MAPPING (in storage varchar,\n"
"in qmrawid varchar, in qmid varchar, in qmparentid varchar,\n"
"in qmv_g any, in qmv_s any, in qmv_p any, in qmv_o any,\n"
"in is_real integer, in rowfilter varchar, in opts any ) returns any\n"
"{\n"
"declare graphiri, old_actual_type varchar;\n"
"declare tablename, qmvid_g, qmvid_s, qmvid_p, qmvid_o varchar;\n"
"declare qmvfix_g, qmvfix_s, qmvfix_p, qmvfix_o any;\n"
"declare qm_exclusive, qm_empty, qm_is_default, qmusersubmaps varchar;\n"
"declare qm_order integer;\n"
"\n"
"graphiri := DB.DBA.JSO_SYS_GRAPH ();\n"
"DB.DBA.RDF_QM_ASSERT_STORAGE_FLAG (storage, 1);\n"
"\n"
"old_actual_type := coalesce ((sparql define input:storage \"\"\n"
"prefix virtrdf: <http://www.openlinksw.com/schemas/virtrdf#>\n"
"select ?t where {\n"
"graph ?:graphiri { `iri(?:qmid)` rdf:type ?t } } ));\n"
"if (old_actual_type is not null)\n"
"{\n"
"declare old_lstiri, old_side_use varchar;\n"
"if (old_actual_type <> \'http://www.openlinksw.com/schemas/virtrdf#QuadMap\')\n"
"signal (\'22023\', \'The RDF QM schema object <\' || qmid || \'> already exists, type <\' || old_actual_type || \'>\');\n"
"old_lstiri := (sparql define input:storage \"\"\n"
"select ?lst where { graph ?:graphiri {\n"
"`iri(?:storage)` virtrdf:qsUserMaps ?lst } } );\n"
"old_side_use := coalesce ((sparql define input:storage \"\"\n"
"select ?s where {\n"
"graph ?:graphiri { ?s ?p `iri(?:qmid)` filter ((?s != iri(?:storage)) && (?s != iri(?:old_lstiri))) } } ) );\n"
"if (old_side_use is not null)\n"
"signal (\'22023\', \'Can not re-create the RDF Quad Mapping <\' || qmid || \'> because it is referenced by <\' || old_side_use || \'>\');\n"
"DB.DBA.RDF_QM_DELETE_MAPPING_FROM_STORAGE (storage, NULL, qmid);\n"
"DB.DBA.RDF_QM_GC_SUBTREE (qmid);\n"
"}\n"
"if (qmparentid is not null)\n"
"DB.DBA.RDF_QM_ASSERT_JSO_TYPE (qmparentid, \'http://www.openlinksw.com/schemas/virtrdf#QuadMap\');\n"
"DB.DBA.RDF_QM_ASSERT_STORAGE_CONTAINS_MAPPING (storage, qmid, 0);\n"
"tablename := NULL;\n"
"DB.DBA.RDF_QM_NORMALIZE_QMV (qmv_g, qmvfix_g, qmvid_g, 0, \'graph\', tablename);\n"
"DB.DBA.RDF_QM_NORMALIZE_QMV (qmv_s, qmvfix_s, qmvid_s, 0, \'subject\', tablename);\n"
"DB.DBA.RDF_QM_NORMALIZE_QMV (qmv_p, qmvfix_p, qmvid_p, 0, \'predicate\', tablename);\n"
"DB.DBA.RDF_QM_NORMALIZE_QMV (qmv_o, qmvfix_o, qmvid_o, 1, \'object\', tablename);\n"
"if (\'\' = tablename)\n"
"tablename := NULL;\n"
"if (get_keyword_ucase (\'EXCLUSIVE\', opts))\n"
"qm_exclusive := \'http://www.openlinksw.com/schemas/virtrdf#SPART_QM_EXCLUSIVE\';\n"
"else\n"
"qm_exclusive := NULL;\n"
"if (get_keyword_ucase (\'OK_FOR_ANY_QUAD\', opts))\n"
"qm_is_default := \'http://www.openlinksw.com/schemas/virtrdf#SPART_QM_OK_FOR_ANY_QUAD\';\n"
"else\n"
"qm_is_default := NULL;\n"
"if (not is_real)\n"
"{\n"
"qm_empty := \'http://www.openlinksw.com/schemas/virtrdf#SPART_QM_EMPTY\';\n"
"if (tablename is null)\n"
"tablename := \'\';\n"
"}\n"
"else\n"
"{\n"
"qm_empty := NULL;\n"
"if (tablename is null)\n"
"signal (\'22023\', \'At least one field of a quad map should be map value, not a constant\');\n"
"}\n"
"qm_order := get_keyword_ucase (\'ORDER\', opts);\n"
"if (not is_real)\n"
"qmusersubmaps := qmid || \'--UserSubMaps\';\n"
"else\n"
"qmusersubmaps := NULL;\n"
"if (qm_is_default is not null)\n"
"{\n"
"if (qm_order is not null)\n"
"signal (\'22023\', \'ORDER option is not applicable to default quad map\');\n"
"if (qmparentid is not null)\n"
"signal (\'22023\', \'A default quad map can not be a sub-map of other quad map\');\n"
"}\n"
"for (sparql define input:storage \"\"\n"
"prefix rdfdf: <http://www.openlinksw.com/virtrdf-data-formats#>\n"
"insert in graph ?:graphiri {\n"
"`iri(?:qmid)`\n"
"rdf:type virtrdf:QuadMap ;\n"
"virtrdf:qmGraphRange-rvrFixedValue ?:qmvfix_g ;\n"
"virtrdf:qmGraphMap `iri(?:qmvid_g)` ;\n"
"virtrdf:qmSubjectRange-rvrFixedValue ?:qmvfix_s ;\n"
"virtrdf:qmSubjectMap `iri(?:qmvid_s)` ;\n"
"virtrdf:qmPredicateRange-rvrFixedValue ?:qmvfix_p ;\n"
"virtrdf:qmPredicateMap `iri(?:qmvid_p)` ;\n"
"virtrdf:qmObjectRange-rvrFixedValue ?:qmvfix_o ;\n"
"virtrdf:qmObjectMap `iri(?:qmvid_o)` ;\n"
"virtrdf:qmTableName ?:tablename ;\n"
"virtrdf:qmTableRowFilter ?:rowfilter ;\n"
"virtrdf:qmUserSubMaps `iri(?:qmusersubmaps)` ;\n"
"virtrdf:qmMatchingFlags `iri(?:qm_exclusive)` ;\n"
"virtrdf:qmMatchingFlags `iri(?:qm_empty)` ;\n"
"virtrdf:qmMatchingFlags `iri(?:qm_is_default)` ;\n"
"virtrdf:qmPriorityOrder ?:qm_order .\n"
"`iri(?:qmusersubmaps)`\n"
"rdf:type virtrdf:array-of-QuadMap .\n"
"}\n"
"where {} ) do {;}\n"
"if (qm_is_default is not null)\n"
"DB.DBA.RDF_QM_SET_DEFAULT_MAPPING (storage, qmid);\n"
"else\n"
"DB.DBA.RDF_QM_ADD_MAPPING_TO_STORAGE (storage, qmparentid, qmid, qm_order);\n"
"return vector (vector (\'00000\', \'Quad map <\' || qmid || \'> has been created and added to the <\' || storage || \'>\'));\n"
"}\n"
"--src sparql.sql:3722\n";

static const char *proc132 = 
"create function DB.DBA.RDF_QM_ATTACH_MAPPING (in storage varchar, in source varchar, in opts any) returns any\n"
"{\n"
"declare graphiri varchar;\n"
"declare qmid, qmgraph varchar;\n"
"declare qm_order, qm_is_default integer;\n"
"graphiri := DB.DBA.JSO_SYS_GRAPH ();\n"
"qmid := get_keyword_ucase (\'ID\', opts, NULL);\n"
"qmgraph := get_keyword_ucase (\'GRAPH\', opts, NULL);\n"
"DB.DBA.RDF_QM_ASSERT_STORAGE_FLAG (storage, 1);\n"
"DB.DBA.RDF_QM_ASSERT_STORAGE_FLAG (source, 0);\n"
"if (qmid is null)\n"
"{\n"
"qmid := coalesce ((sparql define input:storage \"\"\n"
"prefix virtrdf: <http://www.openlinksw.com/schemas/virtrdf#>\n"
"select ?s where {\n"
"graph ?:graphiri {\n"
"?s rdf:type virtrdf:QuadMap .\n"
"?s virtrdf:qmGraphRange-rvrFixedValue `iri(?:qmgraph)` .\n"
"?s virtrdf:qmMatchingFlags virtrdf:SPART_QM_EMPTY .\n"
"} } ));\n"
"if (qmid is null)\n"
"return vector (vector (\'00100\', \'Quad map for graph <\' || qmgraph || \'> is not found\'));\n"
"}\n"
"qm_order := coalesce ((sparql define input:storage \"\"\n"
"select ?o where { graph ?:graphiri {\n"
"`iri(?:qmid)` virtrdf:qmPriorityOrder ?o } } ) );\n"
"if (exists (sparql define input:storage \"\"\n"
"ask where { graph ?:graphiri {\n"
"`iri(?:qmid)` virtrdf:qmMatchingFlags virtrdf:SPART_QM_OK_FOR_ANY_QUAD } } ) )\n"
"qm_is_default := 1;\n"
"else\n"
"qm_is_default := 0;\n"
"DB.DBA.RDF_QM_ASSERT_STORAGE_CONTAINS_MAPPING (storage, qmid, 0);\n"
"DB.DBA.RDF_QM_ASSERT_STORAGE_CONTAINS_MAPPING (source, qmid, 1);\n"
"DB.DBA.RDF_QM_ASSERT_JSO_TYPE (qmid, \'http://www.openlinksw.com/schemas/virtrdf#QuadMap\');\n"
"if (qm_is_default)\n"
"DB.DBA.RDF_QM_SET_DEFAULT_MAPPING (storage, qmid);\n"
"else\n"
"DB.DBA.RDF_QM_ADD_MAPPING_TO_STORAGE (storage, NULL, qmid, NULL /* !!!TBD: place real value instead of constant NULL */);\n"
"return vector (vector (\'00000\', \'Quad map <\' || qmid || \'> is added to the storage <\' || storage || \'>\'));\n"
"}\n"
"--src sparql.sql:3830\n";

static const char *proc133 = 
"create procedure DB.DBA.RDF_QM_ADD_MAPPING_TO_STORAGE (in storage varchar, in qmparent varchar, in qmid varchar, in qmorder integer)\n"
"{\n"
"declare graphiri, lstiri varchar;\n"
"declare iris_and_orders any;\n"
"declare ctr, qmid_is_printed integer;\n"
"\n"
"graphiri := DB.DBA.JSO_SYS_GRAPH ();\n"
"if (qmparent is not null)\n"
"lstiri := (sparql define input:storage \"\"\n"
"select ?lst where { graph ?:graphiri {\n"
"`iri(?:qmparent)` virtrdf:qmUserSubMaps ?lst } } );\n"
"else\n"
"lstiri := (sparql define input:storage \"\"\n"
"select ?lst where { graph ?:graphiri {\n"
"`iri(?:storage)` virtrdf:qsUserMaps ?lst } } );\n"
"\n"
"if (qmorder is null)\n"
"qmorder := 1999;\n"
"iris_and_orders := (\n"
"select VECTOR_AGG (vector (sub.\"id\", sub.\"p\", sub.\"ord1\"))\n"
"from (\n"
"select sp.\"id\", sp.\"p\", sp.\"ord1\"\n"
"from (\n"
"sparql define input:storage \"\"\n"
"select ?id ?p\n"
"(bif:coalesce (?ord,\n"
"1000 + bif:aref (\n"
"bif:sprintf_inverse (\n"
"str(?p),\n"
"bif:concat (str (rdf:_), \"%d\"),\n"
"2),\n"
"0 ) ) ) as ?ord1\n"
"where { graph ?:graphiri {\n"
"`iri(?:lstiri)` ?p ?id .\n"
"filter (! bif:isnull (bif:aref (\n"
"bif:sprintf_inverse (\n"
"str(?p),\n"
"bif:concat (str (rdf:_), \"%d\"),\n"
"2),\n"
"0 ) ) ) .\n"
"optional {?id virtrdf:qmPriorityOrder ?ord} } } ) as sp\n"
"order by 2, 1 ) as sub );\n"
"\n"
"foreach (any itm in iris_and_orders) do\n"
"{\n"
"declare id, p varchar;\n"
"id := itm[0];\n"
"p := itm[1];\n"
"for (sparql define input:storage \"\"\n"
"delete from graph ?:graphiri {\n"
"`iri(?:lstiri)` `iri(?:p)` `iri(?:id)` . }\n"
"where { } ) do {;}\n"
"}\n"
"ctr := 1;\n"
"qmid_is_printed := 0;\n"
"foreach (any itm in iris_and_orders) do\n"
"{\n"
"declare id varchar;\n"
"declare ord integer;\n"
"id := itm[0];\n"
"ord := itm[2];\n"
"if (ord > qmorder)\n"
"{\n"
"for (sparql define input:storage \"\"\n"
"insert in graph ?:graphiri {\n"
"`iri(?:lstiri)`\n"
"`iri(bif:sprintf(\"%s%d\", str(rdf:_), ?:ctr))`\n"
"`iri(?:qmid)` . }\n"
"where { } ) do {;}\n"
"\n"
"ctr := ctr + 1;\n"
"qmid_is_printed := 1;\n"
"}\n"
"for (sparql define input:storage \"\"\n"
"insert in graph ?:graphiri {\n"
"`iri(?:lstiri)`\n"
"`iri(bif:sprintf(\"%s%d\", str(rdf:_), ?:ctr))`\n"
"`iri(?:id)` . }\n"
"where { } ) do {;}\n"
"ctr := ctr + 1;\n"
"}\n"
"if (not qmid_is_printed)\n"
"{\n"
"for (sparql define input:storage \"\"\n"
"insert in graph ?:graphiri {\n"
"`iri(?:lstiri)`\n"
"`iri(bif:sprintf(\"%s%d\", str(rdf:_), ?:ctr))`\n"
"`iri(?:qmid)` . }\n"
"where { } ) do {;}\n"
"\n"
"ctr := ctr + 1;\n"
"}\n"
"}\n"
"--src sparql.sql:3873\n";

static const char *proc134 = 
"create procedure DB.DBA.RDF_QM_DELETE_MAPPING_FROM_STORAGE (in storage varchar, in qmparent varchar, in qmid varchar)\n"
"{\n"
"declare graphiri, lstiri varchar;\n"
"declare iris_and_orders any;\n"
"declare ctr integer;\n"
"\n"
"graphiri := DB.DBA.JSO_SYS_GRAPH ();\n"
"if (qmparent is not null)\n"
"lstiri := (sparql define input:storage \"\"\n"
"select ?lst where { graph ?:graphiri {\n"
"`iri(?:qmparent)` virtrdf:qmUserSubMaps ?lst } } );\n"
"else\n"
"lstiri := (sparql define input:storage \"\"\n"
"select ?lst where { graph ?:graphiri {\n"
"`iri(?:storage)` virtrdf:qsUserMaps ?lst } } );\n"
"\n"
"iris_and_orders := (\n"
"select VECTOR_AGG (vector (sub.\"id\", sub.\"p\", sub.\"ord1\"))\n"
"from (\n"
"select sp.\"id\", sp.\"p\", sp.\"ord1\"\n"
"from (\n"
"sparql define input:storage \"\"\n"
"select ?id ?p\n"
"(bif:coalesce (?ord,\n"
"1000 + bif:aref (\n"
"bif:sprintf_inverse (\n"
"str(?p),\n"
"bif:concat (str (rdf:_), \"%d\"),\n"
"2),\n"
"0 ) ) ) as ?ord1\n"
"where { graph ?:graphiri {\n"
"`iri(?:lstiri)` ?p ?id .\n"
"filter (! bif:isnull (bif:aref (\n"
"bif:sprintf_inverse (\n"
"str(?p),\n"
"bif:concat (str (rdf:_), \"%d\"),\n"
"2),\n"
"0 ) ) ) .\n"
"optional {?id virtrdf:qmPriorityOrder ?ord} } } ) as sp\n"
"order by 2, 1 ) as sub );\n"
"\n"
"foreach (any itm in iris_and_orders) do\n"
"{\n"
"declare id, p varchar;\n"
"id := itm[0];\n"
"p := itm[1];\n"
"for (sparql define input:storage \"\"\n"
"delete from graph ?:graphiri {\n"
"`iri(?:lstiri)` `iri(?:p)` `iri(?:id)` . }\n"
"where { } ) do {;}\n"
"}\n"
"ctr := 1;\n"
"foreach (any itm in iris_and_orders) do\n"
"{\n"
"declare id varchar;\n"
"declare ord integer;\n"
"id := itm[0];\n"
"ord := itm[2];\n"
"if (id <> qmid)\n"
"{\n"
"for (sparql define input:storage \"\"\n"
"insert in graph ?:graphiri {\n"
"`iri(?:lstiri)`\n"
"`iri(bif:sprintf(\"%s%d\", str(rdf:_), ?:ctr))`\n"
"`iri(?:id)` . }\n"
"where { } ) do {;}\n"
"ctr := ctr + 1;\n"
"}\n"
"else\n"
"{\n"
"\n"
" ;\n"
"}\n"
"}\n"
"}\n"
"--src sparql.sql:3968\n";

static const char *proc135 = 
"create procedure DB.DBA.RDF_QM_SET_DEFAULT_MAPPING (in storage varchar, in qmid varchar)\n"
"{\n"
"declare graphiri, old_qmid varchar;\n"
"\n"
"graphiri := DB.DBA.JSO_SYS_GRAPH ();\n"
"old_qmid := coalesce ((sparql define input:storage \"\"\n"
"select ?qm where { graph ?:graphiri {\n"
"`iri(?:storage)` virtrdf:qsDefaultMap ?qm } } ) );\n"
"if (old_qmid is not null)\n"
"{\n"
"if (cast (old_qmid as varchar) = cast (qmid as varchar))\n"
"return;\n"
"signal (\'22023\', \'Quad map storage <\' || storage || \'> has set a default quad map <\' || old_qmid || \'>, drop it before adding <\' || qmid || \'>\');\n"
"}\n"
"for (sparql define input:storage \"\"\n"
"insert in graph ?:graphiri { `iri(?:storage)` virtrdf:qsDefaultMap `iri(?:qmid)` . }\n"
"where {} ) do {;}\n"
"}\n"
"--src sparql.sql:4045\n";

static const char *proc136 = 
"create function DB.DBA.RDF_QM_AUTO_LITERAL_FIELD (in qtable varchar, in alias varchar, in col varchar) returns any\n"
"{\n"
"return vector (UNAME\'literal\', vector (vector (coalesce (alias, \'\'), qtable)), vector (vector (qtable, alias, col)), vector ());\n"
"}\n"
"--src sparql.sql:4065\n";

static const char *proc137 = 
"create procedure\n"
"DB.DBA.SPARQL_GRABBER (in seed varchar, in iter varchar, in final varchar, in limit integer, in depth integer, in base_iri varchar, in resolver varchar)\n"
"{\n"
"declare grabbed any;\n"
"declare rctr, rcount, colcount, iter_ctr integer;\n"
"declare stat, msg varchar;\n"
"declare params any;\n"
"declare metas, rset any;\n"
"result_names (rset);\n"
"params := vector();\n"
"grabbed := dict_new ();\n"
"for (iter_ctr := 0; iter_ctr <= depth; iter_ctr := iter_ctr + 1)\n"
"{\n"
"declare new_found integer;\n"
"new_found := 0;\n"
"stat := \'00000\';\n"
"exec (case (iter_ctr) when 0 then seed else iter end, stat, msg, params, limit, metas, rset);\n"
"if (stat <> \'00000\')\n"
"signal (stat, msg);\n"
"rcount := length (rset);\n"
"colcount := length (metas[0]);\n"
"for (rctr := 0; rctr < rcount; rctr := rctr + 1)\n"
"{\n"
"declare colctr integer;\n"
"for (colctr := 0; colctr < colcount; colctr := colctr + 1)\n"
"{\n"
"declare val any;\n"
"val := rset[rctr][colctr];\n"
"if (isiri_id (val) and (val < #i1000000000))\n"
"{\n"
"declare url, get_method varchar;\n"
"call (resolver) (base_iri, val, url, get_method);\n"
"if (url is not null and not dict_get (grabbed, url, 0))\n"
"{\n"
"whenever sqlstate \'*\' goto end_of_sponge;\n"
"DB.DBA.RDF_SPONGE_UP (url, vector (UNAME\'get:soft\', \'YES\', UNAME\'get:method\', get_method));\n"
"new_found := 1;\n"
"end_of_sponge:\n"
"dict_put (grabbed, url, 1);\n"
"}\n"
"}\n"
"}\n"
"}\n"
"if (not new_found)\n"
"goto final_exec;\n"
"}\n"
"\n"
"final_exec:\n"
"stat := \'00000\';\n"
"exec (final, stat, msg, params, limit, metas, rset);\n"
"if (stat <> \'00000\')\n"
"signal (stat, msg);\n"
"rcount := length (rset);\n"
"for (rctr := 0; rctr < rcount; rctr := rctr + 1)\n"
"result (rset[rctr]);\n"
"}\n"
"--src sparql.sql:4075\n";

static const char *proc138 = 
"create function DB.DBA.SPARQL_GRABBER_DEFAULT_RESOLVER (in base varchar, in rel_iid IRI_ID, out abs_uri varchar, out get_method varchar)\n"
"{\n"
"declare rel varchar;\n"
"rel := DB.DBA.RDF_QNAME_OF_IID (rel_iid);\n"
"if (base = \'\')\n"
"abs_uri := rel;\n"
"else\n"
"abs_uri := XML_URI_RESOLVE_LIKE_GET (base, rel);\n"
"if ((abs_uri like \'%/\') or (abs_uri like \'%#%\'))\n"
"get_method := \'MGET\';\n"
"else\n"
"get_method := \'GET\';\n"
"}\n"
"--src sparql.sql:4133\n";

static const char *proc139 = 
"create procedure DB.DBA.SPARQL_EVAL_TO_ARRAY (in query varchar, in dflt_graph varchar, in maxrows integer)\n"
"{\n"
"declare sqltext, state, msg varchar;\n"
"declare metas, rset any;\n"
"sparql_explain (query);\n"
"sqltext := string_output_string (sparql_to_sql_text (query));\n"
"state := \'00000\';\n"
"metas := null;\n"
"rset := null;\n"
"connection_set (\':default_graph\', dflt_graph);\n"
"exec (sqltext, state, msg, vector(), maxrows, metas, rset);\n"
"\n"
"if (state <> \'00000\')\n"
"signal (state, msg);\n"
"return rset;\n"
"}\n"
"--src sparql.sql:4152\n";

static const char *proc140 = 
"create procedure DB.DBA.SPARQL_EVAL (in query varchar, in dflt_graph varchar, in maxrows integer)\n"
"{\n"
"declare sqltext, state, msg varchar;\n"
"declare metas, rset any;\n"
"sparql_explain (query);\n"
"sqltext := string_output_string (sparql_to_sql_text (query));\n"
"state := \'00000\';\n"
"metas := null;\n"
"rset := null;\n"
"\n"
"\n"
"\n"
"state := \'00000\';\n"
"metas := null;\n"
"rset := null;\n"
"connection_set (\':default_graph\', dflt_graph);\n"
"exec (sqltext, state, msg, vector(), maxrows, metas, rset);\n"
"if (state <> \'00000\')\n"
"signal (state, msg);\n"
"\n"
"exec_result_names (metas[0]);\n"
"foreach (any row in rset) do\n"
"{\n"
"exec_result (row);\n"
"}\n"
"}\n"
"--src sparql.sql:4170\n";

static const char *proc141 = 
"create procedure DB.DBA.SPARQL_REXEC_INT (\n"
"in res_mode integer,\n"
"in service varchar,\n"
"in query varchar,\n"
"in dflt_graph varchar,\n"
"inout named_graphs any,\n"
"inout req_hdr any,\n"
"in maxrows integer,\n"
"inout metas any,\n"
"inout bnode_dict any\n"
")\n"
"{\n"
"declare req_uri, req_method, req_body, local_req_hdr, ret_body, ret_hdr any;\n"
"declare ret_content_type varchar;\n"
"req_body := string_output();\n"
"http (\'query=\', req_body);\n"
"http_url (query, 0, req_body);\n"
"if (dflt_graph is not null and dflt_graph <> \'\')\n"
"{\n"
"http (\'&default-graph-uri=\', req_body);\n"
"http_url (dflt_graph, 0, req_body);\n"
"}\n"
"foreach (varchar uri in named_graphs) do\n"
"{\n"
"http (\'&named-graph-uri=\', req_body);\n"
"http_url (uri, 0, req_body);\n"
"}\n"
"req_body := string_output_string (req_body);\n"
"local_req_hdr := \'Accept: application/sparql-results+xml, text/rdf+n3, application/rdf+xml, application/xml\';\n"
"if (length (req_body) + length (service) >= 1900)\n"
"{\n"
"req_method := \'POST\';\n"
"req_uri := service;\n"
"local_req_hdr := local_req_hdr || \'\\r\\nContent-Type: application/x-www-form-urlencoded\';\n"
"}\n"
"else\n"
"{\n"
"req_method := \'GET\';\n"
"req_uri := service || \'?\' || req_body;\n"
"req_body := \'\';\n"
"}\n"
"if (length (req_hdr) > 0)\n"
"req_hdr := concat (req_hdr, \'\\r\\n\', local_req_hdr );\n"
"else\n"
"req_hdr := local_req_hdr;\n"
"\n"
"\n"
"\n"
"ret_body := http_get (req_uri, ret_hdr, req_method, req_hdr, req_body);\n"
"\n"
"\n"
"ret_content_type := http_request_header (ret_hdr, \'Content-Type\', null, null);\n"
"if (ret_content_type is null or\n"
"(strstr (ret_content_type, \'application/sparql-results+xml\') is null and\n"
"strstr (ret_content_type, \'application/rdf+xml\') is null and\n"
"strstr (ret_content_type, \'text/rdf+n3\') is null ) )\n"
"{\n"
"declare ret_begin, ret_html any;\n"
"ret_begin := \"LEFT\" (ret_body, 1024);\n"
"ret_html := xtree_doc (ret_begin, 2);\n"
"if (xpath_eval (\'/html|/xhtml\', ret_html) is not null)\n"
"ret_content_type := \'text/html\';\n"
"else if (xpath_eval (\'[xmlns:rset=\"http://www.w3.org/2005/sparql-results#\"] /rset:sparql\', ret_html) is not null\n"
"or xpath_eval (\'[xmlns:rset2=\"http://www.w3.org/2001/sw/DataAccess/rf1/result2\"] /rset2:sparql\', ret_html) is not null)\n"
"ret_content_type := \'application/sparql-results+xml\';\n"
"else if (xpath_eval (\'[xmlns:rdf=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\"] /rdf:rdf\', ret_html) is not null)\n"
"ret_content_type := \'application/rdf+xml\';\n"
"else if (strstr (ret_begin, \'<html>\') is not null or\n"
"strstr (ret_begin, \'<xhtml>\') is not null )\n"
"ret_content_type := \'text/html\';\n"
"else\n"
"{\n"
"ret_content_type := \'text/plain\';\n"
"}\n"
"}\n"
"if (strstr (ret_content_type, \'application/sparql-results+xml\') is not null)\n"
"{\n"
"declare ret_xml, var_list, var_metas, ret_row, out_nulls any;\n"
"declare var_ctr, var_count integer;\n"
"declare vect_acc any;\n"
"declare row_inx integer;\n"
"\n"
"ret_xml := xtree_doc (ret_body, 0);\n"
"var_list := xpath_eval (\'[xmlns:rset=\"http://www.w3.org/2005/sparql-results#\"] [xmlns:rset2=\"http://www.w3.org/2001/sw/DataAccess/rf1/result2\"]\n"
"/rset:sparql/rset:head/rset:variable | /rset2:sparql/rset2:head/rset2:variable\', ret_xml, 0);\n"
"if (0 = length (var_list))\n"
"{\n"
"declare bool_ret any;\n"
"bool_ret := xpath_eval (\'[xmlns:rset=\"http://www.w3.org/2005/sparql-results#\"] [xmlns:rset2=\"http://www.w3.org/2001/sw/DataAccess/rf1/result2\"]\n"
"/rset:sparql/rset:boolean | /rset2:sparql/rset2:boolean\', ret_xml);\n"
"if (bool_ret is not null)\n"
"{\n"
"bool_ret := cast (bool_ret as varchar);\n"
"if (\'true\' = bool_ret)\n"
"bool_ret := 1;\n"
"else if (\'false\' = bool_ret)\n"
"bool_ret := 0;\n"
"else\n"
"signal (\'RDFZZ\', sprintf (\n"
"\'DB.DBA.SPARQL_REXEC(\'\'%.300s\'\', ...) has received invalid boolean value \'\'%.300s\'\'\',\n"
"service, bool_ret ) );\n"
"metas :=\n"
"vector (\n"
"vector (\n"
"vector (\'__ask_retval\', 242, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0) ),\n"
"1 );\n"
"if (0 = res_mode)\n"
"{\n"
"declare __ask_retval integer;\n"
"result_names (__ask_retval);\n"
"result (bool_ret);\n"
"}\n"
"else if (1 = res_mode)\n"
"return vector (vector (bool_ret));\n"
"return;\n"
"}\n"
"signal (\'RDFZZ\', sprintf (\n"
"\'DB.DBA.SPARQL_REXEC(\'\'%.300s\'\', ...) has received result with no variables\',\n"
"service ) );\n"
"}\n"
"var_count := length (var_list);\n"
"var_metas := make_array (var_count, \'any\');\n"
"out_nulls := make_array (var_count, \'any\');\n"
"for (var_ctr := var_count - 1; var_ctr >= 0; var_ctr := var_ctr - 1)\n"
"{\n"
"declare var_name varchar;\n"
"var_name := cast (xpath_eval (\'@name\', var_list[var_ctr]) as varchar);\n"
"var_list [var_ctr] := var_name;\n"
"var_metas [var_ctr] := vector (var_name, 242, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0);\n"
"out_nulls [var_ctr] := null;\n"
"}\n"
"\n"
"if (0 = res_mode)\n"
"exec_result_names (var_metas);\n"
"else if (1 = res_mode)\n"
"vectorbld_init (vect_acc);\n"
"row_inx := 0;\n"
"for (ret_row := xpath_eval (\'[xmlns:rset=\"http://www.w3.org/2005/sparql-results#\"] [xmlns:rset2=\"http://www.w3.org/2001/sw/DataAccess/rf1/result2\"]\n"
"/rset:sparql/rset:results/rset:result | /rset2:sparql/rset2:results/rset2:result\', ret_xml);\n"
"ret_row is not null;\n"
"ret_row := xpath_eval (\'[xmlns:rset=\"http://www.w3.org/2005/sparql-results#\"] [xmlns:rset2=\"http://www.w3.org/2001/sw/DataAccess/rf1/result2\"]\n"
"following-sibling::rset:result | following-sibling::rset2:result\', ret_row) )\n"
"{\n"
"declare out_fields, ret_cols any;\n"
"declare col_ctr, col_count integer;\n"
"out_fields := out_nulls;\n"
"ret_cols := xpath_eval (\'[xmlns:rset=\"http://www.w3.org/2005/sparql-results#\"] [xmlns:rset2=\"http://www.w3.org/2001/sw/DataAccess/rf1/result2\"]\n"
"rset:binding | rset2:binding\', ret_row, 0);\n"
"col_count := length (ret_cols);\n"
"for (col_ctr := col_count - 1; col_ctr >= 0; col_ctr := col_ctr - 1)\n"
"{\n"
"declare ret_col any;\n"
"declare var_name, var_type, var_strval varchar;\n"
"declare var_pos integer;\n"
"ret_col := ret_cols[col_ctr];\n"
"var_name := cast (xpath_eval (\'string(@name)\', ret_col) as varchar);\n"
"\n"
"var_pos := position (var_name, var_list) - 1;\n"
"if (var_pos >= 0)\n"
"{\n"
"var_type := cast (xpath_eval (\'local-name(*)\', ret_col) as varchar);\n"
"var_strval := charset_recode (xpath_eval (\'string(*)\', ret_col), \'_WIDE_\', \'UTF-8\');\n"
"\n"
"if (\'uri\' = var_type)\n"
"out_fields [var_pos] := DB.DBA.RDF_MAKE_IID_OF_QNAME (var_strval);\n"
"else if (\'bnode\' = var_type)\n"
"{\n"
"declare local_iid IRI_ID;\n"
"if (bnode_dict is null)\n"
"{\n"
"bnode_dict := dict_new ();\n"
"local_iid := iri_id_from_num (sequence_next (\'RDF_URL_IID_BLANK\'));\n"
"dict_put (bnode_dict, var_strval, local_iid);\n"
"}\n"
"else\n"
"{\n"
"local_iid := dict_get (bnode_dict, var_strval, null);\n"
"if (local_iid is null)\n"
"{\n"
"local_iid := iri_id_from_num (sequence_next (\'RDF_URL_IID_BLANK\'));\n"
"dict_put (bnode_dict, var_strval, local_iid);\n"
"}\n"
"}\n"
"out_fields [var_pos] := local_iid;\n"
"}\n"
"else if (\'literal\' = var_type)\n"
"{\n"
"declare lang, dt varchar;\n"
"lang := charset_recode (xpath_eval (\'*/@xml:lang\', ret_col), \'_WIDE_\', \'UTF-8\');\n"
"dt := charset_recode (xpath_eval (\'*/datatype\', ret_col), \'_WIDE_\', \'UTF-8\');\n"
"out_fields [var_pos] := DB.DBA.RDF_MAKE_LONG_OF_TYPEDSQLVAL_STRINGS (\n"
"var_strval, dt, lang );\n"
"}\n"
"else\n"
"signal (\'RDFZZ\', sprintf (\n"
"\'DB.DBA.SPARQL_REXEC(\'\'%.300s\'\', ...) contains unsupported type of bound value \'\'%.300s\'\'\',\n"
"service, var_type ) );\n"
"}\n"
"}\n"
"if (0 = res_mode)\n"
"exec_result (out_fields);\n"
"else if (1 = res_mode)\n"
"vectorbld_acc (vect_acc, out_fields);\n"
"row_inx := row_inx + 1;\n"
"if (maxrows is not null and maxrows > 0 and row_inx >= maxrows)\n"
"ret_row := xpath_eval (\'[xmlns:rset=\"http://www.w3.org/2005/sparql-results#\"] [xmlns:rset2=\"http://www.w3.org/2001/sw/DataAccess/rf1/result2\"]\n"
"../rset:result[position() = last()] | ../rset2:result[position() = last()]\', ret_row);\n"
"}\n"
"metas := vector (var_metas, 1);\n"
"if (0 = res_mode)\n"
"{\n"
"return;\n"
"}\n"
"else if (1 = res_mode)\n"
"{\n"
"vectorbld_final (vect_acc);\n"
"return vect_acc;\n"
"}\n"
"}\n"
"if (strstr (ret_content_type, \'application/rdf+xml\') is not null)\n"
"{\n"
"declare res_dict any;\n"
"res_dict := DB.DBA.RDF_RDFXML_TO_DICT (ret_body,\'http://local.virt/tmp\',\'\');\n"
"metas := vector (vector (vector (\'res_dict\', 242, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0)), 1);\n"
"if (0 = res_mode)\n"
"{\n"
"result_names (res_dict);\n"
"result (res_dict);\n"
"return;\n"
"}\n"
"else if (1 = res_mode)\n"
"return vector (vector (res_dict));\n"
"}\n"
"if (strstr (ret_content_type, \'text/rdf+n3\') is not null)\n"
"{\n"
"declare res_dict any;\n"
"res_dict := DB.DBA.RDF_TTL2HASH (ret_body, \'\');\n"
"metas := vector (vector (vector (\'res_dict\', 242, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0)), 1);\n"
"if (0 = res_mode)\n"
"{\n"
"result_names (res_dict);\n"
"result (res_dict);\n"
"return;\n"
"}\n"
"else if (1 = res_mode)\n"
"return vector (vector (res_dict));\n"
"}\n"
"if (strstr (ret_content_type, \'text/plain\') is not null)\n"
"{\n"
"signal (\'RDFZZ\', sprintf (\n"
"\'DB.DBA.SPARQL_REXEC(\'\'%.300s\'\', ...) returned Content-Type \'\'%.300s\'\' status \'\'%.300s\'\'\\n%.1000s\',\n"
"service, ret_content_type, ret_hdr[0], \"LEFT\" (ret_body, 1024) ) );\n"
"}\n"
"if (strstr (ret_content_type, \'text/html\') is not null)\n"
"{\n"
"signal (\'RDFZZ\', sprintf (\n"
"\'DB.DBA.SPARQL_REXEC(\'\'%.300s\'\', ...) returned Content-Type \'\'%.300s\'\' status \'\'%.300s\'\'\\n%.1000s\',\n"
"service, ret_content_type, ret_hdr[0],\n"
"cast (xtree_doc (ret_body, 2) as varchar) ) );\n"
"}\n"
"signal (\'RDFZZ\', sprintf (\n"
"\'DB.DBA.SPARQL_REXEC(\'\'%.300s\'\', ...) returned unsupported Content-Type \'\'%.300s\'\'\',\n"
"service, ret_content_type ) );\n"
"}\n"
"--src sparql.sql:4202\n";

static const char *proc142 = 
"create procedure DB.DBA.SPARQL_REXEC (\n"
"in service varchar,\n"
"in query varchar,\n"
"in dflt_graph varchar,\n"
"in named_graphs any,\n"
"in req_hdr any,\n"
"in maxrows integer,\n"
"in bnode_dict any\n"
")\n"
"{\n"
"declare metas any;\n"
"DB.DBA.SPARQL_REXEC_INT (0, service, query, dflt_graph, named_graphs, req_hdr, maxrows, metas, bnode_dict);\n"
"}\n"
"--src sparql.sql:4468\n";

static const char *proc143 = 
"create function DB.DBA.SPARQL_REXEC_TO_ARRAY (\n"
"in service varchar,\n"
"in query varchar,\n"
"in dflt_graph varchar,\n"
"in named_graphs any,\n"
"in req_hdr any,\n"
"in maxrows integer,\n"
"in bnode_dict any\n"
") returns any\n"
"{\n"
"declare metas any;\n"
"return DB.DBA.SPARQL_REXEC_INT (1, service, query, dflt_graph, named_graphs, req_hdr, maxrows, metas, bnode_dict);\n"
"}\n"
"--src sparql.sql:4483\n";

static const char *proc144 = 
"create procedure DB.DBA.SPARQL_REXEC_WITH_META (\n"
"in service varchar,\n"
"in query varchar,\n"
"in dflt_graph varchar,\n"
"in named_graphs any,\n"
"in req_hdr any,\n"
"in maxrows integer,\n"
"in bnode_dict any,\n"
"out metadata any,\n"
"out resultset any\n"
")\n"
"{\n"
"resultset := DB.DBA.SPARQL_REXEC_INT (1, service, query, dflt_graph, named_graphs, req_hdr, maxrows, metadata, bnode_dict);\n"
"}\n"
"--src sparql.sql:4498\n";

static const char *proc145 = 
"create procedure \"querySoap\"  (in  \"Command\" varchar\n"
", in  \"Properties\" any\n"
", out \"Error\" any __soap_fault \'__XML__\'\n"
", out \"ws_sparql_xsd\" any\n"
")\n"
"__soap_options ( __soap_type:=\'__ANY__\',\n"
"\"soapAction\":=\'urn:FIXME:querySoap\',\n"
"\"RequestNamespace\":=\'urn:http://www.w3.org/2005/08/sparql-protocol-query/#\',\n"
"\"ResponseNamespace\":=\'urn:http://www.w3.org/2005/08/sparql-protocol-query/#\',\n"
"\"PartName\":=\'return\'\n"
")\n"
"{\n"
"declare stmt, state, msg, mdta, dta, res, ses any;\n"
"\n"
"stmt := get_keyword (\'Statement\', \"Command\");\n"
"ses := string_output ();\n"
"\n"
"\n"
"res := exec (stmt, state, msg, vector (), 0, mdta, dta);\n"
"\n"
"SPARQL_RESULTS_XML_WRITE_NS (ses);\n"
"SPARQL_RESULTS_XML_WRITE_HEAD (ses, mdta);\n"
"SPARQL_RESULTS_XML_WRITE_RES (ses, mdta, dta);\n"
"\n"
"\n"
"http (\'</sparql>\', ses);\n"
"\n"
"ses := string_output_string (ses);\n"
"string_to_file (\'out.xml\', ses, -2);\n"
"res := xml_tree_doc (ses);\n"
"return res;\n"
"\n"
"}\n"
"--src sparql.sql:4518\n";

static const char *proc146 = 
"create procedure SPARQL_RESULTS_XML_WRITE_NS (inout ses any)\n"
"{\n"
"http (\'<sparql xmlns=\"http://www.w3.org/2005/sparql-results#\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xsi:schemaLocation=\"http://www.w3.org/2001/sw/DataAccess/rf1/result2.xsd\">\', ses);\n"
"}\n"
"--src sparql.sql:4553\n";

static const char *proc147 = 
"create procedure SPARQL_RESULTS_XML_WRITE_HEAD (inout ses any, in mdta any)\n"
"{\n"
"declare i, l integer;\n"
"\n"
"http (\'\\n <head>\', ses);\n"
"\n"
"mdta := mdta[0];\n"
"i := 0; l := length (mdta);\n"
"while (i < l)\n"
"{\n"
"declare _name varchar;\n"
"declare _type, _type_name, nill int;\n"
"_name := mdta[i][0];\n"
"_type := mdta[i][1];\n"
"if (length (mdta[i]) > 4)\n"
"nill := mdta[i][4];\n"
"else\n"
"nill := 0;\n"
"http (sprintf (\'\\n  <variable name=\"%s\"/>\', _name), ses);\n"
"i := i + 1;\n"
"}\n"
"\n"
"http (\'\\n </head>\', ses);\n"
"}\n"
"--src sparql.sql:4559\n";

static const char *proc148 = 
"create procedure SPARQL_RESULTS_XML_WRITE_RES (inout ses any, in mdta any, inout dta any)\n"
"{\n"
"http (\'\\n <results distinct=\"false\" ordered=\"true\">\', ses);\n"
"\n"
"for (declare ctr integer, ctr := 0; ctr < length (dta); ctr := ctr + 1)\n"
"SPARQL_RESULTS_XML_WRITE_ROW (ses, mdta, dta[ctr]);\n"
"\n"
"http (\'\\n </results>\', ses);\n"
"}\n"
"--src sparql.sql:4585\n";

static const char *proc149 = 
"create procedure SPARQL_RESULTS_XML_WRITE_ROW (inout ses any, in mdta any, inout dta any)\n"
"{\n"
"\n"
"http (\'\\n  <result>\', ses);\n"
"mdta := mdta[0];\n"
"\n"
"for (declare x any, x := 0; x < length (mdta); x := x + 1)\n"
"{\n"
"declare _name varchar;\n"
"declare _val any;\n"
"_name := mdta[x][0];\n"
"_val := dta[x];\n"
"if (_val is null)\n"
"goto end_of_binding;\n"
"if (isiri_id (_val))\n"
"{\n"
"if (_val >= #i1000000000)\n"
"{\n"
"http (sprintf (\'\\n   <binding name=\"%s\"><bnode>nodeID://%d</bnode></binding>\', _name, iri_id_num (_val)), ses);\n"
"}\n"
"else\n"
"{\n"
"declare res varchar;\n"
"res := id_to_iri (_val);\n"
"\n"
"if (res is null)\n"
"res := sprintf (\'bad://%d\', iri_id_num (_val));\n"
"http (sprintf (\'\\n   <binding name=\"%s\"><uri>\', _name), ses);\n"
"http_value (res, 0, ses);\n"
"http (\'</uri></binding>\', ses);\n"
"}\n"
"}\n"
"else\n"
"{\n"
"declare lang, dt varchar;\n"
"if (__tag (_val) = 185)\n"
"{\n"
"http (sprintf (\'\\n   <binding name=\"%s\"><literal>\', _name), ses);\n"
"http_value (_val, 0, ses);\n"
"http (\'</literal></binding>\', ses);\n"
"goto end_of_binding;\n"
"}\n"
"lang := DB.DBA.RDF_LANGUAGE_OF_LONG (_val);\n"
"dt := DB.DBA.RDF_DATATYPE_OF_LONG (_val);\n"
"if (lang is not null)\n"
"{\n"
"if (dt is not null)\n"
"http (sprintf (\'\\n   <binding name=\"%s\"><literal xml:lang=\"%V\" datatype=\"%V\">\',\n"
"_name, cast (lang as varchar), cast (dt as varchar)), ses);\n"
"else\n"
"http (sprintf (\'\\n   <binding name=\"%s\"><literal xml:lang=\"%V\">\',\n"
"_name, cast (lang as varchar)), ses);\n"
"}\n"
"else\n"
"{\n"
"if (dt is not null)\n"
"http (sprintf (\'\\n   <binding name=\"%s\"><literal datatype=\"%V\">\',\n"
"_name, cast (dt as varchar)), ses);\n"
"else\n"
"http (sprintf (\'\\n   <binding name=\"%s\"><literal>\',\n"
"_name), ses);\n"
"}\n"
"http_value (DB.DBA.RDF_SQLVAL_OF_LONG (_val), 0, ses);\n"
"http (\'</literal></binding>\', ses);\n"
"}\n"
"end_of_binding: ;\n"
"}\n"
"\n"
"http (\'\\n  </result>\', ses);\n"
"}\n"
"--src sparql.sql:4596\n";

static const char *proc150 = 
"create procedure SPARQL_RESULTS_RDFXML_WRITE_NS (inout ses any)\n"
"{\n"
"http (\'<rdf:RDF xmlns:res=\"http://www.w3.org/2005/sparql-results#\" xmlns:rdf=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\">\n"
"<rdf:Description rdf:nodeID=\"rset\">\n"
"<rdf:type rdf:resource=\"http://www.w3.org/2005/sparql-results#ResultSet\" />\', ses);\n"
"}\n"
"--src sparql.sql:4668\n";

static const char *proc151 = 
"create procedure SPARQL_RESULTS_RDFXML_WRITE_HEAD (inout ses any, in mdta any)\n"
"{\n"
"declare i, l integer;\n"
"mdta := mdta[0];\n"
"i := 0; l := length (mdta);\n"
"while (i < l)\n"
"{\n"
"declare _name varchar;\n"
"declare _type, _type_name, nill int;\n"
"_name := mdta[i][0];\n"
"_type := mdta[i][1];\n"
"if (length (mdta[i]) > 4)\n"
"nill := mdta[i][4];\n"
"else\n"
"nill := 0;\n"
"http (sprintf (\'\\n    <res:resultVariable>%V</res:resultVariable>\', _name), ses);\n"
"i := i + 1;\n"
"}\n"
"}\n"
"--src sparql.sql:4676\n";

static const char *proc152 = 
"create procedure SPARQL_RESULTS_RDFXML_WRITE_RES (inout ses any, in mdta any, inout dta any)\n"
"{\n"
"for (declare ctr integer, ctr := 0; ctr < length (dta); ctr := ctr + 1)\n"
"{\n"
"http ( sprintf (\'\\n    <res:solution rdf:nodeID=\"r%d\">\', ctr), ses);\n"
"SPARQL_RESULTS_RDFXML_WRITE_ROW (ses, mdta, dta, ctr);\n"
"http (\'\\n    </res:solution>\', ses);\n"
"}\n"
"}\n"
"--src sparql.sql:4697\n";

static const char *proc153 = 
"create procedure SPARQL_RESULTS_RDFXML_WRITE_ROW (inout ses any, in mdta any, inout dta any, in rowno integer)\n"
"{\n"
"mdta := mdta[0];\n"
"for (declare x any, x := 0; x < length (mdta); x := x + 1)\n"
"{\n"
"declare _name varchar;\n"
"declare _val any;\n"
"_name := mdta[x][0];\n"
"_val := dta[rowno][x];\n"
"if (_val is null)\n"
"goto end_of_binding;\n"
"http (sprintf (\'\\n      <res:binding rdf:nodeID=\"r%dc%d\"><res:variable>%V</res:variable><res:value\', rowno, x, _name), ses);\n"
"if (isiri_id (_val))\n"
"{\n"
"if (_val >= #i1000000000)\n"
"{\n"
"http (sprintf (\' rdf:nodeID=\"nodeID://%d\"/></res:binding>\', iri_id_num (_val)), ses);\n"
"}\n"
"else\n"
"{\n"
"declare res varchar;\n"
"res := id_to_iri (_val);\n"
"\n"
"if (res is null)\n"
"res := sprintf (\'bad://%d\', iri_id_num (_val));\n"
"http (sprintf (\' rdf:resource=\"%V\"/></res:binding>\', res), ses);\n"
"}\n"
"}\n"
"else\n"
"{\n"
"declare lang, dt varchar;\n"
"if (__tag (_val) = 185)\n"
"{\n"
"http (\'>\', ses);\n"
"http_value (_val, 0, ses);\n"
"http (\'</res:value></res:binding>\', ses);\n"
"goto end_of_binding;\n"
"}\n"
"lang := DB.DBA.RDF_LANGUAGE_OF_LONG (_val);\n"
"dt := DB.DBA.RDF_DATATYPE_OF_LONG (_val);\n"
"if (lang is not null)\n"
"{\n"
"if (dt is not null)\n"
"http (sprintf (\' xml:lang=\"%V\" datatype=\"%V\">\', cast (lang as varchar), cast (dt as varchar)), ses);\n"
"else\n"
"http (sprintf (\' xml:lang=\"%V\">\', cast (lang as varchar)), ses);\n"
"}\n"
"else\n"
"{\n"
"if (dt is not null)\n"
"http (sprintf (\' datatype=\"%V\">\', cast (dt as varchar)), ses);\n"
"else\n"
"http (\'>\', ses);\n"
"}\n"
"http_value (DB.DBA.RDF_SQLVAL_OF_LONG (_val), 0, ses);\n"
"http (\'</res:value></res:binding>\', ses);\n"
"}\n"
"end_of_binding: ;\n"
"}\n"
"}\n"
"--src sparql.sql:4708\n";

static const char *proc154 = 
"create procedure SPARQL_RESULTS_TTL_WRITE_NS (inout ses any)\n"
"{\n"
"http (\'@prefix res: <http://www.w3.org/2005/sparql-results#> .\n"
"@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .\n"
"[] rdf:type res:ResultSet ;\', ses);\n"
"}\n"
"--src sparql.sql:4770\n";

static const char *proc155 = 
"create procedure SPARQL_RESULTS_TTL_WRITE_HEAD (inout ses any, in mdta any)\n"
"{\n"
"declare i, l integer;\n"
"mdta := mdta[0];\n"
"i := 0; l := length (mdta);\n"
"while (i < l)\n"
"{\n"
"declare _name varchar;\n"
"declare _type, _type_name, nill int;\n"
"_name := mdta[i][0];\n"
"_type := mdta[i][1];\n"
"if (length (mdta[i]) > 4)\n"
"nill := mdta[i][4];\n"
"else\n"
"nill := 0;\n"
"http (\'\\n  res:resultVariable \"\', ses);\n"
"http_escape (_name, 11, ses, 0, 1);\n"
"http (\'\"\', ses);\n"
"i := i + 1;\n"
"if (i < l)\n"
"http (\' ;\', ses);\n"
"}\n"
"}\n"
"--src sparql.sql:4778\n";

static const char *proc156 = 
"create procedure SPARQL_RESULTS_TTL_WRITE_RES (inout ses any, in mdta any, inout dta any)\n"
"{\n"
"declare ctr, len integer;\n"
"ctr := 0; len := length (dta);\n"
"while (ctr < len)\n"
"{\n"
"http (\'\\n  res:solution [\', ses);\n"
"SPARQL_RESULTS_TTL_WRITE_ROW (ses, mdta, dta, ctr);\n"
"http (\' ]\', ses);\n"
"ctr := ctr + 1;\n"
"if (ctr < len)\n"
"http (\' ;\', ses);\n"
"}\n"
"}\n"
"--src sparql.sql:4803\n";

static const char *proc157 = 
"create procedure SPARQL_RESULTS_TTL_WRITE_ROW (inout ses any, in mdta any, inout dta any, in rowno integer)\n"
"{\n"
"declare need_semicolon integer;\n"
"mdta := mdta[0];\n"
"need_semicolon := 0;\n"
"for (declare x any, x := 0; x < length (mdta); x := x + 1)\n"
"{\n"
"declare _name varchar;\n"
"declare _val any;\n"
"_name := mdta[x][0];\n"
"_val := dta[rowno][x];\n"
"if (_val is not null)\n"
"{\n"
"if (need_semicolon)\n"
"http (\' ;\', ses);\n"
"else\n"
"need_semicolon := 1;\n"
"http (\'\\n      res:binding [ res:variable \"\', ses);\n"
"http_escape (_name, 11, ses, 0, 1);\n"
"http (\'\" ; res:value \', ses);\n"
"DB.DBA.RDF_LONG_TO_TTL (_val, ses);\n"
"http (\' ]\', ses);\n"
"}\n"
"}\n"
"}\n"
"--src sparql.sql:4819\n";

static const char *proc158 = 
"create procedure SPARQL_RESULTS_JAVASCRIPT_HTML_WRITE (inout ses any, inout metas any, inout rset any, in is_js integer := 0)\n"
"{\n"
"declare varctr, varcount, resctr, rescount integer;\n"
"varcount := length (metas[0]);\n"
"rescount := length (rset);\n"
"if (is_js)\n"
"{\n"
"declare tmp_str varchar;\n"
"declare tmp_ses any;\n"
"tmp_ses := string_output();\n"
"http (\'document.writeln(\'\'\', tmp_ses);\n"
"SPARQL_RESULTS_JAVASCRIPT_HTML_WRITE(tmp_ses,metas,rset,0);\n"
"tmp_str := string_output_string(tmp_ses);\n"
"tmp_str := replace(tmp_str, \'\\n\', \'\'\');\\ndocument.writeln(\'\'\');\n"
"http (tmp_str, ses);\n"
"http (\'\'\');\', ses);\n"
"return;\n"
"}\n"
"http (\'<table class=\"sparql\" border=\"1\">\', ses);\n"
"http (\'\\n  <tr>\', ses);\n"
"\n"
"for (varctr := 0; varctr < varcount; varctr := varctr + 1)\n"
"{\n"
"http(\'\\n    <th>\', ses);\n"
"http_escape (metas[0][varctr][0], 11, ses, 0, 1);\n"
"http(\'</th>\', ses);\n"
"}\n"
"http (\'\\n  </tr>\', ses);\n"
"for (resctr := 0; resctr < rescount; resctr := resctr + 1)\n"
"{\n"
"http(\'\\n  <tr>\', ses);\n"
"\n"
"\n"
"\n"
"for (varctr := 0; varctr < varcount; varctr := varctr + 1)\n"
"{\n"
"declare val any;\n"
"val := rset[resctr][varctr];\n"
"if (val is null)\n"
"{\n"
"http(\'\\n    <td></td>\', ses);\n"
"goto end_of_val_print;\n"
"}\n"
"http(\'\\n    <td>\', ses);\n"
"if (isiri_id (val))\n"
"{\n"
"http_escape (DB.DBA.RDF_QNAME_OF_IID (val), 11, ses, 0, 1);\n"
"}\n"
"else if (182 = __tag (val))\n"
"{\n"
"http_escape (val, 11, ses, 1, 1);\n"
"}\n"
"else if (193 = __tag (val))\n"
"{\n"
"http_escape (val[1], 11, ses, 1, 1);\n"
"}\n"
"else if (185 = __tag (val))\n"
"{\n"
"http_escape (cast (val as varchar), 11, ses, 1, 1);\n"
"}\n"
"else\n"
"{\n"
"http_escape (DB.DBA.RDF_STRSQLVAL_OF_LONG (val), 11, ses, 1, 1);\n"
"}\n"
"http (\'</td>\', ses);\n"
"end_of_val_print: ;\n"
"}\n"
"http(\'\\n  </tr>\', ses);\n"
"}\n"
"http (\'\\n</table>\', ses);\n"
"}\n"
"--src sparql.sql:4846\n";

static const char *proc159 = 
"create procedure SPARQL_RESULTS_JSON_WRITE (inout ses any, inout metas any, inout rset any)\n"
"{\n"
"declare varctr, varcount, resctr, rescount integer;\n"
"varcount := length (metas[0]);\n"
"rescount := length (rset);\n"
"http (\'\\n{ \"head\": { \"link\": [], \"vars\": [\', ses);\n"
"for (varctr := 0; varctr < varcount; varctr := varctr + 1)\n"
"{\n"
"if (varctr > 0)\n"
"http(\', \"\', ses);\n"
"else\n"
"http(\'\"\', ses);\n"
"http_escape (metas[0][varctr][0], 11, ses, 0, 1);\n"
"http(\'\"\', ses);\n"
"}\n"
"http (\'] },\\n  \"results\": { \"distinct\": false, \"ordered\": true, \"bindings\": [\', ses);\n"
"for (resctr := 0; resctr < rescount; resctr := resctr + 1)\n"
"{\n"
"declare need_comma integer;\n"
"if (resctr > 0)\n"
"http(\',\\n    {\', ses);\n"
"else\n"
"http(\'\\n    {\', ses);\n"
"need_comma := 0;\n"
"for (varctr := 0; varctr < varcount; varctr := varctr + 1)\n"
"{\n"
"declare val any;\n"
"val := rset[resctr][varctr];\n"
"if (val is null)\n"
"goto end_of_val_print;\n"
"if (need_comma)\n"
"http(\'\\t, \"\', ses);\n"
"else\n"
"{\n"
"http(\' \"\', ses);\n"
"need_comma := 1;\n"
"}\n"
"http_escape (metas[0][varctr][0], 11, ses, 0, 1);\n"
"http(\'\": { \', ses);\n"
"if (isiri_id (val))\n"
"{\n"
"if (val > #i1000000000)\n"
"http (sprintf (\'\"type\": \"bnode\", \"value\": \"nodeID://%d\', iri_id_num (val)), ses);\n"
"else\n"
"{\n"
"http (\'\"type\": \"uri\", \"value\": \"\', ses);\n"
"http_escape (DB.DBA.RDF_QNAME_OF_IID (val), 11, ses, 0, 1);\n"
"}\n"
"}\n"
"else if (193 = __tag (val))\n"
"{\n"
"declare res varchar;\n"
"if (257 <> val[0])\n"
"{\n"
"http (\'\"type\": \"typed-literal\", \"datatype\": \"\', ses);\n"
"res := coalesce ((select RDT_QNAME from DB.DBA.RDF_DATATYPE where RDT_TWOBYTE = val[0]));\n"
"http_escape (res, 11, ses, 1, 1);\n"
"http (\'\", \"value\": \"\', ses);\n"
"}\n"
"else if (257 <> val[2])\n"
"{\n"
"http (\'\"type\": \"literal\", \"xml:lang\": \"\', ses);\n"
"res := coalesce ((select RL_ID from DB.DBA.RDF_LANGUAGE where RL_TWOBYTE = val[2]));\n"
"http_escape (res, 11, ses, 1, 1);\n"
"http (\'\", \"value\": \"\', ses);\n"
"}\n"
"else\n"
"http (\'\"type\": \"literal\", \"value\": \"\', ses);\n"
"http_escape (val[1], 11, ses, 1, 1);\n"
"}\n"
"else if (182 = __tag (val))\n"
"{\n"
"http (\'\"type\": \"literal\", \"value\": \"\', ses);\n"
"http_escape (val, 11, ses, 1, 1);\n"
"}\n"
"else if (185 = __tag (val))\n"
"{\n"
"http (\'\"type\": \"literal\", \"value\": \"\', ses);\n"
"http_escape (cast (val as varchar), 11, ses, 1, 1);\n"
"}\n"
"else\n"
"{\n"
"http (\'\"type\": \"typed-literal\", \"datatype\": \"\', ses);\n"
"http_escape (cast (DB.DBA.RDF_DATATYPE_OF_TAG (__tag (val)) as varchar), 11, ses, 1, 1);\n"
"http (\'\", \"value\": \"\', ses);\n"
"http_escape (DB.DBA.RDF_STRSQLVAL_OF_LONG (val), 11, ses, 1, 1);\n"
"}\n"
"http (\'\" }\', ses);\n"
"\n"
"end_of_val_print: ;\n"
"}\n"
"http(\'}\', ses);\n"
"}\n"
"http (\' ] } }\', ses);\n"
"}\n"
"--src sparql.sql:4919\n";

static const char *proc160 = 
"create function DB.DBA.SPARQL_RESULTS_WRITE (inout ses any, inout metas any, inout rset any, in accept varchar, in add_http_headers integer) returns varchar\n"
"{\n"
"declare singlefield varchar;\n"
"declare ret_mime varchar;\n"
"if ((1 >= length (rset)) and (1 = length (metas[0])))\n"
"singlefield := metas[0][0][0];\n"
"else\n"
"singlefield := NULL;\n"
"\n"
"if (\'__ask_retval\' = singlefield)\n"
"{\n"
"if (strstr (accept, \'application/sparql-results+json\') is not null or strstr (accept, \'application/json\') is not null)\n"
"{\n"
"if (strstr (accept, \'application/sparql-results+json\') is not null)\n"
"ret_mime := \'application/sparql-results+json\';\n"
"else\n"
"ret_mime := \'application/json\';\n"
"http (\n"
"concat (\n"
"\'{  \"head\": { \"link\": [] }, \"boolean\": \',\n"
"case (length (rset)) when 0 then \'false\' else \'true\' end,\n"
"\'}\'),\n"
"ses );\n"
"}\n"
"else\n"
"{\n"
"ret_mime := \'application/sparql-results+xml\';\n"
"SPARQL_RESULTS_XML_WRITE_NS (ses);\n"
"http (\n"
"concat (\n"
"\'\\n <head></head>\\n <boolean>\',\n"
"case (length (rset)) when 0 then \'false\' else \'true\' end,\n"
"\'</boolean>\\n</sparql>\'),\n"
"ses );\n"
"}\n"
"goto body_complete;\n"
"}\n"
"if ((1 = length (rset)) and\n"
"(1 = length (rset[0])) and\n"
"(214 = __tag (rset[0][0])) )\n"
"{\n"
"declare triples any;\n"
"triples := dict_list_keys (rset[0][0], 1);\n"
"if (strstr (accept, \'text/rdf+n3\') is not null or (accept = \'auto\'))\n"
"{\n"
"ret_mime := \'text/rdf+n3\';\n"
"DB.DBA.RDF_TRIPLES_TO_TTL (triples, ses);\n"
"}\n"
"else\n"
"{\n"
"ret_mime := \'application/rdf+xml\';\n"
"DB.DBA.RDF_TRIPLES_TO_RDF_XML_TEXT (triples, 1, ses);\n"
"}\n"
"goto body_complete;\n"
"}\n"
"if (strstr (accept, \'application/sparql-results+json\') is not null or strstr (accept, \'application/json\') is not null)\n"
"{\n"
"if (strstr (accept, \'application/sparql-results+json\') is not null)\n"
"ret_mime := \'application/sparql-results+json\';\n"
"else\n"
"ret_mime := \'application/json\';\n"
"if ((\'callretRDF/XML-0\' = singlefield) or (\'callretTURTLE-0\' = singlefield) or (\'callretTTL-0\' = singlefield))\n"
"{\n"
"http(\'\"\', ses);\n"
"http_escape (cast (rset[0][0] as varchar), 11, ses, 0, 1);\n"
"http(\'\"\', ses);\n"
"}\n"
"else\n"
"SPARQL_RESULTS_JSON_WRITE (ses, metas, rset);\n"
"goto body_complete;\n"
"}\n"
"if (strstr (accept, \'text/html\') is not null)\n"
"{\n"
"ret_mime := \'text/html\';\n"
"SPARQL_RESULTS_JAVASCRIPT_HTML_WRITE(ses, metas, rset, 0);\n"
"goto body_complete;\n"
"}\n"
"if (strstr (accept, \'application/javascript\') is not null)\n"
"{\n"
"ret_mime := \'application/javascript\';\n"
"SPARQL_RESULTS_JAVASCRIPT_HTML_WRITE(ses, metas, rset, 1);\n"
"goto body_complete;\n"
"}\n"
"if (strstr (accept, \'application/soap+xml\') is not null)\n"
"{\n"
"ret_mime := \'application/soap+xml\';\n"
"http (\'<soapenv:Envelope xmlns:soapenv=\"http://www.w3.org/2003/05/soap-envelope\"><soapenv:Body><query-result xmlns=\"http://www.w3.org/2005/09/sparql-protocol-types/#\">\', ses);\n"
"SPARQL_RESULTS_XML_WRITE_NS (ses);\n"
"SPARQL_RESULTS_XML_WRITE_HEAD (ses, metas);\n"
"SPARQL_RESULTS_XML_WRITE_RES (ses, metas, rset);\n"
"http (\'\\n</sparql>\', ses);\n"
"http (\'</query-result></soapenv:Body></soapenv:Envelope>\', ses);\n"
"goto body_complete;\n"
"}\n"
"if ((\'callretRDF/XML-0\' = singlefield) and (\'auto\' = accept))\n"
"{\n"
"ret_mime := \'application/rdf+xml\';\n"
"http (rset[0][0]);\n"
"goto body_complete;\n"
"}\n"
"if (((\'callretTURTLE-0\' = singlefield) or (\'callretTTL-0\' = singlefield)) and (\'auto\' = accept))\n"
"{\n"
"ret_mime := \'text/rdf+n3\';\n"
"http (rset[0][0]);\n"
"goto body_complete;\n"
"}\n"
"if (strstr (accept, \'application/sparql-results+xml\') is null)\n"
"{\n"
"if (strstr (accept, \'text/rdf+n3\') is not null)\n"
"{\n"
"ret_mime := \'text/rdf+n3\';\n"
"SPARQL_RESULTS_TTL_WRITE_NS (ses);\n"
"SPARQL_RESULTS_TTL_WRITE_HEAD (ses, metas);\n"
"if (length (rset) > 0)\n"
"http (\' ;\', ses);\n"
"SPARQL_RESULTS_TTL_WRITE_RES (ses, metas, rset);\n"
"http (\' .\', ses);\n"
"goto body_complete;\n"
"}\n"
"if (strstr (accept, \'application/rdf+xml\') is not null)\n"
"{\n"
"ret_mime := \'application/rdf+xml\';\n"
"SPARQL_RESULTS_RDFXML_WRITE_NS (ses);\n"
"SPARQL_RESULTS_RDFXML_WRITE_HEAD (ses, metas);\n"
"SPARQL_RESULTS_RDFXML_WRITE_RES (ses, metas, rset);\n"
"http (\'\\n  </rdf:Description>\', ses);\n"
"http (\'\\n</rdf:RDF>\', ses);\n"
"goto body_complete;\n"
"}\n"
"}\n"
"ret_mime := \'application/sparql-results+xml\';\n"
"SPARQL_RESULTS_XML_WRITE_NS (ses);\n"
"SPARQL_RESULTS_XML_WRITE_HEAD (ses, metas);\n"
"SPARQL_RESULTS_XML_WRITE_RES (ses, metas, rset);\n"
"http (\'\\n</sparql>\', ses);\n"
"\n"
"body_complete:\n"
"if (add_http_headers)\n"
"http_header (\'Content-Type: \' || ret_mime || \'\\r\\n\');\n"
"return ret_mime;\n"
"}\n"
"--src sparql.sql:5016\n";

static const char *proc161 = 
"create procedure WS.WS.SPARQL_VHOST_RESET ()\n"
"{\n"
"if (not exists (select 1 from \"DB\".\"DBA\".\"SYS_USERS\" where U_NAME = \'SPARQL\'))\n"
"{\n"
"DB.DBA.USER_CREATE (\'SPARQL\', uuid(), vector (\'DISABLED\', 1, \'LOGIN_QUALIFIER\', \'SPARQL\'));\n"
"DB.DBA.EXEC_STMT (\'grant SPARQL_SELECT to \"SPARQL\"\', 0);\n"
"}\n"
"DB.DBA.VHOST_REMOVE (lpath=>\'/SPARQL\');\n"
"DB.DBA.VHOST_REMOVE (lpath=>\'/sparql\');\n"
"DB.DBA.VHOST_REMOVE (lpath=>\'/services/sparql-query\');\n"
"DB.DBA.VHOST_DEFINE (lpath=>\'/sparql/\', ppath => \'/!sparql/\', is_dav => 1, vsp_user => \'dba\', opts => vector(\'noinherit\', 1));\n"
"\n"
"\n"
"\n"
"}\n"
"--src sparql.sql:5163\n";

static const char *proc162 = 
"create procedure DB.DBA.SPARQL_PROTOCOL_ERROR_REPORT (\n"
"inout path varchar, inout params any, inout lines any,\n"
"in httpcode varchar, in httpstatus varchar,\n"
"in query varchar, in state varchar, in msg varchar, in accept varchar := null)\n"
"{\n"
"\n"
"if (httpstatus is null)\n"
"{\n"
"declare errtitle varchar;\n"
"declare delim varchar;\n"
"delim := strchr (msg, \'\\n\');\n"
"if (delim is null)\n"
"errtitle := msg;\n"
"else\n"
"errtitle := subseq (msg, 0, delim);\n"
"httpstatus := sprintf (\'Error %s %s\', state, errtitle);\n"
"}\n"
"if (accept is not null and accept = \'application/soap+xml\')\n"
"{\n"
"declare err_str any;\n"
"http_request_status (sprintf (\'HTTP/1.1 500 %s\', httpstatus));\n"
"err_str := soap_make_error (\'320\', state, msg, 12);\n"
"http (err_str);\n"
"return;\n"
"}\n"
"\n"
"http_request_status (sprintf (\'HTTP/1.1 %s %s\', httpcode, httpstatus));\n"
"http_header (\'Content-Type: text/plain\\r\\n\');\n"
"http (concat (state, \' Error \', msg));\n"
"if (query is not null)\n"
"{\n"
"http (\'\\n\\nSPARQL query:\\n\');\n"
"http (query);\n"
"}\n"
"}\n"
"--src sparql.sql:5183\n";

static const char *proc163 = 
"create procedure DB.DBA.SPARQL_WSDL (in lines any)\n"
"{\n"
"declare host any;\n"
"host := http_request_header (lines, \'Host\', null, null);\n"
"http (sprintf (\'<?xml version=\"1.0\" encoding=\"utf-8\"?>\n"
"<description xmlns=\"http://www.w3.org/2006/01/wsdl\"\n"
"xmlns:tns=\"http://www.w3.org/2005/08/sparql-protocol-query/#\"\n"
"targetNamespace=\"http://www.w3.org/2005/08/sparql-protocol-query/#\">\n"
"<include location=\"http://www.w3.org/TR/rdf-sparql-protocol/sparql-protocol-query.wsdl\" />\n"
"<service name=\"SparqlService\" interface=\"tns:SparqlQuery\">\n"
"<endpoint name=\"SparqlEndpoint\" binding=\"tns:querySoap\" address=\"http://%s/sparql\"/>\n"
"</service>\n"
"</description>\', host));\n"
"}\n"
"--src sparql.sql:5220\n";

static const char *proc164 = 
"create procedure WS.WS.\"/!sparql/\" (inout path varchar, inout params any, inout lines any)\n"
"{\n"
"declare query, dflt_graph, full_query, format varchar;\n"
"declare named_graphs any;\n"
"declare paramctr, paramcount, maxrows, can_sponge, should_sponge integer;\n"
"declare ses, content any;\n"
"declare def_max, add_http_headers int;\n"
"declare http_meth, content_type, ini_dflt_graph varchar;\n"
"\n"
"ses := 0;\n"
"query := null;\n"
"dflt_graph := null;\n"
"format := \'\';\n"
"add_http_headers := 1;\n"
"named_graphs := vector ();\n"
"maxrows := 1024*1024;\n"
"http_meth := http_request_get (\'REQUEST_METHOD\');\n"
"ini_dflt_graph := cfg_item_value (virtuoso_ini_path (), \'SPARQL\', \'DefaultGraph\');\n"
"content_type := http_request_header (lines, \'Content-Type\', null, \'\');\n"
"content := null;\n"
"can_sponge := coalesce ((select top 1 1\n"
"from DB.DBA.SYS_USERS as sup\n"
"join DB.DBA.SYS_ROLE_GRANTS as g on (sup.U_ID = g.GI_SUPER)\n"
"join DB.DBA.SYS_USERS as sub on (g.GI_SUB = sub.U_ID)\n"
"where sup.U_NAME = \'SPARQL\' and sub.U_NAME = \'SPARQL_UPDATE\' ), 0);\n"
"declare exit handler for sqlstate \'*\' {\n"
"DB.DBA.SPARQL_PROTOCOL_ERROR_REPORT (path, params, lines,\n"
"\'500\', \'SPARQL Request Failed\',\n"
"query, __SQL_STATE, __SQL_MESSAGE, format);\n"
"return;\n"
"};\n"
"\n"
"\n"
"if (http_path () = \'/sparql/services.wsdl\')\n"
"{\n"
"http_header (\'Content-Type: application/wsdl+xml\\r\\n\');\n"
"\n"
"DB.DBA.SPARQL_WSDL (lines);\n"
"return;\n"
"}\n"
"\n"
"paramcount := length (params);\n"
"if (((0 = paramcount) or ((2 = paramcount) and (\'Content\' = params[0]))) and content_type <> \'application/soap+xml\')\n"
"{\n"
"declare redir varchar;\n"
"redir := registry_get (\'WS.WS.SPARQL_DEFAULT_REDIRECT\');\n"
"if (isstring (redir))\n"
"{\n"
"http_request_status (\'HTTP/1.1 301 Moved Permanently\');\n"
"http_header (sprintf (\'Location: %s\\r\\n\', redir));\n"
"return;\n"
"}\n"
"http(\'<html xmlns=\"http://www.w3.org/1999/xhtml\">\\n\');\n"
"http(\'	<head>\\n\');\n"
"http(\'		<title>Virtuoso SPARQL Query Form</title>\\n\');\n"
"http(\'		<style type=\"text/css\">\\n\');\n"
"http(\'label.n\');\n"
"http(\'{ display: inline; margin-top: 10pt; }\\n\');\n"
"http(\'body { font-family: arial, helvetica, sans-serif; font-size: 9pt; color: #234; }\\n\');\n"
"http(\'fieldset { border: 2px solid #86b9d9; }\\n\');\n"
"http(\'legend { font-size: 12pt; color: #86b9d9; }\\n\');\n"
"http(\'label { font-weight: bold; }\\n\');\n"
"http(\'h1 { width: 100%; background-color: #86b9d9; font-size: 18pt; font-weight: normal; color: #fff; height: 4ex; text-align: right; vertical-align: middle; padding-right:  8px; }\\n\');\n"
"http(\'		</style>\\n\');\n"
"http(\'		<script language=\"JavaScript\">\\n\');\n"
"http(\'var last_format = 1;\\n\');\n"
"http(\'function format_select(query_obg)\\n\');\n"
"http(\'{\\n\');\n"
"http(\'  var query = query_obg.value; \\n\');\n"
"http(\'  var format = query_obg.form.format;\\n\');\n"
"http(\'\\n\');\n"
"http(\'  if (query.match(/construct/i) && last_format == 1) {\\n\');\n"
"http(\'    for(var i = format.options.length; i > 0; i--)\\n\');\n"
"http(\'      format.options[i] = null;\');\n"
"http(\'    format.options[1] = new Option(\\\'N3/Turtle\\\',\\\'text/rdf+n3\\\');\\n\');\n"
"http(\'    format.options[2] = new Option(\\\'RDF/XML\\\',\\\'application/rdf+xml\\\');\\n\');\n"
"http(\'    format.selectedIndex = 1;\\n\');\n"
"http(\'    last_format = 2;\\n\');\n"
"http(\'  }\\n\');\n"
"http(\'\\n\');\n"
"http(\'  if (!query.match(/construct/i) && last_format == 2) {\\n\');\n"
"http(\'    for(var i = format.options.length; i > 0; i--)\\n\');\n"
"http(\'      format.options[i] = null;\\n\');\n"
"http(\'    format.options[1] = new Option(\\\'HTML\\\',\\\'text/html\\\');\\n\');\n"
"http(\'    format.options[2] = new Option(\\\'XML\\\',\\\'application/sparql-results+xml\\\');\\n\');\n"
"http(\'    format.options[3] = new Option(\\\'JSON\\\',\\\'application/sparql-results+json\\\');\\n\');\n"
"http(\'    format.options[4] = new Option(\\\'Javascript\\\',\\\'application/javascript\\\');\\n\');\n"
"http(\'    format.selectedIndex = 1;\\n\');\n"
"http(\'    last_format = 1;\\n\');\n"
"http(\'  }\\n\');\n"
"http(\'}\\n\');\n"
"http(\'		</script>\\n\');\n"
"http(\'	</head>\\n\');\n"
"http(\'	<body>\\n\');\n"
"http(\'		<div id=\"header\">\\n\');\n"
"http(\'			<h1>OpenLink Virtuoso SPARQL Query</h1>\\n\');\n"
"http(\'		</div>\\n\');\n"
"http(\'		<div id=\"main\">\\n\');\n"
"http(\'			<p>This query page is designed to help you test Openlink Virtuoso SPARQL protocol endpoint. <br/>\\n\');\n"
"http(\'			Consult the <a href=\"http://virtuoso.openlinksw.com/wiki/main/Main/VOSSparqlProtocol\">Virtuoso Wiki page</a> describing the service \\n\');\n"
"http(\'			or the <a href=\"http://docs.openlinksw.com/virtuoso/\">Online Virtuoso Documentation</a> section <a href=\"http://docs.openlinksw.com/virtuoso/rdfandsparql.html\">RDF Database and SPARQL</a>.</p>\\n\');\n"
"http(\'			<p>There is also a rich Web based user interface with sample queries. \\n\');\n"
"if (DB.DBA.VAD_CHECK_VERSION(\'iSPARQL\') is null)\n"
"http(\'			In order to use it you must install the iSPARQL package (isparql_dav.vad).</p>\\n\');\n"
"else\n"
"http(\'			You can access it at: <a href=\"/isparql\">/isparql</a>.</p>\\n\');\n"
"http(\'			<form action=\"\" method=\"GET\">\\n\');\n"
"http(\'			<fieldset>\\n\');\n"
"http(\'			<legend>Query</legend>\\n\');\n"
"http(\'			  <label for=\"default-graph-uri\">Default Graph URI</label>\\n\');\n"
"http(\'			  <br />\\n\');\n"
"http(\'			  <input type=\"text\" name=\"default-graph-uri\" id=\"default-graph-uri\"\\n\');\n"
"http(sprintf (\'				  	value=\"%s\" size=\"80\"/>\\n\', coalesce (ini_dflt_graph, \'\') ));\n"
"http(\'			  <br /><br />\\n\');\n"
"if (can_sponge)\n"
"{\n"
"http(\'			  <input type=\"checkbox\"\' ||\n"
"case (isnull (get_keyword (\'should-sponge\', params))) when 0 then \' checked=\"checked\"\' else \'\' end ||\n"
"\' name=\"should-sponge\" id=\"should-sponge\" value=\"soft\"/>\\n\');\n"
"http(\'			  <label for=\"should-sponge\">Retrieve remote RDF data for all missing source graphs</label>\\n\');\n"
"http(\'			  <br /><br />\\n\');\n"
"}\n"
"else\n"
"{\n"
"http(\'			  <i>Security restrictions of this server does not allow you to retrieve remote RDF data.\n"
"DBA may wish to grant \"SPARQL_UPDATE\" privilege to \"SPARQL\" account to remove the restriction.</i>\\n\');\n"
"http(\'			  <br /><br />\\n\');\n"
"}\n"
"http(\'			  <label for=\"query\">Query text</label>\\n\');\n"
"http(\'			  <br />\\n\');\n"
"http(\'			  <textarea rows=\"10\" cols=\"60\" name=\"query\" id=\"query\" onchange=\"format_select(this)\" onkeyup=\"format_select(this)\">SELECT * WHERE {?s ?p ?o}</textarea>\\n\');\n"
"http(\'			  <br /><br />\\n\');\n"
"\n"
"\n"
"\n"
"\n"
"http(\'			  <label for=\"format\" class=\"n\">Display Results As:</label>\\n\');\n"
"http(\'			  <select name=\"format\">\\n\');\n"
"http(\'			    <option value=\"auto\">Auto</option>\\n\');\n"
"http(\'			    <option value=\"text/html\" selected=\"selected\">HTML</option>\\n\');\n"
"http(\'			    <option value=\"application/sparql-results+xml\">XML</option>\\n\');\n"
"http(\'			    <option value=\"application/sparql-results+json\">JSON</option>\\n\');\n"
"http(\'			    <option value=\"application/javascript\">Javascript</option>\\n\');\n"
"http(\'			  </select>\\n\');\n"
"http(\'			  <input type=\"submit\" value=\"Run Query\"/>\\n\');\n"
"http(\'			  <input type=\"reset\" value=\"Reset\"/>\\n\');\n"
"http(\'			</fieldset>\\n\');\n"
"http(\'			</form>\\n\');\n"
"http(\'		</div>\\n\');\n"
"http(\'	</body>\\n\');\n"
"http(\'</html>\\n\');\n"
"return;\n"
"}\n"
"for (paramctr := 0; paramctr < paramcount; paramctr := paramctr + 2)\n"
"{\n"
"declare pname, pvalue varchar;\n"
"pname := params [paramctr];\n"
"pvalue := params [paramctr+1];\n"
"if (\'query\' = pname)\n"
"query := pvalue;\n"
"else if (\'default-graph-uri\' = pname and length (pvalue))\n"
"dflt_graph := pvalue;\n"
"else if (\'named-graph-uri\' = pname)\n"
"{\n"
"if (position (pvalue, named_graphs) < 0)\n"
"named_graphs := vector_concat (named_graphs, vector (pvalue));\n"
"}\n"
"else if (\'maxrows\' = pname)\n"
"{\n"
"maxrows := cast (pvalue as integer);\n"
"}\n"
"else if (\'should-sponge\' = pname)\n"
"{\n"
"if (can_sponge)\n"
"should_sponge := 1;\n"
"}\n"
"else if (\'format\' = pname)\n"
"{\n"
"format := pvalue;\n"
"}\n"
"else if (query is null and \'query-uri\' = pname and length (pvalue))\n"
"{\n"
"if (cfg_item_value (virtuoso_ini_path (), \'SPARQL\', \'ExternalQuerySource\') = \'1\')\n"
"{\n"
"declare uri varchar;\n"
"declare hf, hdr, charset any;\n"
"uri := pvalue;\n"
"if (uri like \'http://%\' and uri not like \'http://localdav.virt/%\' and uri not like \'http://local.virt/dav/%\')\n"
"{\n"
"query := http_get (uri, hdr);\n"
"if (hdr[0] not like \'% 200%\')\n"
"signal (\'22023\', concat (\'HTTP request failed: \', hdr[0], \'for URI \', uri));\n"
"charset := http_request_header (hdr, \'Content-Type\', \'charset\', \'\');\n"
"if (charset <> \'\')\n"
"{\n"
"query := charset_recode (query, charset, \'UTF-8\');\n"
"}\n"
"}\n"
"else\n"
"{\n"
"query := XML_URI_GET (\'\', pvalue);\n"
"}\n"
"}\n"
"else\n"
"{\n"
"DB.DBA.SPARQL_PROTOCOL_ERROR_REPORT (path, params, lines,\n"
"\'403\', \'Prohibited\', query, \'22023\', \'The external query sources are prohibited.\');\n"
"return;\n"
"}\n"
"}\n"
"else if (\'xslt-uri\' = pname and length (pvalue))\n"
"{\n"
"if (cfg_item_value (virtuoso_ini_path (), \'SPARQL\', \'ExternalXsltSource\') = \'1\')\n"
"{\n"
"add_http_headers := 0;\n"
"http_xslt (pvalue);\n"
"}\n"
"else\n"
"{\n"
"DB.DBA.SPARQL_PROTOCOL_ERROR_REPORT (path, params, lines,\n"
"\'403\', \'Prohibited\', query, \'22023\', \'The XSL-T transformation is prohibited\');\n"
"return;\n"
"}\n"
"}\n"
"}\n"
"def_max := atoi (coalesce (cfg_item_value (virtuoso_ini_path (), \'SPARQL\', \'ResultSetMaxRows\'), \'-1\'));\n"
"if (def_max > 0 and def_max < maxrows)\n"
"maxrows := def_max;\n"
"\n"
"\n"
"\n"
"\n"
"\n"
"\n"
"if (http_meth = \'POST\' and content_type = \'application/soap+xml\')\n"
"{\n"
"declare xt, ng any;\n"
"content := http_body_read ();\n"
"\n"
"xt := xtree_doc (content);\n"
"query := charset_recode (xpath_eval (\'[ xmlns:soap=\"http://www.w3.org/2003/05/soap-envelope\" xmlns:sp=\"http://www.w3.org/2005/09/sparql-protocol-types/#\" ] string (/soap:Envelope/soap:Body/sp:query-request/sp:query)\', xt), \'_WIDE_\', \'UTF-8\');\n"
"dflt_graph := charset_recode (xpath_eval (\'[ xmlns:soap=\"http://www.w3.org/2003/05/soap-envelope\" xmlns:sp=\"http://www.w3.org/2005/09/sparql-protocol-types/#\" ] string (/soap:Envelope/soap:Body/sp:query-request/sp:default-graph-uri)\', xt), \'_WIDE_\', \'UTF-8\');\n"
"ng := xpath_eval (\'[ xmlns:soap=\"http://www.w3.org/2003/05/soap-envelope\" xmlns:sp=\"http://www.w3.org/2005/09/sparql-protocol-types/#\" ] /soap:Envelope/soap:Body/sp:query-request/sp:named-graph-uri\', xt, 0);\n"
"\n"
"foreach (any frag in ng) do\n"
"{\n"
"declare pvalue varchar;\n"
"pvalue := charset_recode (xpath_eval (\'string(.)\', frag), \'_WIDE_\', \'UTF-8\');\n"
"if (position (pvalue, named_graphs) < 0)\n"
"named_graphs := vector_concat (named_graphs, vector (pvalue));\n"
"}\n"
"format := \'application/soap+xml\';\n"
"}\n"
"\n"
"if (query is null)\n"
"{\n"
"if (strstr (content_type, \'application/xml\') is not null)\n"
"{\n"
"DB.DBA.SPARQL_PROTOCOL_ERROR_REPORT (path, params, lines,\n"
"\'400\', \'Bad Request\',\n"
"query, \'22023\', \'XML notation of SPARQL queries is not supported\' );\n"
"return;\n"
"}\n"
"DB.DBA.SPARQL_PROTOCOL_ERROR_REPORT (path, params, lines,\n"
"\'400\', \'Bad Request\',\n"
"query, \'22023\', \'The request does not contain text of SPARQL query\', format);\n"
"return;\n"
"}\n"
"if (dflt_graph is null)\n"
"{\n"
"declare req_hosts varchar;\n"
"declare req_hosts_split any;\n"
"declare hctr integer;\n"
"req_hosts := http_request_header (lines, \'Host\', null, null);\n"
"req_hosts := replace (req_hosts, \', \', \',\');\n"
"req_hosts_split := split_and_decode (req_hosts, 0, \'\\0\\0,\');\n"
"for (hctr := length (req_hosts_split) - 1;\n"
"(hctr >= 0) and dflt_graph is null;\n"
"hctr := hctr - 1)\n"
"{\n"
"dflt_graph := coalesce ((\n"
"select SH_GRAPH_URI from DB.DBA.SYS_SPARQL_HOST\n"
"where SH_HOST = req_hosts_split [hctr] ) );\n"
"}\n"
"full_query := query;\n"
"}\n"
"else\n"
"{\n"
"full_query := concat (\'define input:default-graph-uri \"\', dflt_graph, \'\" \', query);\n"
"}\n"
"if (can_sponge and should_sponge)\n"
"full_query := concat (\'define get:soft \"soft\" \', full_query);\n"
"full_query := concat (\'define output:valmode \"LONG\" \', full_query);\n"
"declare state, msg varchar;\n"
"declare metas, rset any;\n"
"state := \'00000\';\n"
"metas := null;\n"
"rset := null;\n"
"exec (\'isnull (sparql_to_sql_text (?))\', state, msg, vector (full_query));\n"
"if (state <> \'00000\')\n"
"{\n"
"DB.DBA.SPARQL_PROTOCOL_ERROR_REPORT (path, params, lines,\n"
"\'400\', \'Bad Request\',\n"
"query, state, msg, format);\n"
"return;\n"
"}\n"
"state := \'00000\';\n"
"metas := null;\n"
"rset := null;\n"
"connection_set (\':default_graph\', dflt_graph);\n"
"connection_set (\':named_graphs\', named_graphs);\n"
"http_header (sprintf (\'X-SPARQL-default-graph: %U\\r\\n\', dflt_graph));\n"
"\n"
"\n"
"set_user_id (\'SPARQL\');\n"
"exec ( concat (\'sparql \', full_query), state, msg, vector(), maxrows, metas, rset);\n"
"\n"
"if (state <> \'00000\')\n"
"{\n"
"DB.DBA.SPARQL_PROTOCOL_ERROR_REPORT (path, params, lines,\n"
"\'500\', \'SPARQL Request Failed\',\n"
"query, state, msg, format);\n"
"return;\n"
"}\n"
"declare accept varchar;\n"
"accept := http_request_header (lines, \'Accept\', null, \'\');\n"
"if (format <> \'\')\n"
"accept := format;\n"
"DB.DBA.SPARQL_RESULTS_WRITE (ses, metas, rset, accept, add_http_headers);\n"
"}\n"
"--src sparql.sql:5236\n";

static const char *other13 = 
"registry_set (\'/!sparql/\', \'no_vsp_recompile\')\n";
static const char *proc165 = 
"create procedure DB.DBA.TTLP_EXEC_TRIPLE_W (\n"
"in g_iid IRI_ID,\n"
"in s_iid IRI_ID,\n"
"in p_iid IRI_ID,\n"
"in o_iid any)\n"
"{\n"
"log_enable (0);\n"
"insert soft DB.DBA.RDF_QUAD (G,S,P,O)\n"
"values (g_iid, s_iid, p_iid, o_iid);\n"
"commit work;\n"
"}\n"
"--src sparql.sql:5579\n";

static const char *proc166 = 
"create procedure DB.DBA.TTLP_EXEC_TRIPLE_W_L (\n"
"in g_iid IRI_ID,\n"
"in s_iid IRI_ID,\n"
"in p_iid IRI_ID,\n"
"in o_val any, in o_type any, in o_lang any)\n"
"{\n"
"log_enable (0);\n"
"if (\'http://www.w3.org/2001/XMLSchema#boolean\' = o_type)\n"
"{\n"
"if ((\'true\' = o_val) or (\'1\' = o_val))\n"
"o_val := 1;\n"
"else if ((\'false\' = o_val) or (\'0\' = o_val))\n"
"o_val := 0;\n"
"else signal (\'RDFXX\', \'Invalid notation of boolean literal\');\n"
"}\n"
"else if (\'http://www.w3.org/2001/XMLSchema#dateTime\' = o_type)\n"
"{\n"
"o_val := __xqf_str_parse (\'dateTime\', o_val);\n"
"}\n"
"else if (\'http://www.w3.org/2001/XMLSchema#double\' = o_type)\n"
"{\n"
"o_val := cast (o_val as double precision);\n"
"}\n"
"else if (\'http://www.w3.org/2001/XMLSchema#float\' = o_type)\n"
"{\n"
"o_val := cast (o_val as float);\n"
"}\n"
"else if (\'http://www.w3.org/2001/XMLSchema#integer\' = o_type)\n"
"{\n"
"o_val := cast (o_val as int);\n"
"}\n"
"else if (isstring (o_type) or isstring (o_lang))\n"
"{\n"
"if (not isstring (o_type))\n"
"o_type := null;\n"
"if (not isstring (o_lang))\n"
"o_lang := null;\n"
"\n"
"insert soft DB.DBA.RDF_QUAD (G,S,P,O)\n"
"values (g_iid, s_iid, p_iid,\n"
"DB.DBA.RDF_MAKE_OBJ_OF_TYPEDSQLVAL (o_val,\n"
"iri_to_id (o_type, 1),\n"
"o_lang ) );\n"
"commit work;\n"
"return;\n"
"}\n"
"\n"
"insert soft DB.DBA.RDF_QUAD (G,S,P,O)\n"
"values (g_iid, s_iid, p_iid, DB.DBA.RDF_MAKE_OBJ_OF_SQLVAL (o_val));\n"
"commit work;\n"
"}\n"
"--src sparql.sql:5593\n";

static const char *proc167 = 
"create procedure DB.DBA.TTLP_EXEC_TRIPLE_A (\n"
"in g_uri varchar, in g_iid IRI_ID,\n"
"in s_uri varchar, in s_iid IRI_ID,\n"
"in p_uri varchar, in p_iid IRI_ID,\n"
"in o_uri varchar, in o_iid IRI_ID,\n"
"inout app_env any )\n"
"{\n"
"declare err any;\n"
"app_env[1] := aq_request (app_env[0], \'DB.DBA.TTLP_EXEC_TRIPLE_W\', vector (g_iid, s_iid, p_iid, o_iid));\n"
"if (mod (app_env[1], 4000) = 0)\n"
"{\n"
"commit work;\n"
"aq_wait (app_env[0], app_env[1], err, 1);\n"
"aq_wait_all (app_env[0]);\n"
"}\n"
"}\n"
"--src sparql.sql:5647\n";

static const char *proc168 = 
"create procedure DB.DBA.TTLP_EXEC_TRIPLE_L_A (\n"
"in g_uri varchar, in g_iid IRI_ID,\n"
"in s_uri varchar, in s_iid IRI_ID,\n"
"in p_uri varchar, in p_iid IRI_ID,\n"
"in o_val any, in o_type varchar, in o_lang varchar,\n"
"inout app_env any )\n"
"{\n"
"declare err any;\n"
"app_env[1] := aq_request (app_env[0], \'DB.DBA.TTLP_EXEC_TRIPLE_W_L\', vector  (g_iid, s_iid, p_iid, o_val, o_type, o_lang));\n"
"if (mod (app_env[1], 4000) = 0)\n"
"{\n"
"commit work;\n"
"aq_wait (app_env[0], app_env[1], err, 1);\n"
"aq_wait_all (app_env[0]);\n"
"}\n"
"}\n"
"--src sparql.sql:5666\n";

static const char *proc169 = 
"create procedure DB.DBA.TTLP_MT (in strg varchar, in base varchar, in graph varchar := null, in flags integer := 0)\n"
"{\n"
"declare app_env, err any;\n"
"app_env := vector (async_queue (3), 0);\n"
"rdf_load_turtle (strg, base, graph, flags,\n"
"vector (\n"
"\'DB.DBA.TTLP_EXEC_NEW_GRAPH(?,?)\',\n"
"\'select DB.DBA.TTLP_EXEC_NEW_BLANK(?,?)\',\n"
"\'select iri_to_id(cast (? as varchar), 1), ?, ?\',\n"
"\'DB.DBA.TTLP_EXEC_TRIPLE_A(?,?, ?,?, ?,?, ?,?, ?)\',\n"
"\'DB.DBA.TTLP_EXEC_TRIPLE_L_A(?,?, ?,?, ?,?, ?,?,?, ?)\',\n"
"\'commit work\' ),\n"
"app_env);\n"
"commit work;\n"
"aq_wait (app_env[0], app_env[1], err, 1);\n"
"aq_wait_all (app_env[0]);\n"
"}\n"
"--src sparql.sql:5685\n";

static const char *proc170 = 
"create procedure DB.DBA.RDF_LOAD_RDFXML_MT (in strg varchar, in base varchar, in graph varchar)\n"
"{\n"
"declare app_env, err any;\n"
"app_env := vector (async_queue (3), 0);\n"
"rdf_load_rdfxml (strg, 0,\n"
"graph,\n"
"vector (\n"
"\'DB.DBA.TTLP_EXEC_NEW_GRAPH(?,?)\',\n"
"\'select DB.DBA.TTLP_EXEC_NEW_BLANK(?,?)\',\n"
"\'select iri_to_id (cast (? as varchar), 1), ?, ?\',\n"
"\'DB.DBA.TTLP_EXEC_TRIPLE_A(?,?, ?,?, ?,?, ?,?, ?)\',\n"
"\'DB.DBA.TTLP_EXEC_TRIPLE_L_A(?,?, ?,?, ?,?, ?,?,?, ?)\',\n"
"\'commit work\' ),\n"
"app_env,\n"
"base );\n"
"commit work;\n"
"aq_wait (app_env[0], app_env[1], err, 1);\n"
"aq_wait_all (app_env[0]);\n"
"\n"
"return graph;\n"
"}\n"
"--src sparql.sql:5704\n";

static const char *tbl6 = 
"create table DB.DBA.SYS_HTTP_SPONGE (\n"
"HS_LOCAL_IRI varchar not null,\n"
"HS_PARSER varchar not null,\n"
"HS_ORIGIN_URI varchar not null,\n"
"HS_ORIGIN_LOGIN varchar,\n"
"HS_LAST_LOAD datetime,\n"
"HS_LAST_ETAG varchar,\n"
"HS_LAST_READ datetime,\n"
"HS_EXP_IS_TRUE integer,\n"
"HS_EXPIRATION datetime,\n"
"HS_LAST_MODIFIED datetime,\n"
"HS_DOWNLOAD_SIZE integer,\n"
"HS_DOWNLOAD_MSEC_TIME integer,\n"
"HS_READ_COUNT integer,\n"
"HS_SQL_STATE varchar,\n"
"HS_SQL_MESSAGE varchar,\n"
"HS_QUALITY double precision,\n"
"primary key (HS_LOCAL_IRI, HS_PARSER)\n"
")\n"
"create index SYS_HTTP_SPONGE_EXPIRATION on DB.DBA.SYS_HTTP_SPONGE (HS_EXPIRATION desc)\n";

static const char *proc171 = 
"create function DB.DBA.SYS_HTTP_SPONGE_UP (in local_iri varchar, in parser varchar, in eraser varchar, in options any)\n"
"{\n"
"declare new_origin_uri, new_origin_login, new_last_etag varchar;\n"
"declare old_origin_uri, old_origin_login, old_last_etag varchar;\n"
"declare new_last_load, new_expiration datetime;\n"
"declare old_last_load, old_expiration, old_last_modified datetime;\n"
"declare load_begin_msec, load_end_msec, old_exp_is_true,\n"
"old_download_size, old_download_msec_time, old_read_count,\n"
"new_download_size, explicit_refresh integer;\n"
"declare get_method varchar;\n"
"declare ret_hdr any;\n"
"declare req_hdr varchar;\n"
"declare ret_body, ret_content_type, ret_etag, ret_last_modified, ret_date, ret_last_modif, ret_expires varchar;\n"
"declare get_proxy varchar;\n"
"declare ret_dt_date, ret_dt_last_modified, ret_dt_expires datetime;\n"
"declare ret_304_not_modified integer;\n"
"\n"
"new_origin_uri := cast (get_keyword_ucase (\'get:uri\', options, local_iri) as varchar);\n"
"new_origin_login := cast (get_keyword_ucase (\'get:login\', options) as varchar);\n"
"explicit_refresh := get_keyword_ucase (\'get:refresh\', options);\n"
"set isolation=\'serializable\';\n"
"whenever not found goto add_new_origin;\n"
"select HS_ORIGIN_URI, HS_ORIGIN_LOGIN, HS_LAST_LOAD, HS_LAST_ETAG,\n"
"HS_EXP_IS_TRUE, HS_EXPIRATION, HS_LAST_MODIFIED,\n"
"HS_DOWNLOAD_SIZE, HS_DOWNLOAD_MSEC_TIME, HS_READ_COUNT\n"
"into old_origin_uri, old_origin_login, old_last_load, old_last_etag,\n"
"old_exp_is_true, old_expiration, old_last_modified,\n"
"old_download_size, old_download_msec_time, old_read_count\n"
"from DB.DBA.SYS_HTTP_SPONGE where HS_LOCAL_IRI = local_iri and HS_PARSER = parser;\n"
"if (new_origin_uri <> old_origin_uri)\n"
"signal (\'RDFXX\',\n"
"sprintf (\'Can not get-and-cache RDF graph <%.500s> from <%.500s> because is has been loaded from <%.500s>\',\n"
"local_iri, new_origin_uri, old_origin_uri) );\n"
"if (coalesce (new_origin_login, \'\') <> coalesce (old_origin_login, \'\') and\n"
"old_expiration is not null )\n"
"signal (\'RDFXX\',\n"
"sprintf (\'Can not get-and-cache RDF graph <%.500s> from <%.500s> using %s because is has been loaded using %s\',\n"
"local_iri, new_origin_uri,\n"
"case (isnull (new_origin_login)) when 0 then sprintf (\'login \"%.100s\"\', new_origin_login) else \'anonymous access\' end,\n"
"case (isnull (old_origin_login)) when 0 then sprintf (\'login \"%.100s\"\', old_origin_login) else \'anonymous access\' end ) );\n"
"\n"
"\n"
"if (old_expiration is not null)\n"
"{\n"
"if ((old_expiration >= now()) and (\n"
"explicit_refresh is null or\n"
"old_exp_is_true or\n"
"(dateadd (\'second\', explicit_refresh, old_last_load) >= now()) ) )\n"
"{\n"
"\n"
"update DB.DBA.SYS_HTTP_SPONGE\n"
"set HS_LAST_READ = now(), HS_READ_COUNT = old_read_count + 1\n"
"where HS_LOCAL_IRI = local_iri and HS_LAST_READ < now();\n"
"return local_iri;\n"
"}\n"
"}\n"
"else\n"
"{\n"
"if (old_last_load >= now() and old_expiration is null)\n"
"{\n"
"\n"
"return local_iri;\n"
"}\n"
"}\n"
"\n"
"update_old_origin:\n"
"\n"
"update DB.DBA.SYS_HTTP_SPONGE\n"
"set HS_LAST_LOAD = now(), HS_LAST_ETAG = NULL, HS_LAST_READ = NULL,\n"
"HS_EXP_IS_TRUE = 0, HS_EXPIRATION = NULL, HS_LAST_MODIFIED = NULL,\n"
"HS_DOWNLOAD_SIZE = NULL, HS_DOWNLOAD_MSEC_TIME = NULL,\n"
"HS_READ_COUNT = 0,\n"
"HS_SQL_STATE = NULL, HS_SQL_MESSAGE = NULL\n"
"where\n"
"HS_LOCAL_IRI = local_iri and HS_PARSER = parser;\n"
"goto perform_actual_load;\n"
"\n"
"add_new_origin:\n"
"\n"
"old_origin_uri := NULL; old_origin_login := NULL; old_last_load := NULL; old_last_etag := NULL;\n"
"old_expiration := NULL; old_download_size := NULL; old_download_msec_time := NULL;\n"
"old_exp_is_true := 0; old_read_count := 0;\n"
"insert into DB.DBA.SYS_HTTP_SPONGE (HS_LOCAL_IRI, HS_PARSER, HS_ORIGIN_URI, HS_ORIGIN_LOGIN, HS_LAST_LOAD)\n"
"values (local_iri, parser, new_origin_uri, new_origin_login, now());\n"
"goto perform_actual_load;\n"
"\n"
"perform_actual_load:\n"
"new_expiration := NULL;\n"
"new_last_etag := NULL;\n"
"ret_304_not_modified := 0;\n"
"load_begin_msec := msec_time();\n"
"set isolation=\'committed\';\n"
"\n"
"get_method := cast (get_keyword_ucase (\'get:method\', options, \'GET\') as varchar);\n"
"if (get_method in (\'POST\', \'GET\'))\n"
"{\n"
"req_hdr := NULL;\n"
"get_proxy := get_keyword_ucase (\'get:proxy\', options);\n"
"\n"
"\n"
"if (old_last_etag is not null)\n"
"req_hdr := \'If-None-Match: \' || old_last_etag;\n"
"\n"
"ret_body := http_get (new_origin_uri, ret_hdr, get_method, req_hdr, NULL, get_proxy);\n"
"\n"
"if (ret_hdr[0] like \'HTTP%404%\')\n"
"signal (\'HT404\', sprintf (\'Resource \"%.1000s\" not found\', new_origin_uri));\n"
"if (ret_hdr[0] like \'HTTP%304%\')\n"
"{\n"
"ret_304_not_modified := 1;\n"
"goto resp_received;\n"
"}\n"
"goto resp_received;\n"
"}\n"
"\n"
"call (eraser) (local_iri, new_origin_uri, options);\n"
"signal (\'RDFZZ\', sprintf (\n"
"\'Unable to get data from \"%.1000s\": This version of Virtuoso does not support OPTION (get:method \"%.100s\")\',\n"
"new_origin_uri, get_method ) );\n"
"\n"
"resp_received:\n"
"ret_content_type := http_request_header (ret_hdr, \'Content-Type\', null, null);\n"
"ret_etag := http_request_header (ret_hdr, \'ETag\', null, null);\n"
"ret_date := http_request_header (ret_hdr, \'Date\', null, null);\n"
"ret_expires := http_request_header (ret_hdr, \'Expires\', null, null);\n"
"ret_last_modif := http_request_header (ret_hdr, \'Last-Modified\', null, null);\n"
"ret_dt_date := http_string_date (ret_date, NULL, NULL);\n"
"ret_dt_expires := http_string_date (ret_expires, NULL, now());\n"
"ret_dt_last_modified := http_string_date (ret_last_modif, NULL, now());\n"
"if (http_request_header (ret_hdr, \'Pragma\', null, null) = \'no-cache\' or\n"
"http_request_header (ret_hdr, \'Cache-Control\', null, null) like \'no-cache%\' )\n"
"ret_dt_expires := now ();\n"
"if (ret_304_not_modified and ret_dt_last_modified is null)\n"
"ret_dt_last_modified := old_last_modified;\n"
"if (ret_dt_date is not null)\n"
"{\n"
"if (ret_dt_expires is not null)\n"
"ret_dt_expires := dateadd (\'second\', datediff (\'second\', ret_dt_date, now()), ret_dt_expires);\n"
"if (ret_dt_last_modified is not null)\n"
"ret_dt_last_modified := dateadd (\'second\', datediff (\'second\', ret_dt_date, now()), ret_dt_last_modified);\n"
"}\n"
"if (ret_dt_expires is not null and\n"
"(ret_dt_expires < coalesce (ret_dt_date, ret_dt_last_modified, now ())) )\n"
"ret_dt_expires := NULL;\n"
"if (ret_dt_expires is not null)\n"
"new_expiration := ret_dt_expires;\n"
"else\n"
"{\n"
"if (ret_dt_date is not null and ret_dt_last_modified is not null and (ret_dt_date >= ret_dt_last_modified))\n"
"new_expiration := dateadd (\'second\', __min (3600 * 24 * 7, 0.7 * datediff (\'second\', ret_dt_last_modified, ret_dt_date)), now());\n"
"}\n"
"if (ret_304_not_modified)\n"
"{\n"
"if (new_expiration is null and explicit_refresh is not null)\n"
"new_expiration := dateadd (\'second\', 0.7 * explicit_refresh, now());\n"
"if (ret_dt_expires is null and new_expiration is not null and explicit_refresh is not null)\n"
"new_expiration := __min (new_expiration, dateadd (\'second\', explicit_refresh, now()));\n"
"update DB.DBA.SYS_HTTP_SPONGE\n"
"set HS_LAST_LOAD = now(), HS_LAST_ETAG = old_last_etag, HS_LAST_READ = now(),\n"
"HS_EXP_IS_TRUE = case (isnull (ret_dt_expires)) when 1 then 0 else 1 end,\n"
"HS_EXPIRATION = coalesce (ret_dt_expires, new_expiration, now()),\n"
"HS_LAST_MODIFIED = coalesce (old_last_modified, ret_dt_last_modified),\n"
"HS_DOWNLOAD_SIZE = old_download_size,\n"
"HS_DOWNLOAD_MSEC_TIME = old_download_msec_time,\n"
"HS_READ_COUNT = old_read_count + 1,\n"
"HS_SQL_STATE = NULL, HS_SQL_MESSAGE = NULL\n"
"where\n"
"HS_LOCAL_IRI = local_iri;\n"
"return local_iri;\n"
"}\n"
"if (ret_body is null)\n"
"signal (\'RDFXX\', sprintf (\'Unable to retrieve RDF data from \"%.500s\": %.500s\', new_origin_uri, ret_hdr[0]));\n"
"\n"
"new_download_size := length (ret_body);\n"
"\n"
"whenever sqlstate \'*\' goto error_during_load;\n"
"call (parser) (local_iri, new_origin_uri, ret_content_type, ret_hdr, ret_body, options);\n"
"new_last_etag := ret_etag;\n"
"load_end_msec := msec_time();\n"
"if (new_expiration is null)\n"
"new_expiration := dateadd (\'second\', load_end_msec - load_begin_msec, now());\n"
"if (ret_dt_expires is null and explicit_refresh is not null)\n"
"new_expiration := __min (new_expiration, dateadd (\'second\', 0.7 * explicit_refresh, now()));\n"
"update DB.DBA.SYS_HTTP_SPONGE\n"
"set HS_LAST_LOAD = now(), HS_LAST_ETAG = new_last_etag, HS_LAST_READ = now(),\n"
"HS_EXP_IS_TRUE = case (isnull (ret_dt_expires)) when 1 then 0 else 1 end,\n"
"HS_EXPIRATION = coalesce (ret_dt_expires, new_expiration, now()),\n"
"HS_LAST_MODIFIED = ret_dt_last_modified,\n"
"HS_DOWNLOAD_SIZE = new_download_size,\n"
"HS_DOWNLOAD_MSEC_TIME = load_end_msec - load_begin_msec,\n"
"HS_READ_COUNT = 1,\n"
"HS_SQL_STATE = NULL, HS_SQL_MESSAGE = NULL\n"
"where\n"
"HS_LOCAL_IRI = local_iri and HS_PARSER = parser;\n"
"return local_iri;\n"
"\n"
"error_during_load:\n"
"\n"
"update DB.DBA.SYS_HTTP_SPONGE\n"
"set HS_SQL_STATE = __SQL_STATE,\n"
"HS_SQL_MESSAGE = __SQL_MESSAGE\n"
"where\n"
"HS_LOCAL_IRI = local_iri and HS_PARSER = parser;\n"
"return local_iri;\n"
"}\n"
"--src sparql.sql:5753\n";

static const char *proc172 = 
"create function DB.DBA.RDF_SPONGE_GUESS_CONTENT_TYPE (in origin_uri varchar, in ret_content_type varchar, inout ret_body any) returns varchar\n"
"{\n"
"if (ret_content_type is not null)\n"
"{\n"
"if (strstr (ret_content_type, \'application/sparql-results+xml\') is not null)\n"
"return \'application/sparql-results+xml\';\n"
"if (strstr (ret_content_type, \'application/rdf+xml\') is not null)\n"
"return \'application/rdf+xml\';\n"
"if (strstr (ret_content_type, \'text/rdf+n3\') is not null)\n"
"return \'text/rdf+n3\';\n"
"}\n"
"declare ret_begin, ret_html any;\n"
"ret_begin := \"LEFT\" (ret_body, 1024);\n"
"ret_html := xtree_doc (ret_begin, 2);\n"
"if (xpath_eval (\'/html|/xhtml\', ret_html) is not null)\n"
"return \'text/html\';\n"
"if (xpath_eval (\'[xmlns:rset=\"http://www.w3.org/2005/sparql-results#\"] /rset:sparql\', ret_html) is not null\n"
"or xpath_eval (\'[xmlns:rset2=\"http://www.w3.org/2001/sw/DataAccess/rf1/result2\"] /rset2:sparql\', ret_html) is not null)\n"
"return \'application/sparql-results+xml\';\n"
"if (xpath_eval (\'[xmlns:rdf=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\"] /rdf:rdf\', ret_html) is not null)\n"
"return \'application/rdf+xml\';\n"
"if (strstr (ret_begin, \'<html>\') is not null or\n"
"strstr (ret_begin, \'<xhtml>\') is not null )\n"
"return \'text/html\';\n"
"if (ret_content_type is null or\n"
"strstr (ret_content_type, \'text/plain\') is not null or\n"
"strstr (ret_content_type, \'application/octet-stream\') is not null )\n"
"{\n"
"declare ret_lines any;\n"
"declare ret_lcount, ret_lctr integer;\n"
"ret_lines := split_and_decode (ret_begin, 0, \'\\0\\t\\n\');\n"
"ret_lcount := length (ret_lines);\n"
"for (ret_lctr := 0; ret_lctr < ret_lcount; ret_lctr := ret_lctr + 1)\n"
"{\n"
"declare l varchar;\n"
"l := rtrim (replace (ret_lines [ret_lctr], \'\\r\', \'\'));\n"
"\n"
"if ((\"LEFT\" (l, 7) = \'@prefix\') or (\"LEFT\" (l, 5) = \'@base\') or (\"LEFT\" (l, 8) = \'@keyword\'))\n"
"return \'text/rdf+n3\';\n"
"if (((\"LEFT\" (l, 1) = \'<\') or (\"LEFT\" (l, 1) = \'[\')) and (\"RIGHT\" (origin_uri, 4) in (\'.ttl\', \'.TTL\', \'.n3\', \'.N3\')))\n"
"return \'text/rdf+n3\';\n"
"if (not ((l like \'#%\') or (l=\'\')))\n"
"return \'text/plain\';\n"
"}\n"
"}\n"
"return ret_content_type;\n"
"}\n"
"--src sparql.sql:5960\n";

static const char *proc173 = 
"create procedure DB.DBA.RDF_LOAD_HTTP_RESPONSE (in graph_iri varchar, in new_origin_uri varchar, inout ret_content_type varchar, inout ret_hdr any, inout ret_body any, inout options any)\n"
"{\n"
"\n"
"ret_content_type := DB.DBA.RDF_SPONGE_GUESS_CONTENT_TYPE (new_origin_uri, ret_content_type, ret_body);\n"
"\n"
"if (strstr (ret_content_type, \'application/sparql-results+xml\') is not null)\n"
"signal (\'RDFXX\', sprintf (\'Unable to load RDF graph <%.500s> from <%.500s>: the sparql-results XML answer does not contain triples\', graph_iri, new_origin_uri));\n"
"if (strstr (ret_content_type, \'application/rdf+xml\') is not null)\n"
"{\n"
"delete from DB.DBA.RDF_QUAD where G = DB.DBA.RDF_MAKE_IID_OF_QNAME (graph_iri);\n"
"DB.DBA.RDF_LOAD_RDFXML (ret_body, new_origin_uri, graph_iri);\n"
"return;\n"
"}\n"
"if (strstr (ret_content_type, \'text/rdf+n3\') is not null)\n"
"{\n"
"delete from DB.DBA.RDF_QUAD where G = DB.DBA.RDF_MAKE_IID_OF_QNAME (graph_iri);\n"
"DB.DBA.TTLP (ret_body, new_origin_uri, graph_iri);\n"
"return;\n"
"}\n"
"delete from DB.DBA.RDF_QUAD where G = DB.DBA.RDF_MAKE_IID_OF_QNAME (graph_iri);\n"
"if (strstr (ret_content_type, \'text/plain\') is not null)\n"
"{\n"
"signal (\'RDFXX\', sprintf (\n"
"\'Unable to load RDF graph <%.500s> from <%.500s>: returned Content-Type \'\'%.300s\'\' status \'\'%.300s\'\'\\n%.1000s\',\n"
"graph_iri, new_origin_uri, ret_content_type, ret_hdr[0], \"LEFT\" (ret_body, 1024) ) );\n"
"}\n"
"if (strstr (ret_content_type, \'text/html\') is not null)\n"
"{\n"
"signal (\'RDFZZ\', sprintf (\n"
"\'Unable to load RDF graph <%.500s> from <%.500s>: returned Content-Type \'\'%.300s\'\' status \'\'%.300s\'\'\\n%.1000s\',\n"
"graph_iri, new_origin_uri, ret_content_type, ret_hdr[0],\n"
"cast (xtree_doc (ret_body, 2) as varchar) ) );\n"
"}\n"
"signal (\'RDFZZ\', sprintf (\n"
"\'Unable to load RDF graph <%.500s> from <%.500s>: returned unsupported Content-Type \'\'%.300s\'\'\',\n"
"graph_iri, new_origin_uri, ret_content_type ) );\n"
"}\n"
"--src sparql.sql:6009\n";

static const char *proc174 = 
"create procedure DB.DBA.RDF_FORGET_HTTP_RESPONSE (in graph_iri varchar, in new_origin_uri varchar, inout options any)\n"
"{\n"
"delete from DB.DBA.RDF_QUAD where G = DB.DBA.RDF_MAKE_IID_OF_QNAME (graph_iri);\n"
"}\n"
"--src sparql.sql:6048\n";

static const char *proc175 = 
"create function DB.DBA.RDF_SPONGE_UP (in graph_iri varchar, in options any)\n"
"{\n"
"declare get_soft varchar;\n"
"get_soft := get_keyword (\'get:soft\', options);\n"
"if (\'soft\' = get_soft)\n"
"{\n"
"if (\n"
"exists (select top 1 1 from DB.DBA.RDF_QUAD\n"
"where G = DB.DBA.RDF_MAKE_IID_OF_QNAME (graph_iri) ) and\n"
"not exists (select top 1 1 from DB.DBA.SYS_HTTP_SPONGE\n"
"where HS_LOCAL_IRI = graph_iri and HS_PARSER = \'DB.DBA.RDF_LOAD_HTTP_RESPONSE\' ) )\n"
"return graph_iri;\n"
"}\n"
"return DB.DBA.SYS_HTTP_SPONGE_UP (graph_iri, \'DB.DBA.RDF_LOAD_HTTP_RESPONSE\', \'DB.DBA.RDF_FORGET_HTTP_RESPONSE\', options);\n"
"}\n"
"--src sparql.sql:6054\n";

static const char *proc176 = 
"create procedure DB.DBA.SPARQL_RELOAD_QM_GRAPH ()\n"
"{\n"
"if (not exists (sparql define input:storage \"\" ask where {\n"
"graph <http://www.openlinksw.com/schemas/virtrdf#> {\n"
"<http://www.openlinksw.com/sparql/virtrdf-data-formats.ttl>\n"
"virtrdf:version \'2006-11-12 0001\'\n"
"} } ) )\n"
"{\n"
"declare txt1, txt2 varchar;\n"
"declare jso_sys_g_iid IRI_ID;\n"
"declare dict1, lst1, dict2, lst2, sum_lst any;\n"
"txt1 := cast ( XML_URI_GET (\n"
"\'http://www.openlinksw.com/sparql/virtrdf-data-formats.ttl\', \'\' ) as varchar );\n"
"txt2 := \'\n"
"@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .\n"
"@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .\n"
"@prefix owl: <http://www.w3.org/2002/07/owl#> .\n"
"@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .\n"
"@prefix virtrdf: <http://www.openlinksw.com/schemas/virtrdf#> .\n"
"@prefix rdfdf: <http://www.openlinksw.com/virtrdf-data-formats#> .\n"
"\n"
"virtrdf:DefaultQuadStorage\n"
"rdf:type virtrdf:QuadStorage ;\n"
"virtrdf:qsUserMaps virtrdf:DefaultQuadStorage-UserMaps ;\n"
"virtrdf:qsDefaultMap virtrdf:DefaultQuadMap .\n"
"virtrdf:DefaultQuadStorage-UserMaps\n"
"rdf:type virtrdf:array-of-QuadMap .\n"
"\';\n"
"jso_sys_g_iid := DB.DBA.RDF_MAKE_IID_OF_QNAME (JSO_SYS_GRAPH ());\n"
"dict1 := DB.DBA.RDF_TTL2HASH (txt1, \'\');\n"
"dict2 := DB.DBA.RDF_TTL2HASH (txt2, \'\');\n"
"lst1 := dict_list_keys (dict1, 1);\n"
"lst2 := dict_list_keys (dict2, 1);\n"
"sum_lst := vector_concat (lst1, lst2);\n"
"foreach (any triple in sum_lst) do\n"
"{\n"
"delete from DB.DBA.RDF_QUAD where G = jso_sys_g_iid and S = triple[0] and P = triple[1];\n"
"}\n"
"foreach (any triple in sum_lst) do\n"
"{\n"
"insert into DB.DBA.RDF_QUAD (G,S,P,O) values (jso_sys_g_iid, triple[0], triple[1], DB.DBA.RDF_OBJ_OF_LONG (triple[2]));\n"
"}\n"
"commit work;\n"
"}\n"
"JSO_LOAD_GRAPH (JSO_SYS_GRAPH (), 0);\n"
"JSO_PIN_GRAPH (JSO_SYS_GRAPH ());\n"
"}\n"
"--src sparql.sql:6075\n";

static const char *proc177 = 
"create procedure DB.DBA.RDF_CREATE_SPARQL_ROLES ()\n"
"{\n"
"declare state, msg varchar;\n"
"declare cmds any;\n"
"cmds := vector (\n"
"\'create role SPARQL_SELECT\',\n"
"\'create role SPARQL_UPDATE\',\n"
"\'grant SPARQL_SELECT to SPARQL_UPDATE\',\n"
"\'grant select on DB.DBA.RDF_QUAD to SPARQL_SELECT\',\n"
"\'grant all on DB.DBA.RDF_QUAD to SPARQL_UPDATE\',\n"
"\'grant select on DB.DBA.RDF_URL to SPARQL_SELECT\',\n"
"\'grant all on DB.DBA.RDF_URL to SPARQL_UPDATE\',\n"
"\'grant select on DB.DBA.RDF_PREFIX to SPARQL_SELECT\',\n"
"\'grant all on DB.DBA.RDF_PREFIX to SPARQL_UPDATE\',\n"
"\'grant select on DB.DBA.RDF_IRI to SPARQL_SELECT\',\n"
"\'grant all on DB.DBA.RDF_IRI to SPARQL_UPDATE\',\n"
"\'grant select on DB.DBA.RDF_OBJ to SPARQL_SELECT\',\n"
"\'grant all on DB.DBA.RDF_OBJ to SPARQL_UPDATE\',\n"
"\'grant select on DB.DBA.RDF_DATATYPE to SPARQL_SELECT\',\n"
"\'grant all on DB.DBA.RDF_DATATYPE to SPARQL_UPDATE\',\n"
"\'grant select on DB.DBA.RDF_LANGUAGE to SPARQL_SELECT\',\n"
"\'grant all on DB.DBA.RDF_LANGUAGE to SPARQL_UPDATE\',\n"
"\'grant select on DB.DBA.SYS_SPARQL_HOST to SPARQL_SELECT\',\n"
"\'grant all on DB.DBA.SYS_SPARQL_HOST to SPARQL_UPDATE\',\n"
"\'grant execute on DB.DBA.RDF_GLOBAL_RESET to SPARQL_UPDATE\',\n"
"\'grant execute on DB.DBA.RDF_MAKE_IID_OF_QNAME to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_MAKE_IID_OF_QNAME_SAFE to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_MAKE_IID_OF_LONG to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_QNAME_OF_IID to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_IID_OF_QNAME_SAFE to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_IID_OF_QNAME to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_MAKE_GRAPH_IIDS_OF_QNAMES to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_TWOBYTE_OF_DATATYPE to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_TWOBYTE_OF_LANGUAGE to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_DATATYPE_OF_TAG to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RQ_LONG_OF_O to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RQ_SQLVAL_OF_O to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_BOOL_OF_O to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RQ_IID_OF_O to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RQ_O_IS_LIT to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_MAKE_RO_ID_OF_STRING to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_FIND_RO_ID_OF_STRING to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_MAKE_OBJ_OF_SQLVAL to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_MAKE_OBJ_OF_TYPEDSQLVAL to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_MAKE_OBJ_OF_TYPEDSQLVAL_STRINGS to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_LONG_OF_OBJ to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_DATATYPE_OF_OBJ to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_LANGUAGE_OF_OBJ to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_SQLVAL_OF_OBJ to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_BOOL_OF_OBJ to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_QNAME_OF_OBJ to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_STRSQLVAL_OF_OBJ to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_OBJ_OF_LONG to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_OBJ_OF_SQLVAL to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_MAKE_LONG_OF_SQLVAL to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_MAKE_LONG_OF_TYPEDSQLVAL to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_MAKE_LONG_OF_TYPEDSQLVAL_STRINGS to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_SQLVAL_OF_LONG to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_BOOL_OF_LONG to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_DATATYPE_OF_LONG to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_LANGUAGE_OF_LONG to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_STRSQLVAL_OF_LONG to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_LONG_OF_SQLVAL to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_DATATYPE_OF_SQLVAL to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_QUAD_URI to SPARQL_UPDATE\',\n"
"\'grant execute on DB.DBA.RDF_QUAD_URI_L to SPARQL_UPDATE\',\n"
"\'grant execute on DB.DBA.RDF_QUAD_URI_L_TYPED to SPARQL_UPDATE\',\n"
"\'grant execute on DB.DBA.TTLP_EXEC_NEW_GRAPH to SPARQL_UPDATE\',\n"
"\'grant execute on DB.DBA.TTLP_EXEC_NEW_BLANK to SPARQL_UPDATE\',\n"
"\'grant execute on DB.DBA.TTLP_EXEC_GET_IID to SPARQL_UPDATE\',\n"
"\'grant execute on DB.DBA.TTLP_EXEC_TRIPLE to SPARQL_UPDATE\',\n"
"\'grant execute on DB.DBA.TTLP_EXEC_TRIPLE_L to SPARQL_UPDATE\',\n"
"\'grant execute on DB.DBA.TTLP to SPARQL_UPDATE\',\n"
"\'grant execute on DB.DBA.RDF_TTL2HASH_EXEC_NEW_GRAPH to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_TTL2HASH_EXEC_NEW_BLANK to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_TTL2HASH_EXEC_GET_IID to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_TTL2HASH_EXEC_TRIPLE to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_TTL2HASH_EXEC_TRIPLE_L to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_TTL2HASH to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_LOAD_RDFXML to SPARQL_UPDATE\',\n"
"\'grant execute on DB.DBA.RDF_RDFXML_TO_DICT to SPARQL_UPDATE\',\n"
"\'grant execute on DB.DBA.RDF_LONG_TO_TTL to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_TRIPLES_TO_TTL to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_TRIPLES_TO_RDF_XML_TEXT to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_FORMAT_RESULT_SET_AS_TTL_INIT to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_FORMAT_RESULT_SET_AS_TTL_ACC to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_FORMAT_RESULT_SET_AS_TTL_FIN to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_FORMAT_RESULT_SET_AS_RDF_XML_INIT to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_FORMAT_RESULT_SET_AS_RDF_XML_ACC to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_FORMAT_RESULT_SET_AS_RDF_XML_FIN to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_FORMAT_TRIPLE_DICT_AS_TTL to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_FORMAT_TRIPLE_DICT_AS_RDF_XML to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_INSERT_TRIPLES to SPARQL_UPDATE\',\n"
"\'grant execute on DB.DBA.RDF_DELETE_TRIPLES to SPARQL_UPDATE\',\n"
"\'grant execute on DB.DBA.SPARQL_INSERT_DICT_CONTENT to SPARQL_UPDATE\',\n"
"\'grant execute on DB.DBA.SPARQL_DELETE_DICT_CONTENT to SPARQL_UPDATE\',\n"
"\'grant execute on DB.DBA.SPARQL_DESCRIBE_INIT to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.SPARQL_DESCRIBE_ACC to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.SPARQL_DESCRIBE_FIN to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_DESCRIBE_PUT to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_TYPEMIN_OF_OBJ to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_TYPEMAX_OF_OBJ to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_IID_CMP to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_OBJ_CMP to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.RDF_LONG_CMP to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.SPARQL_EVAL_TO_ARRAY to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.SPARQL_EVAL to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.SPARQL_REXEC to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.SPARQL_REXEC_TO_ARRAY to SPARQL_SELECT\',\n"
"\'grant execute on DB.DBA.SPARQL_REXEC_WITH_META to SPARQL_SELECT\',\n"
"\'grant execute on WS.WS.\"/!sparql/\" to \"SPARQL\"\',\n"
"\'grant execute on DB.DBA.TTLP_MT  to SPARQL_UPDATE\',\n"
"\'grant execute on DB.DBA.TTLP_EXEC_TRIPLE_W  to SPARQL_UPDATE\',\n"
"\'grant execute on DB.DBA.TTLP_EXEC_TRIPLE_A  to SPARQL_UPDATE\',\n"
"\'grant execute on DB.DBA.TTLP_EXEC_TRIPLE_L_A  to SPARQL_UPDATE\',\n"
"\'grant execute on DB.DBA.RDF_LOAD_RDFXML_MT to SPARQL_UPDATE\',\n"
"\'grant execute on DB.DBA.RDF_LOAD_HTTP_RESPONSE to SPARQL_UPDATE\',\n"
"\'grant execute on DB.DBA.RDF_FORGET_HTTP_RESPONSE to SPARQL_UPDATE\',\n"
"\'grant execute on DB.DBA.RDF_SPONGE_UP to SPARQL_UPDATE\' );\n"
"foreach (varchar cmd in cmds) do\n"
"{\n"
"exec (cmd, state, msg);\n"
"}\n"
"}\n"
"--src sparql.sql:6124\n";


static const char *other14 = 
" DB.DBA.RDF_CREATE_SPARQL_ROLES ()\n";

static const char *other15 = 
" DB.DBA.SPARQL_RELOAD_QM_GRAPH ()\n";
static int
sch_proc_def_exists (client_connection_t *cli, const char *proc_name)
{
  query_t *proc = NULL;
  char *full_name = sch_full_proc_name (isp_schema(NULL), proc_name,
	cli->cli_qualifier, CLI_OWNER (cli));
  if (full_name)
    proc = sch_proc_def (isp_schema(NULL), full_name);
  return (proc != NULL);
}

#define DEFINE_PROC(name, proc) \
   if (!sch_proc_def_exists (bootstrap_cli, (name))) \
     ddl_std_proc_1 (proc, 0x0, 1)


#define DEFINE_PUBLIC_PROC(name, proc) \
   if (!sch_proc_def_exists (bootstrap_cli, (name))) \
     ddl_std_proc_1 (proc, 0x1, 1)


#define DEFINE_OVERWRITE_PROC(name, proc) \
   ddl_std_proc_1 (proc, 0x1, 1)


void
sqls_define_sparql (void)
{
  /* sparql.sql */

  ddl_ensure_table ("DB.DBA.RDF_QUAD", tbl0);
  ddl_ensure_table ("DB.DBA.RDF_URL", tbl1);
  ddl_ensure_table ("DB.DBA.RDF_OBJ", tbl2);
  ddl_ensure_table ("DB.DBA.RDF_DATATYPE", tbl3);
  ddl_ensure_table ("DB.DBA.RDF_LANGUAGE", tbl4);
  ddl_ensure_table ("DB.DBA.SYS_SPARQL_HOST", tbl5);
  ddl_ensure_table ("do this always", other0);
  ddl_ensure_table ("do this always", other1);
  ddl_ensure_table ("do this always", other2);
  ddl_ensure_table ("do this always", other3);
  ddl_ensure_table ("do this always", other4);
  ddl_ensure_table ("do this always", other5);
  DEFINE_PROC ("DB.DBA.RDF_GLOBAL_RESET", proc0);
  DEFINE_PROC ("DB.DBA.RDF_MAKE_IID_OF_QNAME", proc1);
  DEFINE_PROC ("DB.DBA.RDF_MIGRATE_URL_TO_IRI", proc2);
  ddl_ensure_table ("do this always", other6);
  DEFINE_PROC ("DB.DBA.RDF_MAKE_IID_OF_QNAME_SAFE", proc3);
  DEFINE_PROC ("DB.DBA.RDF_MAKE_IID_OF_LONG", proc4);
  DEFINE_PROC ("DB.DBA.RDF_QNAME_OF_IID", proc5);
  DEFINE_PROC ("DB.DBA.RDF_IID_OF_QNAME_SAFE", proc6);
  DEFINE_PROC ("DB.DBA.RDF_IID_OF_QNAME", proc7);
  DEFINE_PROC ("DB.DBA.RDF_MAKE_GRAPH_IIDS_OF_QNAMES", proc8);
  DEFINE_PROC ("DB.DBA.RDF_TWOBYTE_OF_DATATYPE", proc9);
  DEFINE_PROC ("DB.DBA.RDF_TWOBYTE_OF_LANGUAGE", proc10);
  DEFINE_PROC ("DB.DBA.RDF_DATATYPE_OF_TAG", proc11);
  DEFINE_PROC ("DB.DBA.RQ_LONG_OF_O", proc12);
  DEFINE_PROC ("DB.DBA.RQ_SQLVAL_OF_O", proc13);
  DEFINE_PROC ("DB.DBA.RQ_BOOL_OF_O", proc14);
  DEFINE_PROC ("DB.DBA.RQ_IID_OF_O", proc15);
  DEFINE_PROC ("DB.DBA.RQ_O_IS_LIT", proc16);
  DEFINE_PROC ("DB.DBA.RDF_MAKE_RO_ID_OF_STRING", proc17);
  DEFINE_PROC ("DB.DBA.RDF_FIND_RO_ID_OF_STRING", proc18);
  DEFINE_PROC ("DB.DBA.RDF_MAKE_OBJ_OF_SQLVAL", proc19);
  DEFINE_PROC ("DB.DBA.RDF_MAKE_OBJ_OF_TYPEDSQLVAL", proc20);
  DEFINE_PROC ("DB.DBA.RDF_MAKE_OBJ_OF_TYPEDSQLVAL_STRINGS", proc21);
  DEFINE_PROC ("DB.DBA.RDF_LONG_OF_OBJ", proc22);
  DEFINE_PROC ("DB.DBA.RDF_DATATYPE_OF_OBJ", proc23);
  DEFINE_PROC ("DB.DBA.RDF_LANGUAGE_OF_OBJ", proc24);
  DEFINE_PROC ("DB.DBA.RDF_SQLVAL_OF_OBJ", proc25);
  DEFINE_PROC ("DB.DBA.RDF_BOOL_OF_OBJ", proc26);
  DEFINE_PROC ("DB.DBA.RDF_QNAME_OF_OBJ", proc27);
  DEFINE_PROC ("DB.DBA.RDF_STRSQLVAL_OF_OBJ", proc28);
  DEFINE_PROC ("DB.DBA.RDF_OBJ_OF_LONG", proc29);
  DEFINE_PROC ("DB.DBA.RDF_OBJ_OF_SQLVAL", proc30);
  DEFINE_PROC ("DB.DBA.RDF_MAKE_LONG_OF_SQLVAL", proc31);
  DEFINE_PROC ("DB.DBA.RDF_MAKE_LONG_OF_TYPEDSQLVAL", proc32);
  DEFINE_PROC ("DB.DBA.RDF_MAKE_LONG_OF_TYPEDSQLVAL_STRINGS", proc33);
  DEFINE_PROC ("DB.DBA.RDF_SQLVAL_OF_LONG", proc34);
  DEFINE_PROC ("DB.DBA.RDF_BOOL_OF_LONG", proc35);
  DEFINE_PROC ("DB.DBA.RDF_DATATYPE_OF_LONG", proc36);
  DEFINE_PROC ("DB.DBA.RDF_LANGUAGE_OF_LONG", proc37);
  DEFINE_PROC ("DB.DBA.RDF_STRSQLVAL_OF_LONG", proc38);
  DEFINE_PROC ("DB.DBA.RDF_LONG_OF_SQLVAL", proc39); 
  DEFINE_PUBLIC_PROC ("DB.DBA.RDF_STRSQLVAL_OF_SQLVAL", proc40); 
  DEFINE_PUBLIC_PROC ("DB.DBA.RDF_DATATYPE_OF_SQLVAL", proc41); 
  DEFINE_PUBLIC_PROC ("DB.DBA.RDF_LANGUAGE_OF_SQLVAL", proc42); 
  DEFINE_PUBLIC_PROC ("DB.DBA.RDF_IS_BLANK_REF", proc43); 
  DEFINE_PUBLIC_PROC ("DB.DBA.RDF_IS_URI_REF", proc44); 
  DEFINE_PUBLIC_PROC ("DB.DBA.http://www.w3.org/2001/XMLSchema#boolean", proc45); 
  DEFINE_PUBLIC_PROC ("DB.DBA.http://www.w3.org/2001/XMLSchema#dateTime", proc46); 
  DEFINE_PUBLIC_PROC ("DB.DBA.http://www.w3.org/2001/XMLSchema#double", proc47); 
  DEFINE_PUBLIC_PROC ("DB.DBA.http://www.w3.org/2001/XMLSchema#float", proc48); 
  DEFINE_PUBLIC_PROC ("DB.DBA.http://www.w3.org/2001/XMLSchema#integer", proc49); 
  DEFINE_PUBLIC_PROC ("DB.DBA.__and", proc50); 
  DEFINE_PUBLIC_PROC ("DB.DBA.__or", proc51); 
  DEFINE_PUBLIC_PROC ("DB.DBA.__not", proc52);
  DEFINE_PROC ("DB.DBA.RDF_QUAD_URI", proc53);
  DEFINE_PROC ("DB.DBA.RDF_QUAD_URI_L", proc54);
  DEFINE_PROC ("DB.DBA.RDF_QUAD_URI_L_TYPED", proc55);
  DEFINE_PROC ("DB.DBA.TTLP_EXEC_NEW_GRAPH", proc56);
  DEFINE_PROC ("DB.DBA.TTLP_EXEC_NEW_BLANK", proc57);
  DEFINE_PROC ("DB.DBA.TTLP_EXEC_GET_IID", proc58);
  DEFINE_PROC ("DB.DBA.TTLP_EXEC_TRIPLE", proc59);
  DEFINE_PROC ("DB.DBA.TTLP_EXEC_TRIPLE_L", proc60);
  DEFINE_PROC ("DB.DBA.TTLP", proc61);
  DEFINE_PROC ("DB.DBA.RDF_TTL2HASH_EXEC_NEW_GRAPH", proc62);
  DEFINE_PROC ("DB.DBA.RDF_TTL2HASH_EXEC_NEW_BLANK", proc63);
  DEFINE_PROC ("DB.DBA.RDF_TTL2HASH_EXEC_GET_IID", proc64);
  DEFINE_PROC ("DB.DBA.RDF_TTL2HASH_EXEC_TRIPLE", proc65);
  DEFINE_PROC ("DB.DBA.RDF_TTL2HASH_EXEC_TRIPLE_L", proc66);
  DEFINE_PROC ("DB.DBA.RDF_TTL2HASH", proc67);
  DEFINE_PROC ("DB.DBA.RDF_LOAD_RDFXML", proc68);
  DEFINE_PROC ("DB.DBA.RDF_RDFXML_TO_DICT", proc69);
  DEFINE_PROC ("DB.DBA.RDF_LONG_TO_TTL", proc70);
  DEFINE_PROC ("DB.DBA.RDF_TRIPLES_TO_TTL", proc71);
  DEFINE_PROC ("DB.DBA.RDF_TRIPLES_TO_RDF_XML_TEXT", proc72);
  DEFINE_PROC ("DB.DBA.RDF_FORMAT_RESULT_SET_AS_TTL_INIT", proc73);
  DEFINE_PROC ("DB.DBA.RDF_FORMAT_RESULT_SET_AS_TTL_ACC", proc74);
  DEFINE_PROC ("DB.DBA.RDF_FORMAT_RESULT_SET_AS_TTL_FIN", proc75);
  ddl_ensure_table ("do this always", other7);
  DEFINE_PROC ("DB.DBA.RDF_FORMAT_RESULT_SET_AS_RDF_XML_INIT", proc76);
  DEFINE_PROC ("DB.DBA.RDF_FORMAT_RESULT_SET_AS_RDF_XML_ACC", proc77);
  DEFINE_PROC ("DB.DBA.RDF_FORMAT_RESULT_SET_AS_RDF_XML_FIN", proc78);
  ddl_ensure_table ("do this always", other8);
  DEFINE_PROC ("DB.DBA.RDF_FORMAT_TRIPLE_DICT_AS_TTL", proc79);
  DEFINE_PROC ("DB.DBA.RDF_FORMAT_TRIPLE_DICT_AS_RDF_XML", proc80); 
  DEFINE_PUBLIC_PROC ("DB.DBA.RDF_FORMAT_BOOL_RESULT_AS_RDF_XML_INIT", proc81); 
  DEFINE_PUBLIC_PROC ("DB.DBA.RDF_FORMAT_BOOL_RESULT_AS_RDF_XML_ACC", proc82); 
  DEFINE_PUBLIC_PROC ("DB.DBA.RDF_FORMAT_BOOL_RESULT_AS_RDF_XML_FIN", proc83);
  ddl_ensure_table ("do this always", other9); 
  DEFINE_PUBLIC_PROC ("DB.DBA.RDF_FORMAT_BOOL_RESULT_AS_TTL_INIT", proc84); 
  DEFINE_PUBLIC_PROC ("DB.DBA.RDF_FORMAT_BOOL_RESULT_AS_TTL_ACC", proc85); 
  DEFINE_PUBLIC_PROC ("DB.DBA.RDF_FORMAT_BOOL_RESULT_AS_TTL_FIN", proc86);
  ddl_ensure_table ("do this always", other10);
  DEFINE_PROC ("DB.DBA.RDF_INSERT_TRIPLES", proc87);
  DEFINE_PROC ("DB.DBA.RDF_DELETE_TRIPLES", proc88);
  DEFINE_PROC ("DB.DBA.SPARQL_INSERT_DICT_CONTENT", proc89);
  DEFINE_PROC ("DB.DBA.SPARQL_DELETE_DICT_CONTENT", proc90); 
  DEFINE_PUBLIC_PROC ("DB.DBA.RDF_REGEX", proc91); 
  DEFINE_PUBLIC_PROC ("DB.DBA.RDF_LANGMATCHES", proc92); 
  DEFINE_PUBLIC_PROC ("DB.DBA.SPARQL_CONSTRUCT_INIT", proc93); 
  DEFINE_PUBLIC_PROC ("DB.DBA.SPARQL_CONSTRUCT_ACC", proc94); 
  DEFINE_PUBLIC_PROC ("DB.DBA.SPARQL_CONSTRUCT_FIN", proc95); 
  ddl_ensure_table ("do this always", other11);
  DEFINE_PUBLIC_PROC ("DB.DBA.SPARQL_DESCRIBE_INIT", proc96);
  DEFINE_PROC ("DB.DBA.SPARQL_DESCRIBE_ACC", proc97);
  DEFINE_PROC ("DB.DBA.SPARQL_DESCRIBE_FIN", proc98);
  ddl_ensure_table ("do this always", other12);
  DEFINE_PROC ("DB.DBA.RDF_DESCRIBE_PUT", proc99);
  DEFINE_PROC ("DB.DBA.RDF_TYPEMIN_OF_OBJ", proc100);
  DEFINE_PROC ("DB.DBA.RDF_TYPEMAX_OF_OBJ", proc101);
  DEFINE_PROC ("DB.DBA.RDF_IID_CMP", proc102);
  DEFINE_PROC ("DB.DBA.RDF_OBJ_CMP", proc103);
  DEFINE_PROC ("DB.DBA.RDF_LONG_CMP", proc104);
  DEFINE_PROC ("JSO_MAKE_INHERITANCE", proc105);
  DEFINE_PROC ("JSO_LOAD_INSTANCE", proc106);
  DEFINE_PROC ("JSO_LIST_INSTANCES_OF_GRAPH", proc107);
  DEFINE_PROC ("JSO_LOAD_GRAPH", proc108);
  DEFINE_PROC ("JSO_PIN_GRAPH", proc109);
  DEFINE_PROC ("JSO_SYS_GRAPH", proc110);
  DEFINE_PROC ("JSO_DUMP_IRI", proc111);
  DEFINE_PROC ("JSO_DUMP_FLD", proc112);
  DEFINE_PROC ("JSO_DUMP_ALL", proc113);
  DEFINE_PROC ("DB.DBA.RDF_QM_CHANGE", proc114);
  DEFINE_PROC ("DB.DBA.RDF_QM_APPLY_CHANGES", proc115);
  DEFINE_PROC ("DB.DBA.RDF_QM_ASSERT_JSO_TYPE", proc116);
  DEFINE_PROC ("DB.DBA.RDF_QM_ASSERT_STORAGE_FLAG", proc117);
  DEFINE_PROC ("DB.DBA.RDF_QM_ASSERT_STORAGE_CONTAINS_MAPPING", proc118);
  DEFINE_PROC ("DB.DBA.RDF_QM_ASSERT_STORAGE_IS_FLAGGED", proc119);
  DEFINE_PROC ("DB.DBA.RDF_QM_GC_SUBTREE", proc120);
  DEFINE_PROC ("DB.DBA.RDF_QM_DROP_MAPPING", proc121);
  DEFINE_PROC ("DB.DBA.RDF_QM_DEFINE_IRI_CLASS_FORMAT", proc122);
  DEFINE_PROC ("DB.DBA.RDF_QM_DEFINE_IRI_CLASS_FUNCTIONS", proc123);
  DEFINE_PROC ("DB.DBA.RDF_QM_DEFINE_SUBCLASS", proc124);
  DEFINE_PROC ("DB.DBA.RDF_QM_DROP_QUAD_STORAGE", proc125);
  DEFINE_PROC ("DB.DBA.RDF_QM_DEFINE_QUAD_STORAGE", proc126);
  DEFINE_PROC ("DB.DBA.RDF_QM_BEGIN_ALTER_QUAD_STORAGE", proc127);
  DEFINE_PROC ("DB.DBA.RDF_QM_END_ALTER_QUAD_STORAGE", proc128);
  DEFINE_PROC ("DB.DBA.RDF_QM_DEFINE_MAP_VALUE", proc129);
  DEFINE_PROC ("DB.DBA.RDF_QM_NORMALIZE_QMV", proc130);
  DEFINE_PROC ("DB.DBA.RDF_QM_DEFINE_MAPPING", proc131);
  DEFINE_PROC ("DB.DBA.RDF_QM_ATTACH_MAPPING", proc132);
  DEFINE_PROC ("DB.DBA.RDF_QM_ADD_MAPPING_TO_STORAGE", proc133);
  DEFINE_PROC ("DB.DBA.RDF_QM_DELETE_MAPPING_FROM_STORAGE", proc134);
  DEFINE_PROC ("DB.DBA.RDF_QM_SET_DEFAULT_MAPPING", proc135);
  DEFINE_PROC ("DB.DBA.RDF_QM_AUTO_LITERAL_FIELD", proc136);
  DEFINE_PROC ("DB.DBA.SPARQL_GRABBER", proc137);
  DEFINE_PROC ("DB.DBA.SPARQL_GRABBER_DEFAULT_RESOLVER", proc138);
  DEFINE_PROC ("DB.DBA.SPARQL_EVAL_TO_ARRAY", proc139);
  DEFINE_PROC ("DB.DBA.SPARQL_EVAL", proc140);
  DEFINE_PROC ("DB.DBA.SPARQL_REXEC_INT", proc141);
  DEFINE_PROC ("DB.DBA.SPARQL_REXEC", proc142);
  DEFINE_PROC ("DB.DBA.SPARQL_REXEC_TO_ARRAY", proc143);
  DEFINE_PROC ("DB.DBA.SPARQL_REXEC_WITH_META", proc144);
  DEFINE_PROC ("querySoap", proc145);
  DEFINE_PROC ("SPARQL_RESULTS_XML_WRITE_NS", proc146);
  DEFINE_PROC ("SPARQL_RESULTS_XML_WRITE_HEAD", proc147);
  DEFINE_PROC ("SPARQL_RESULTS_XML_WRITE_RES", proc148);
  DEFINE_PROC ("SPARQL_RESULTS_XML_WRITE_ROW", proc149);
  DEFINE_PROC ("SPARQL_RESULTS_RDFXML_WRITE_NS", proc150);
  DEFINE_PROC ("SPARQL_RESULTS_RDFXML_WRITE_HEAD", proc151);
  DEFINE_PROC ("SPARQL_RESULTS_RDFXML_WRITE_RES", proc152);
  DEFINE_PROC ("SPARQL_RESULTS_RDFXML_WRITE_ROW", proc153);
  DEFINE_PROC ("SPARQL_RESULTS_TTL_WRITE_NS", proc154);
  DEFINE_PROC ("SPARQL_RESULTS_TTL_WRITE_HEAD", proc155);
  DEFINE_PROC ("SPARQL_RESULTS_TTL_WRITE_RES", proc156);
  DEFINE_PROC ("SPARQL_RESULTS_TTL_WRITE_ROW", proc157);
  DEFINE_PROC ("SPARQL_RESULTS_JAVASCRIPT_HTML_WRITE", proc158);
  DEFINE_PROC ("SPARQL_RESULTS_JSON_WRITE", proc159);
  DEFINE_PROC ("DB.DBA.SPARQL_RESULTS_WRITE", proc160);
  DEFINE_PROC ("WS.WS.SPARQL_VHOST_RESET", proc161);
  DEFINE_PROC ("DB.DBA.SPARQL_PROTOCOL_ERROR_REPORT", proc162);
  DEFINE_PROC ("DB.DBA.SPARQL_WSDL", proc163);
  DEFINE_PROC ("WS.WS./!sparql/", proc164);
  ddl_ensure_table ("do this always", other13);
  DEFINE_PROC ("DB.DBA.TTLP_EXEC_TRIPLE_W", proc165);
  DEFINE_PROC ("DB.DBA.TTLP_EXEC_TRIPLE_W_L", proc166);
  DEFINE_PROC ("DB.DBA.TTLP_EXEC_TRIPLE_A", proc167);
  DEFINE_PROC ("DB.DBA.TTLP_EXEC_TRIPLE_L_A", proc168);
  DEFINE_PROC ("DB.DBA.TTLP_MT", proc169);
  DEFINE_PROC ("DB.DBA.RDF_LOAD_RDFXML_MT", proc170);
  ddl_ensure_table ("DB.DBA.SYS_HTTP_SPONGE", tbl6);
  DEFINE_PROC ("DB.DBA.SYS_HTTP_SPONGE_UP", proc171);
  DEFINE_PROC ("DB.DBA.RDF_SPONGE_GUESS_CONTENT_TYPE", proc172);
  DEFINE_PROC ("DB.DBA.RDF_LOAD_HTTP_RESPONSE", proc173);
  DEFINE_PROC ("DB.DBA.RDF_FORGET_HTTP_RESPONSE", proc174);
  DEFINE_PROC ("DB.DBA.RDF_SPONGE_UP", proc175);
  DEFINE_PROC ("DB.DBA.SPARQL_RELOAD_QM_GRAPH", proc176);
  DEFINE_PROC ("DB.DBA.RDF_CREATE_SPARQL_ROLES", proc177);
}


void
sqls_arfw_define_sparql (void)
{
  /* sparql.sql */

 if (sch_proc_def_exists (bootstrap_cli, "DB.DBA.USER_CREATE"))
  ddl_ensure_table ("do this always", other14);
 if (sch_proc_def_exists (bootstrap_cli, "DB.DBA.XML_URI_GET"))
  ddl_ensure_table ("do this always", other15);
}
